#========================================================================================
# package      DE_EPAGES::AmazonPayments::API::DTO::AuthorizationDetails
# state        private
# description  Data transfer object for the AuthorizationDetails structure.
#========================================================================================
package DE_EPAGES::AmazonPayments::API::DTO::AuthorizationDetails;

use strict;
use DE_EPAGES::AmazonPayments::API::DTO::Amount;
use DE_EPAGES::AmazonPayments::API::DTO::AddressInfo;
use DE_EPAGES::AmazonPayments::API::Constants qw ( AMAZON_AUTHORIZATION_STATUS );
use DE_EPAGES::Core::API::Error qw ( Error );

use metaclass (
    metaclass   => 'Moose::Meta::Class',
    error_class => 'DE_EPAGES::Core::API::MooseError',
);
use Moose;
use Moose::Util::TypeConstraints;

with 'DE_EPAGES::AmazonPayments::API::DTO::ReferenceId';

has authorizationId => ( is => 'ro' );

has authorizationReferenceId => ( is => 'ro' );

has creationTimestamp => ( is => 'ro' );

# authorizationStatus is the external status. Use transStatus to get the internal authorization status
has authorizationStatus => (
    is  => 'ro',
    isa => enum [ values %{AMAZON_AUTHORIZATION_STATUS()} ]
);

has authorizationAmount => (
    is => 'ro',
    isa => 'DE_EPAGES::AmazonPayments::API::DTO::Amount'
);

has billingAddress => (
    is => 'ro',
    isa => 'DE_EPAGES::AmazonPayments::API::DTO::AddressInfo'
);

around BUILDARGS => sub {
    my $orig    = shift;
    my $class   = shift;
    my ($hArgs) = @_;

    if ( $hArgs->{'authorizationStatus'} ) {
        $hArgs->{'authorizationStatus'} = $hArgs->{'authorizationStatus'}->{'state'};
    } else {
        delete $hArgs->{'authorizationStatus'};
    }
    if ( $hArgs->{'authorizationAmount'} ) {
        $hArgs->{'authorizationAmount'} = DE_EPAGES::AmazonPayments::API::DTO::Amount->new( $hArgs->{'authorizationAmount'} );
    } else {
        delete $hArgs->{'authorizationAmount'};
    }
    if ( $hArgs->{'billingAddress'} ) {
        $hArgs->{'billingAddress'} = DE_EPAGES::AmazonPayments::API::DTO::AddressInfo->new( $hArgs->{'billingAddress'} );
    } else {
        delete $hArgs->{'billingAddress'};
    }
    return $class->$orig($hArgs);
};

#========================================================================================
# function     transStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       $TransStatus = $AuthorizationDetails->transStatus;
#----------------------------------------------------------------------------------------
# description  Transforms the value of <function authorizationStatus> into a value that
#               can be used for the attribute LineItemPaymentAmazonPayments.TransStatus.
#----------------------------------------------------------------------------------------
# return       $TransStatus | line item transaction status | string
#========================================================================================
sub transStatus {
    my $self = shift;

    my $Status = $self->authorizationStatus;
    my $hStatusMap = { reverse %{AMAZON_AUTHORIZATION_STATUS()} };
    Error('InvalidAuthStatus', { 'Status' => $Status }) unless exists $hStatusMap->{$Status};

    return $hStatusMap->{$Status};
}

#========================================================================================
# function     refGuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $GUID = $AuthorizationDetails->refGuid;
#----------------------------------------------------------------------------------------
# description  Transforms the value of <function authorizationReferenceId> into the GUID
#               format.
#----------------------------------------------------------------------------------------
# return       $GUID | line item GUID | string
#========================================================================================
sub refGuid {
    my $self = shift;

    return $self->referenceIdToGuid($self->authorizationReferenceId);
}

1;
