#========================================================================================
# package      DE_EPAGES::Core::API::CSV::BaseExportDriver
# state        public
#----------------------------------------------------------------------------------------
# description  Base class for csv export drivers
#========================================================================================
package DE_EPAGES::Core::API::CSV::BaseExportDriver;

use strict;
use Encode qw( encode FB_DEFAULT WARN_ON_ERR );
use File::Basename  qw ( dirname  );
use FileHandle;

use DE_EPAGES::Core::API::CSV;
use DE_EPAGES::Core::API::Constants qw( BOM DEFAULT_ENCODING );
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError);
use DE_EPAGES::Core::API::Iterator qw( ArrayIterator );
use DE_EPAGES::Core::API::Warning qw ( Warning);
use DE_EPAGES::Core::API::File qw ( ExistsFile CreateDirectory SetAccessRights );
use DE_EPAGES::Core::API::TestParameter qw ( TestString );

my %FILE_FORMAT = (
    'eol'           => "\n",
    'encoding'      => DEFAULT_ENCODING,
    'quote_char'    => '"',
    'escape_char'   => '"',
    'sep_char'      => ',',
    'open_mode'     => '>',
    'usebom'        => 1,
);

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'FILE_OPEN_FAILED'  => 'CSV file (#FileName): can\'t open file (#Code)',
        'FILE_CLOSE_FAILED' => 'CSV file (#FileName): can\'t close file (#Code)',
        'WRITE_LINE_FAILED' => 'CSV file (#FileName), line #Line: can\'t write line (#Code)',
        'WRITE_LINE_WARNING' => 'CSV file (#FileName), line #Line: error writing line (#Message) (#Value)',
        'INVALID_CHARACTER'  => 'CSV file (#FileName), line #Line: character #CharacterCode does not map to encoding #Encoding (#Value)',
    };
}
#ERRORDEFINITION

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Exporter = DE_EPAGES::Core::API::CSV::BaseExportDriver->new(%options);
#----------------------------------------------------------------------------------------
# description  create a new csv export driver
#----------------------------------------------------------------------------------------
# input        %options |  exporter options | hash
# return       $Exporter |  exporter | object
#========================================================================================
sub new {
    my $class = shift;
    my %params = @_;
    return bless \%params, $class;
}

#========================================================================================
# function     exportCSV
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Exporter->exportCSV($FileName, $hFormat, $aColumns, $Items );
#----------------------------------------------------------------------------------------
# description  exports data (2 dim. array) into a file.
#               starts the document, calls function parse and closes the document
#----------------------------------------------------------------------------------------
# input        $FileName |  export file | string
# input        $hFormat | export file format
#                   <ul>
#                     <li>eol - end of line (optional, default: "\n")</li>
#                     <li>encoding - character encoding (optional, default: "iso-8859-1")</li>
#                     <li>quote_char - field delimiter (optional, default: '"')</li>
#                     <li>escape_char - escape character (optional, default: '"')</li>
#                     <li>sep_char - field separator (optional, default: ",")</li>
#                     <li>open_mode - file open mode (optional, default: ">")</li>
#                     <li>usebom - write a byte order mark for unicode files if 'utf8' or
#                         encoding  =~ /^utf/i  (optional, default: 1)</li>
#                   </ul>
#                    | ref.hash
# input        $aColumns | column names (used for header) | ref.array
# input        $Items | data for export | DE_EPAGES::Core::API::Iterator or ref.array.array.string
#========================================================================================
sub exportCSV {
    my $self = shift;
    my ($FileName, $hFormat, $aColumns, $Items) = @_;
    TestString('FileName', $FileName);
    $hFormat = {} unless defined $hFormat;
    $hFormat->{$_} = $FILE_FORMAT{$_} foreach grep {not exists $hFormat->{$_}} keys %FILE_FORMAT;

    $self->{'FileName'}   = $FileName;
    $self->{'Writer'}     = DE_EPAGES::Core::API::CSV->new($hFormat);
    $self->{'LineNumber'} = 0;
    $self->{'Columns'}    = $aColumns;
    $self->{'Items'}      = $Items // [];
    $self->{'Encoding'}   = $hFormat->{'encoding'};
    $self->{'NoHeader'}   = ExistsFile($FileName) && $hFormat->{'open_mode'} eq '>>';

    CreateDirectory( dirname($FileName) );

    my $fh = FileHandle->new;
    $self->{'FileHandle'} = $fh;

    open ($fh, $hFormat->{'open_mode'}, $FileName)    ## no critic qw(RequireBriefOpen)
        or Error('FILE_OPEN_FAILED',{'Mode'=>$hFormat->{'open_mode'}, 'FileName'=>$FileName, 'Code'=>$!});
    $self->_print( BOM ) if $self->encodeTo =~ /^utf/i && $hFormat->{'usebom'};
    eval {
        $fh->autoflush(1);
        $self->parse();
    };
    my $Error = ExistsError() ? GetError() : undef;
    close $fh or
        Error('FILE_CLOSE_FAILED',{'Mode'=>$hFormat->{'open_mode'}, 'FileName'=>$FileName, 'Code'=>$!});;
    Error($Error) if defined $Error;

    SetAccessRights( $FileName );

    return;
}

#========================================================================================
# function     items
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $Items = $self->items();
#----------------------------------------------------------------------------------------
# description  return the items to export
#----------------------------------------------------------------------------------------
# return       $Items | parameter of exportCSV | ref.array.array.string or DE_EPAGES::Core::API::Iterator
#========================================================================================
sub items { return shift->{'Items'}; }

#========================================================================================
# function     iterator
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $Items = $self->iterator;
# example      my $Items = $Driver->iterator;
#               while ( my $Item = <$Items> ) {
#                   ProcessItem( $Item );
#               }
#----------------------------------------------------------------------------------------
# description  returns an iterator for items to export
#----------------------------------------------------------------------------------------
# return       $Items | item iterator | ref.array.array.string or DE_EPAGES::Core::API::Iterator
#========================================================================================
sub iterator {
    my $self = shift;

    my $Items = $self->items;

    return (ref($Items) eq 'ARRAY')
        ? ArrayIterator( [ @$Items ] ) # copy the array, because the iterator would destroy it
        : $Items;
}

#========================================================================================
# function     fileHandle
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->fileHandle;
#----------------------------------------------------------------------------------------
# description  return the current fileHandle
#----------------------------------------------------------------------------------------
# return       $FileHandle | setted by exportCSV | filehandle
#========================================================================================
sub fileHandle { return shift->{'FileHandle'}; }

#========================================================================================
# function     fileName
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->fileName;
#----------------------------------------------------------------------------------------
# description  return the current fileName
#----------------------------------------------------------------------------------------
# return       $FileName | export file name | string
#========================================================================================
sub fileName { return shift->{'FileName'}; }

#========================================================================================
# function     writer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Writer = $self->writer;
#----------------------------------------------------------------------------------------
# description  return the current csv writer
#----------------------------------------------------------------------------------------
# return       $Writer
#========================================================================================
sub writer { return shift->{'Writer'}; }

#========================================================================================
# function     columns
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aColumns = $self->columns;
#----------------------------------------------------------------------------------------
# description  return the column name array
#----------------------------------------------------------------------------------------
# return       $aColumns
#========================================================================================
sub columns { return shift->{'Columns'}; }

#========================================================================================
# function     encodeTo
# state        public
#----------------------------------------------------------------------------------------
# syntax       encodeTo();
# example      $Driver->encodeTo();
#----------------------------------------------------------------------------------------
# description  returns the target encoding (was setted in <function exportCSV>)
#----------------------------------------------------------------------------------------
# return       $Encoding | target encoding (utf8,iso-8859-1,...) | string
#========================================================================================
sub encodeTo { return shift->{'Encoding'}; }

#========================================================================================
# function     incLineNumber
# state        public
#----------------------------------------------------------------------------------------
# syntax       $linenumber = $self->incLineNumber;
#----------------------------------------------------------------------------------------
# description  return the incremented linenumber
#----------------------------------------------------------------------------------------
# return       $linenumber |
#========================================================================================
sub incLineNumber { return ++shift->{'LineNumber'}; }

#========================================================================================
# function     lineNumber
# state        public
#----------------------------------------------------------------------------------------
# syntax       $linenumber = $self->lineNumber;
#----------------------------------------------------------------------------------------
# description  return the current linenumber
#----------------------------------------------------------------------------------------
# return       $linenumber |
#========================================================================================
sub lineNumber { return shift->{'LineNumber'}; }

#========================================================================================
# function     writeLine
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->writeLine($aLineElements);
#----------------------------------------------------------------------------------------
# description  exports a line ( converts and writes to the filehandle )
#----------------------------------------------------------------------------------------
# input        $aLineElements | list of strings for 1 line | ref.array.string
#========================================================================================
sub writeLine {
    my $self = shift;
    my ($aLineElements)  = @_;

    my $value = $self->writer->combine( $aLineElements );

    my $Warning;
    local $SIG{__WARN__} = sub {
        my ($Message) = @_;

        return if defined $Warning; # remember only one warning per line

        $Message =~ s/ at \S+ line \d+\b.*//; # remove code line
        if ($Message =~ /"\\x\{([0-9A-Fa-f]+)\}"\ does\ not\ map\ to\ (\S+)/ix ){
            my $CharacterCode = $1;
            my $Character = chr(hex($CharacterCode));
            my $Encoding  =$2 ;
            $Warning = DE_EPAGES::Core::API::Error->new(
                'Code' => 'INVALID_CHARACTER',
                'Vars' => {
                    'FileName' => $self->fileName,
                    'CharacterCode'=> $CharacterCode,
                    'Character'=> $Character,
                    'Encoding' => $Encoding,
                    'Line' => $self->lineNumber,
                    'Value' => $value,
                },
            );
        }
        else {
            $Warning = DE_EPAGES::Core::API::Error->new(
                'Code' => 'WRITE_LINE_WARNING',
                'Vars' => {
                    'FileName' => $self->fileName,
                    'Message' => $Message,
                    'Line' => $self->lineNumber,
                    'Value' => $value,
                },
            );
        }
    };

    $self->_print($value) or Error( 'WRITE_LINE_FAILED', {
        'FileName' => $self->fileName,
        'Code' => $!,
        'Line' => $self->lineNumber,
        'Value' => $value
    });

    Warning( $Warning ) if defined $Warning;
    $self->incLineNumber;
    return;
}

# write to file using the selected encoding
sub _print {
    my $self = shift;
    my ($Text) = @_;

    # substitute invalid character with '?' instead of \x{0000}
    my $Bytes = encode( $self->encodeTo, $Text, FB_DEFAULT|WARN_ON_ERR );
    return $self->fileHandle->print( $Bytes );
}

#========================================================================================
# function     writeHeader
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->writeHeader();
#----------------------------------------------------------------------------------------
# description  writes the header
#========================================================================================
sub writeHeader {
    my $self = shift;

    $self->writeLine($self->columns);
    return;
}

#========================================================================================
# function     exportLine
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportLine();
#----------------------------------------------------------------------------------------
# description  exports one data line, get the line element from function topID('Item')
#========================================================================================
sub exportLine {
    my $self = shift;

    my $aValues = $self->topID('Item');
    $self->writeLine( $aValues );
    return;
}

#========================================================================================
# function     writeLines
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->writeLines();
#----------------------------------------------------------------------------------------
# description  writes the content lines
#               each item of $self-><function iterator> is exported via <function exportLine>.
#========================================================================================
sub writeLines {
    my $self = shift;

    my $Items = $self->iterator;
    while( my $Item = <$Items> ) {
        $self->pushID('Item', $Item);
        $self->exportLine();
        $self->popID('Item');
    }
    return;
}

#========================================================================================
# function     drive
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->drive();
#----------------------------------------------------------------------------------------
# description  write the data content lines
#========================================================================================
sub drive {
    my $self = shift;

    $self->writeLines();
    return;
}

#========================================================================================
# function     parse
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->parse();
#----------------------------------------------------------------------------------------
# description  calls function writeHeader, calls function drive
#========================================================================================
sub parse {
    my $self = shift;

    $self->writeHeader() unless $self->{'NoHeader'};
    $self->drive();
    return;
}

#========================================================================================
# function     pushID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->pushID( $Key, $Value );
# example      $self->pushID( 'CompanyID', $CompanyID );
#----------------------------------------------------------------------------------------
# description  adds a parameter to the parameter stack. Any previous parameter with the
#               same name will be hidden until $self->popID( $Key ) is called.
#               Use $self->topID( $Key ) to get the parameter value.
#               Use $self->popID( $Key ) to remove the parameter from the stack.
#----------------------------------------------------------------------------------------
# input        $Key | parameter name | string
# input        $Value | parameter value
#========================================================================================
sub pushID {
    my $self = shift;
    my ($Key, $Value) = @_;

    $self->{'IDs'}->{$Key} = [] unless exists $self->{'IDs'}->{$Key};
    push @{$self->{'IDs'}->{$Key}}, $Value;
}

#========================================================================================
# function     popID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Value = $self->popID( $Key );
# example      $CompanyID = $self->popID( 'CompanyID' );
#----------------------------------------------------------------------------------------
# description  removes a parameter from the parameter stack and returns its value.
#----------------------------------------------------------------------------------------
# input        $Key | parameter name | string
# return       $Value | parameter value
#========================================================================================
sub popID {
    my $self = shift;
    my ($Key) = @_;

    return pop @{$self->{'IDs'}->{$Key}};
}

#========================================================================================
# function     topID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Value = $self->topID( $Key );
# example      $CompanyID = $self->topID( 'CompanyID' );
#----------------------------------------------------------------------------------------
# description  returns the current value of a parameter
#----------------------------------------------------------------------------------------
# input        $Key | parameter name | string
# return       $Value | parameter value
#========================================================================================
sub topID {
    my $self = shift;
    my ($Key) = @_;

    return () unless exists $self->{'IDs'}->{$Key};
    return $self->{'IDs'}->{$Key}->[-1];
}

1;
