#========================================================================================
# package      DE_EPAGES::Core::API::MooseError
# state        public
# description  Error handling for Exceptions raised by Moose. It uses the error code
#               "MooseDie". To use it, you have to assign it to the error_class handler
#               in your moose package.
#
#               Newer versions of Moose have replaced all string errors with error exceptions,
#               this no longer works for these versions. Remove the base class so that it doesn't fail
#               compilation when using these newer versions (but checked using PERL version)
#
# example      package MyPackage;
#               use metaclass (
#                   metaclass   => 'Moose::Meta::Class',
#                   error_class => 'DE_EPAGES::Core::API::MooseError',
#               );
#               use Moose;
#               has 'Bool', 'isa' => 'Bool', 'is' => 'rw'; # some moose code.
#               no Moose;
#========================================================================================
package DE_EPAGES::Core::API::MooseError;

use strict;
use DE_EPAGES::Core::API::Constants qw( PERL538 );

our $baseMooseClass; ## no critic qw(ProhibitPackageVars)
BEGIN {
    $baseMooseClass = $] < PERL538 ? 'Moose::Error::Default' : undef;
}

use base $baseMooseClass;
use DE_EPAGES::Core::API::Error;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Error = $class->new( %args  );
#----------------------------------------------------------------------------------------
# description  Constructor called by Moose::Meta::Class. Creates an epages error object
#----------------------------------------------------------------------------------------
# input        %args | error arguments | hash
# return       $Error | Error object | DE_EPAGES::Core::API::Error
#========================================================================================
sub new {
    my $class = shift;
    my ( %args ) = @_;
    my $Error = DE_EPAGES::Core::API::Error->new(
        'Code'        => 'MooseDie',
        'Message'     => delete $args{'message'},
        'SkipCallers' => (delete $args{'depth'}) + 1,
        'Vars'        => \%args,
    );
    return $Error;
}

1;
