#========================================================================================
# package      DE_EPAGES::Core::API::Object::DateTimeFormatter
# state        public
#----------------------------------------------------------------------------------------
# description  Formatter and parser for locale-dependent time and date strings. Based on
#               DateTime::Format::Strptime.
#----------------------------------------------------------------------------------------
# example      my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new(
#                   locale => 'de_DE',
#               );
#               my $Date   = $Formatter->parse_datetime( 'Donnerstag, 7. Oktober 2004', 'date' );
#               my $String = $Formatter->format_datetime( $Date, 'date', 'long' );
#========================================================================================
package DE_EPAGES::Core::API::Object::DateTimeFormatter;
use base DateTime::Format::Strptime;

use strict;
use DateTime::Locale;
use Time::Piece;
use DE_EPAGES::Core::API::Constants qw( PERL538 );
use DE_EPAGES::Core::API::Error qw (CallerError Error ExistsError GetError);
use DE_EPAGES::Core::API::PerlTools qw (GetLocalTimeZone GetLocalLocale GetCurrentDateTime IsA);
use DE_EPAGES::Core::API::TestParameter qw ( TestString );

my %FORMAT_PARSE_MAP = (
    '%{hour}'                => '%H',
    '%{time_zone_long_name}' => '%z',
    '%y'             => '%{ce_year}',
);

#ERRORDEFINITION
sub ErrorDefinition { return {
    'FORMAT_NOT_DATETIME' => 'datetime \'#Dt\' doesn\'t match format \'#Format\'',
    'FORMAT_UNKNOWN' => 'unknown datetime format: #Type (allowed formats are: #Allowed)',
    }
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new(%options);
#----------------------------------------------------------------------------------------
# description  Creates a formatter for dates and/or time strings.
#----------------------------------------------------------------------------------------
# input        %options | keys
#               <ul>
#                   <li>locale       - (optional) locale id, example: 'en_US',
#                                                 default: server locale - string</li>
#                   <li>time_zone    - (optional) time zone, example: '+01:00',
#                                                 default: server time zone - string</li>
#               </ul>| hash
# return       $Formatter | formatter object | object
#========================================================================================
sub new {
    my $class = shift;
    my (%options) = @_;
    $options{'time_zone'} = GetLocalTimeZone() unless defined $options{'time_zone'};
    $options{'locale'}    = GetLocalLocale()   unless defined $options{'locale'};
    $options{'pattern'}   = '%u';
    return bless DateTime::Format::Strptime->new(%options), $class;
}

#========================================================================================
# function     formatString
# state        public
#----------------------------------------------------------------------------------------
# syntax       $formatString = formatString($type, $formattype, $locale);
#----------------------------------------------------------------------------------------
# description  returns strptime pattern for $type and $formattype
#----------------------------------------------------------------------------------------
# input        $type | date/time/datetime | string
# input        $formattype | short/medium/long/full (optional short)| string
# input        $locale | locale id as 'en_US', default: server locale | string
# return       $formatString | pattern for DateTime::Format::Strptime | string
#========================================================================================
sub formatString { ##no critic qw(RequireArgUnpacking)
    my $self = ref $_[0] ? shift : undef;
    my ($type, $formattype, $locale) = @_;
    TestString('type',$type);

    if ($type !~ /^(?:date|time|datetime)$/) {
        CallerError('FORMAT_UNKNOWN', {
            'Type'   => $type,
            'Allowed' => 'date, datetime or time'
        })
    }

    if ( $] < PERL538 ) {
        $locale = DateTime::Locale->load($self->locale);
    } else {
        $locale = $self ? $self->locale : GetLocalLocale() unless defined $locale;
    }

    $formattype = 'short' unless defined $formattype;
    my $functionname = $type.'_format_'.$formattype;

    my $formatString;
    # $locale->_cldr_to_strftime does not exist anymore with 5.38.0
    if ( $] < PERL538 ) {
        $formatString = $locale->_cldr_to_strftime($locale->$functionname()); ## no critic qw(ProhibitPrivateSubroutineUse)
        $formatString =~ s/\Q$_\E/$FORMAT_PARSE_MAP{$_}/ foreach keys %FORMAT_PARSE_MAP;
    } else {
        my $dtLocale = DateTime::Locale->load($locale);
        $formatString = _cldr_to_strftime($dtLocale->$functionname);
    }

    # time zones work only if datetime is given (not just time)
    if ($type eq 'time' and $formatString =~ /%z/ and $formatString !~ /^%Y-%m-%d / ) {
        $formatString = '%Y-%m-%d ' . $formatString;
    }

    return $formatString;
}

#========================================================================================
# function     format_datetime
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FormattedDateTime = $Formatter->format_datetime($dt, $type, $formattype);
#----------------------------------------------------------------------------------------
# description  Formats datetime
#----------------------------------------------------------------------------------------
# input        $dt | datetime | object
# input        $type | date/time/datetime | string
# input        $formattype | (optional) short/medium/long/full
#                             or custom strftime format, for example '%Y-%m-%d'
#                             (default: short)| string
# return       $FormattedDateTime | formatted datetime | string
#========================================================================================
sub format_datetime {
    my $self = shift;
    my ($dt, $type, $formattype) = @_;
    die 'no DateTime object' unless IsA( $dt, 'DateTime' );
    TestString('type',$type);

    if ($type !~ /^(?:date|time|datetime)$/) {
        CallerError('FORMAT_UNKNOWN', {
            'Type'   => $type,
            'Allowed' => 'date, datetime or time'
        })
    }

    my $Locale = $self->locale;
    my $dtLocale = $] < PERL538 ? $Locale : $dt->locale->code;

    # $dt->set(locale => $Locale) does not work anymore with 5.38.0
    $dt->set(locale => $Locale) if $] < PERL538;
    $dt->set_locale($Locale) if $Locale ne $dtLocale;

    my $formatString;
    if (defined $formattype and $formattype =~ /%/) {
        $formatString = $formattype;
    } else {
        $formatString = $self->formatString($type, $formattype, $Locale);
    }

    my $strftime = $dt->strftime( $formatString );
    $dt->set_locale($dtLocale) if $Locale ne $dtLocale;

    # strip leading and trailing spaces
    $strftime =~ s/^\s*(.*?)\s*$/$1/x;
    # strip spaces before numbers if $formatString contains %e/%k/%l
    $strftime =~ s,([:/.-]|\s)\s([1-9])\b,$1$2,gx if $formatString =~ /%[ekl]/;;

    return  $strftime;
}

#========================================================================================
# function     parse_datetime
# state        public
#----------------------------------------------------------------------------------------
# syntax       $dt = $Formatter->parse_datetime($FormattedDateTime, $type, $formattype);
#----------------------------------------------------------------------------------------
# description  Return the datetime object of a formatted datetime
#----------------------------------------------------------------------------------------
# input        $FormattedDateTime | formatted datetime | string
# input        $type | date/time/datetime | string
# input        $formattype | (optional) short/medium/long/full
#                             or custom strftime format, for example '%Y-%m-%d'
#                             (default: short)| string
# return       $dt | datetime object | DateTime object
#========================================================================================
sub parse_datetime {
    my $self = shift;
    my ($string, $type, $formattype) = @_;
    TestString('type', $type);

    if ($type !~ /^(?:date|time|datetime)$/) {
        CallerError('FORMAT_UNKNOWN', {
            'Type'   => $type,
            'Allowed' => 'date, datetime or time'
        })
    }

    my $formatString;
    if (defined $formattype and $formattype =~ /%/) {
        $formatString = $formattype;
    } else {
        $formatString = $self->formatString($type, $formattype);
    }

    # time zones work only if datetime is given (not just time)
    if ($type eq 'time' and $formatString =~ /%z/ and $string !~ /^(19|20)\d\d-[01]\d-[0-3]\d /) {
        my $Now = GetCurrentDateTime;
        $string = $Now->ymd('-') . ' ' . $string;
    }

    my $dt;
    if ($] < PERL538) {
        # $self->pattern($formatString) does not work anymore with 5.38.0
        $self->pattern($formatString);
        $dt = eval { return $self->SUPER::parse_datetime($string); };
        Error( GetError() ) if ExistsError();
        Error('FORMAT_NOT_DATETIME', { 'Dt'=> $string, 'Format' => $formatString })
            if not defined $dt;
    } else {
        my $dtLocale = DateTime::Locale->load($self->locale);
        Time::Piece::mon_list(@{$dtLocale->month_format_abbreviated});
        Time::Piece::day_list(@{$dtLocale->day_format_abbreviated});
        Time::Piece::fullmon_list(@{$dtLocale->month_format_wide});
        Time::Piece::fullday_list(@{$dtLocale->day_format_wide});

        my $parsed_date = eval { return Time::Piece->strptime($string, $formatString); };
        Error('FORMAT_NOT_DATETIME', { 'Dt'=> $string, 'Format' => $formatString })
            if ExistsError() or not defined $parsed_date;

        $dt = DateTime->new(
            year       => $parsed_date->year,
            month      => $parsed_date->mon,
            day        => $parsed_date->mday,
            hour       => $parsed_date->hour,
            minute     => $parsed_date->minute,
            second     => $parsed_date->second,
            time_zone  => $self->time_zone,
            locale     => $self->locale,
            );
        # Time::Piece has problems with time zones
        $dt->add(seconds => $dt->offset) if $formatString =~ /%z/;
    }
    return $dt;
}

# private functions from DateTime::Locale::Base

sub _cldr_to_strftime {
    my ($simple) = @_;
    $simple
        =~ s/(G+|y+|u+|M+|d+|h+|H+|m+|s+|S+|E+|D+|F+|w+|W+|a+|k+|K+|z+)|'((?:[^']|'')*)'/
                        $2 ? _stringify($2) : $1 ? _convert($1) : "'"/eg;
    $simple =~ s,\x{202f}, ,g;
    return $simple;
}

sub _convert {
    my ($simple) = @_;
    # see  https://metacpan.org/pod/DateTime#CLDR-Patterns
      my @JavaPatterns = (
        qr/G/    => '{era}',
        qr/yyyy/ => 'Y',
        qr/yy/   => 'Y', # always use 4 digits for year
        qr/y/    => 'Y', # always use 4 digits for year
        qr/u/    => 'Y',
        qr/MMMM/ => 'B',
        qr/MMM/  => 'b',
        qr/MM/   => 'm',
        qr/M/    => 'm', # was: {month}
        qr/dd/   => 'd',
        qr/d/    => 'e', # was: {day}
        qr/hh/   => 'l',
        qr/h/    => 'l', # was: {hour_12}
        qr/HH/   => 'H',
        qr/H/    => 'H', # was: {hour}
        qr/mm/   => 'M',
        qr/m/    => 'M', # was: {minute}
        qr/ss/   => 'S',
        qr/s/    => 'S', # was: {second}
        qr/S/    => 'N',
        qr/EEEE/ => 'A',
        qr/E/    => 'a',
        qr/D/    => 'j',
        qr/F/    => '{weekday_of_month}',
        qr/w/    => 'V',
        qr/W/    => '{week_month}',
        qr/a/    => 'p',
        qr/k/    => 'k', # was: {hour_1}
        qr/K/    => 'l', # was: {hour_12_0}
        qr/zzzz/ => 'z', # was: {time_zone_long_name}
        qr/z/    => 'z', # was: {time_zone_long_name}
    );
    for ( my $x = 0; $x < @JavaPatterns; $x += 2 ) {
        return '%' . $JavaPatterns[ $x + 1 ] if $simple =~ /$JavaPatterns[$x]/;
        return '%' . $JavaPatterns[ $x + 1 ] if $simple =~ /$JavaPatterns[$x]/;
    }
    die "**Dont know $simple***";
}

sub _stringify {
    my ($string) = @_;
    $string =~ s/%(?:[^%])/%%/g;
    $string =~ s/\'\'/\'/g;
    return $string;
}

1;
