#========================================================================================
# package      DE_EPAGES::Core::API::Transformation::Color
# state        public
#----------------------------------------------------------------------------------------
# description  This module provides color transformation functionality.
#========================================================================================
package DE_EPAGES::Core::API::Transformation::Color;
use base qw( DE_EPAGES::Core::API::Transformation::Transformation Exporter);
use strict;

our @EXPORT_OK = qw(
    String2Color
    Hsv2Rgb
    Rgb2Hsv
    hardFloor
);

use Image::Magick qw ();
use DE_EPAGES::Core::API::Error qw (Error);
use DE_EPAGES::Core::API::TestParameter qw (TestInteger);
use List::Util qw (min max);

#ERRORDEFINITION
sub ErrorDefinition () {
    return {
    }
}
#ERRORDEFINITION END

##no critic qw(ProhibitMagicNumbers)

#========================================================================================
# function     ReplacableAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hReplacableAttributes = $Transformation->ReplacableAttributes();
#----------------------------------------------------------------------------------------
# description  Returns a hash containing all parameter names to be converted to another
#               object
#----------------------------------------------------------------------------------------
# return       $hReplacableAttributes | hash of parameter names indexed by function name
#               | hash.hash
#========================================================================================
sub ReplacableAttributes {
    return {
        # user defined functions -- in this file
        'OverlayColor'      => {'Color' => 'Color'},
    }
}

#========================================================================================
# function     Transform
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Transformation->Transform($hParams, $hObjects, $hVars);
#----------------------------------------------------------------------------------------
# description  Process the function collection of this transformation.
#----------------------------------------------------------------------------------------
# input        $hParams | hash of function parameters | hash
# input        $hObjects | hash of objects indexed by alias | hash
# input        $hVars | some vars provided by the transformation module | hash
#========================================================================================
sub Transform {
    my $self = shift;
    my ($hParams, $hObjects, $hVars) = @_;

    # get color value
    my $Color = String2Color($hObjects->{$hParams->{'Source'}}->{'Value'});

    foreach my $hFunctionParams (@{$hParams->{'Functions'}}) {
        my $FunctionName = $hFunctionParams->{'Function'};
        # remove function name in attributes
        # replace object ids with objects in parameters
        $self->ReplaceObjects($FunctionName, $hFunctionParams, $hObjects, $hVars);

        if ($self->can($FunctionName)) {
            $Color = $self->$FunctionName($Color, $hFunctionParams);
        } else {
            Error( 'UNKNOWN_COLOR_FUNCTION', {
                'Function' => $FunctionName,
            });
        }
    }

    # set color value
    unless (exists($hObjects->{$hParams->{'Target'}})) {
        Error('UNKNOWN_OBJECT', { 'Alias' => $hParams->{'Target'} });
    }
    $hObjects->{$hParams->{'Target'}}->{'Value'} = Color2String($Color);

    return;
}

#========================================================================================
# function     Finalize
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Transformation->Finalize($hParams, $hVars);
#----------------------------------------------------------------------------------------
# description  Called for each object of this type after processing to cleanup and (e.g.
#               delete temporary files) and return another representation of this object
#               if needed. This is used as the result of the calling Transform function.
#----------------------------------------------------------------------------------------
# input        $hParams | hash of function parameters | hash
# input        $hVars | some vars provided by the transformation module | hash
# return       $Value | the return value of this transformation | string
#========================================================================================
sub Finalize {
    my $self = shift;
    my ($hParams, $hVars) = @_;

    my $Color = String2Color($hParams->{'Value'});
    # return as format #RRGGBB - no alpha
    if ($Color->{'A'} == 0.0) {
        $Color = 'transparent';
    } else {
        $Color = sprintf('#%.2X%.2X%.2X',
                        $Color->{'R'} * 255,
                        $Color->{'G'} * 255,
                        $Color->{'B'} * 255);
    }
    return $Color;
}

#========================================================================================
# function     String2Color
# state        private
#----------------------------------------------------------------------------------------
# syntax       String2Color($ColorString);
#----------------------------------------------------------------------------------------
# description  Convert a string to a color hash with keys R, G, B and A. The string has
#               to be in a Image::Magick color notation,
#               see http://www.imagemagick.org/www/color.html for details.
#----------------------------------------------------------------------------------------
# input        $ColorString | color string | string
# return       $hColor | hash with keys R, G, B and A; R, G, B ranges from 0.0 - 1.0;
#               A == 1.0 means totally opaque, A == 0.0 means
#               totally transparent | hash
#========================================================================================
sub String2Color {
    my ($ColorString) = @_;

    my ($R, $G, $B, $A);

    if ($ColorString =~ /^#([0-9A-Z]{2})([0-9A-Z]{2})([0-9A-Z]{2})([0-9A-Z]{2})?$/i) {
        # #RRGGBB, #RRGGBBAA
        ($R, $G, $B, $A) = ($1, $2, $3, $4);
        $R = hex($R) / 255;
        $G = hex($G) / 255;
        $B = hex($B) / 255;
        $A = (defined $A) ? hex($A) / 255 : 1.0;
    }
    elsif ($ColorString =~ /^#([0-9A-Z])([0-9A-Z])([0-9A-Z])([0-9A-Z])?$/i) {
        # #RGB, #RGBA
        ($R, $G, $B, $A) = ($1, $2, $3, $4);
        $R = hex($R) / 15;
        $G = hex($G) / 15;
        $B = hex($B) / 15;
        $A = (defined $A) ? hex($A) / 15 : 1.0;
    }
    elsif ($ColorString =~ /^rgb\((\d{1,3}),(\d{1,3}),(\d{1,3})(?:,(\d{1,3}))?\)$/i) {
        # rgb(R,G,B), rgb(R,G,B,A)
        ($R, $G, $B, $A) = ($1, $2, $3, $4);
        $R = int($R) / 255;
        $G = int($G) / 255;
        $B = int($B) / 255;
        $A = defined $A ? int($A) / 255 : 1.0;
    }
    elsif ($ColorString =~ /^rgba\((\d{1,3}),(\d{1,3}),(\d{1,3}),(\d+(?:\.\d+))\)$/i) {
        # rgba(R,G,B,A)
        ($R, $G, $B, $A) = ($1, $2, $3, $4);
        $R = int($R) / 255;
        $G = int($G) / 255;
        $B = int($B) / 255;
    }
    elsif ($ColorString =~ /none|transparent/) {
        # transparent
        ($R, $G, $B, $A) = (0, 0, 0, 0);
    }
    else {
        # don't know what it is - try it via Image::Magick
        my $Image = Image::Magick->new('size' => '1x1');
        $Image->Read('xc:none');
        $Image->Set('matte' => 'True', 'alpha' => 'Activate');
        $Image->Draw(fill => $ColorString, primitive => 'rectangle', points => '0,0 1,1');
        # decrease the precision by 2**QuantumDepth-1
        ($R, $G, $B, $A) =
            map { $_ / (2 ** Image::Magick->QuantumDepth - 1) }
            split /,/, $Image->Get('Pixel[0,0]');
        $A = 1 - $A;
    }

    return {R => $R, G => $G, B => $B, A => $A};
}

#========================================================================================
# function     Color2String
# state        private
#----------------------------------------------------------------------------------------
# syntax       Color2String($RgbColor);
#----------------------------------------------------------------------------------------
# description  Convert a color hash to a string with notation "rgb(r,g,b,a)".
#----------------------------------------------------------------------------------------
# input        $RgbColor | color hash, keys R, G, B and A | hash
# return       $ColorString | color string with notation rgb(r,g,b,a) | string
#========================================================================================
sub Color2String {
    my ($Rgb) = @_;

    # use this format for alpha support
    return sprintf('rgba(%u,%u,%u,%f)',
                    $Rgb->{'R'} * 255,
                    $Rgb->{'G'} * 255,
                    $Rgb->{'B'} * 255,
                    $Rgb->{'A'});
}

#========================================================================================
# function     hardFloor
# state        private
#----------------------------------------------------------------------------------------
# syntax       hardFloor($Value);
#----------------------------------------------------------------------------------------
# description  Returns the floor value even for pseude integers (floats with a value
#               very close to an integer, are sometimes displayed as integer, but int()
#               returns wrong values).
#----------------------------------------------------------------------------------------
# input        $Value | a wannabe integer | scalar
# return       $IntValue | floored value | int
#========================================================================================
sub hardFloor {
    my ($Value) = @_;

    $Value .= '.';
    if( $Value =~ /(.*?)\..*/) {
        $Value = $1;
    }
    return int($Value);
}

#========================================================================================
# function     Rgb2Hsv
# state        private
#----------------------------------------------------------------------------------------
# syntax       Rgb2Hsv($Rgb);
#----------------------------------------------------------------------------------------
# description  Convert a RGB color hash into HSV, alpha channel is not touched.
#----------------------------------------------------------------------------------------
# input        $hRgb | RGB color | hash
# return       $hHsv | hash with keys H, S, V and A; all values from 0.0 - 1.0 | hash
#========================================================================================
sub Rgb2Hsv {
    my ($Rgb) = @_;

    my ($R, $G, $B) = map { $Rgb->{$_} } qw(R G B);
    my ($H, $S, $V, $Min, $Max, $Diff, $DR, $DG, $DB);

    $Min = min($R, $G, $B);
    $Max = max($R, $G, $B);
    $Diff = $Max - $Min;

    $V = $Max;

    if ($Diff == 0) {
        # gray
        $H = 0;
        $S = 0;
    } else {
        # chromatic color
        $S = $Diff / $Max;

        $DR = ((($Max - $R) / 6) + ($Diff / 2)) / $Diff;
        $DG = ((($Max - $G) / 6) + ($Diff / 2)) / $Diff;
        $DB = ((($Max - $B) / 6) + ($Diff / 2)) / $Diff;

        if ($R == $Max) {
            $H = $DB - $DG;
        } elsif ($G == $Max ) {
            $H = (1 / 3) + $DR - $DB;
        } elsif ($B == $Max) {
            $H = (2 / 3) + $DG - $DR;
        }

        $H += 1 if ($H < 0);
        $H -= 1 if ($H > 1);
    }
    return {H => $H, S => $S, V => $V, A => $Rgb->{'A'}};
}

#========================================================================================
# function     Hsv2Rgb
# state        private
#----------------------------------------------------------------------------------------
# syntax       Hsv2Rgb($Rgb);
#----------------------------------------------------------------------------------------
# description  Convert a HSV color hash into RGB, alpha channel is not touched.
#----------------------------------------------------------------------------------------
# input        $hHsv | HSV color | hash
# return       $hRgb | hash with keys R, G, B and A; all values from 0.0 - 1.0 | hash
#========================================================================================
sub Hsv2Rgb {
    my ($Hsv) = @_;

    my ($H, $S, $V) = map { $Hsv->{$_} } qw(H S V);
    my ($R, $G, $B, $T1, $T2, $T3, $TI, $TH);

    if ($S == 0) {
        $R = $V;
        $G = $V;
        $B = $V;
    } else {
        $TH = $H * 6;
        $TH = 0 if ($TH == 6);
        $TI = int($TH);
        $T1 = $V * (1 - $S);
        $T2 = $V * (1 - $S * ($TH - $TI));
        $T3 = $V * (1 - $S * (1 - ($TH - $TI)));

        if ($TI == 0) {
            ($R, $G, $B) = ($V, $T3, $T1);
        } elsif ($TI == 1) {
            ($R, $G, $B) = ($T2, $V, $T1);
        } elsif ($TI == 2) {
            ($R, $G, $B) = ($T1, $V, $T3);
        } elsif ($TI == 3) {
            ($R, $G, $B) = ($T1, $T2, $V);
        } elsif ($TI == 4) {
            ($R, $G, $B) = ($T3, $T1, $V);
        } else {
            ($R, $G, $B) = ($V, $T1, $T2);
        }
    }
    return { R => $R, G => $G, B => $B, A => $Hsv->{'A'} };
}

#========================================================================================
# function     AddColorOffset
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hColor = $Transformation->AddColorOffset($Color, $hAttributes);
#----------------------------------------------------------------------------------------
# description  Adds an HSV offset to the given color.
#----------------------------------------------------------------------------------------
# input        $hColor | color hash | hash
# input        $hAttributes | function attributes | hash
# return       $hColor | RGBA color string | string
#========================================================================================
sub AddColorOffset {
    my $self = shift;
    my ($Color, $hAttributes) = @_;

    # convert color from RGB to HSV
    $Color = Rgb2Hsv($Color);

    my $DiffH = $hAttributes->{'Hue'};
    my $DiffS = $hAttributes->{'Saturation'};
    my $DiffB = $hAttributes->{'Brightness'};

    $Color->{'H'} += $DiffH / 360 if defined $DiffH;
    $Color->{'S'} += $DiffS / 100 if defined $DiffS;
    $Color->{'V'} += $DiffB / 100 if defined $DiffB;

    # e.g. -0.1 => 0.9
    $Color->{'H'} += 1.0 while ($Color->{'H'} < 0.0);
    $Color->{'H'} -= 1.0 while ($Color->{'H'} > 1.0);
    $Color->{'S'} = max(0.0, min($Color->{'S'}, 1.0));
    $Color->{'V'} = max(0.0, min($Color->{'V'}, 1.0));

    return Hsv2Rgb($Color);
}

#========================================================================================
# function     OverlayColor
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hColor = $Transformation->OverlayColor($Color, $hAttributes);
#----------------------------------------------------------------------------------------
# description  Adds an overlay color with a specified transparency.
#----------------------------------------------------------------------------------------
# input        $hColor | color hash | hash
# input        $hAttributes | function attributes | hash
# return       $hColor | RGBA color string | string
#========================================================================================
sub OverlayColor {
    my $self = shift;
    my ($Color, $hAttributes) = @_;

    my $OverlayColor = String2Color($hAttributes->{'Color'});
    my $T = $hAttributes->{'Transparency'};
    TestInteger('T', $T, 0, 100);
    $T /= 100;

    my $R = ($Color->{'R'} * (1.0 - $T)) + ($OverlayColor->{'R'} * $T);
    my $G = ($Color->{'G'} * (1.0 - $T)) + ($OverlayColor->{'G'} * $T);
    my $B = ($Color->{'B'} * (1.0 - $T)) + ($OverlayColor->{'B'} * $T);

    return {
        'R' => $R,
        'G' => $G,
        'B' => $B,
        'A' => 1.0
    };
}

#========================================================================================
# function     Set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hColor = $Transformation->Set($Color, $hAttributes);
#----------------------------------------------------------------------------------------
# description  Sets one or more values of a color component.
#----------------------------------------------------------------------------------------
# input        $hColor | color hash | hash
# input        $hAttributes | function attributes | hash
# return       $hColor | RGBA color string | string
#========================================================================================
sub Set {
    my $self = shift;
    my ($Color, $hAttributes) = @_;

    my $hAttributesMap = {
        'Red' => 'R',
        'R' => 'R',
        'Green' => 'G',
        'G' => 'G',
        'Blue' => 'B',
        'B' => 'B',
        'Alpha' => 'A',
        'A' => 'A',
    };

    foreach my $key (keys %$hAttributesMap) {
        $Color->{$hAttributesMap->{$key}} = $hAttributes->{$key} if (exists $hAttributes->{$key});
    }

    return $Color;
}

1;
