use strict;
use Test::More tests => 42;
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::String qw( QuoteXML UnquoteXML SplitEMailAddresses Slice
                                     Substr CompareStrings sortCSVLine sortXMLAttributes );
use Math::BigFloat;

##no critic qw(ProhibitMagicNumbers ProhibitInterpolationOfLiterals ProhibitLongLines)

sub Main {

    # sortXMLAttributes

    my $TestString = '<s1 x="eee" b=\'rrr\'/>';
    ok( sortXMLAttributes($TestString) eq $TestString, "not sorted: $TestString");
    $TestString = '<s1 x=eee b=rrr/>';
    ok( sortXMLAttributes($TestString) eq $TestString, "not sorted: $TestString");
    $TestString = '<s1 x="eee" b="rrr">FOO</s1>';
    ok( sortXMLAttributes($TestString) eq $TestString, "not sorted: $TestString");

    $TestString  = '   <s1 x="e\'ee" b="rrr" y="e\'ee" >  ';
    my $Expected = '   <s1 b="rrr" x="e\'ee" y="e\'ee" >  ';
    ok( sortXMLAttributes($TestString) eq $Expected, "sorted: $TestString");

    $TestString = "   <s1   x='e\"ee' b='rrr'   y='e\"ee' >  \n\n\n";
    $Expected   = "   <s1   b='rrr' x='e\"ee' y='e\"ee' >  \n\n\n";
    ok( sortXMLAttributes($TestString) eq $Expected, "sorted: $TestString");

    $TestString = "   <s1   x='e\"ee' b='rrr'\n\ny='e\"ee' >";
    $Expected   = "   <s1   b='rrr' x='e\"ee' y='e\"ee' >";
    ok( sortXMLAttributes($TestString) eq $Expected, "sorted: $TestString");

    $TestString  = "<s1 x='e\'ee' b='rrr' y='e\'ee'>";
    $Expected    = "<s1 e'ee' ee' b=' rrr' y=' x='e'>";
    ok( sortXMLAttributes($TestString) eq $Expected, "invalid XML: $TestString");

    # QuoteXML

    ok( !defined QuoteXML(undef), "QuoteXML undef" );
    ok( defined QuoteXML(''), "QuoteXML ''" );
    ok( QuoteXML '<>&"\'' eq '&lt;&gt;&amp;&quot;&apos;', "QuoteXML '<>&\"\\''" );

    # UnquoteXML

    ok( !defined UnquoteXML(undef), "UnquoteXML undef" );
    ok( defined UnquoteXML(''), "UnquoteXML ''" );
    ok( UnquoteXML '&lt;&gt;&amp;&quot;&apos;' eq '<>&"\'', "UnquoteXML '<>&\"\\''" );
    ok( UnquoteXML '&amp;lt;' eq '&lt;', "UnquoteXML '&amp;lt;'" );

    # SplitEMailAddresses

    my %AddrListTests = (
        "acb;\"\\\\,\\\(\\\"\"cccc, \"rrr\"; a,  ; bd\\\\, abf\\\, (\",\\\\x); aaaddd <x\\\<xx\@h<hh,hhhh><hhhh>>,<sss>" =>
        [ "acb","\"\\\\,\\\(\\\"\"cccc","\"rrr\"","a","bd\\\\","abf\\","(\",\\\\x)","aaaddd <x\\\<xx\@h<hh,hhhh><hhhh>>","<sss>" ],
        '"Manuel Blechschmidt" <mblechschmidt@epages.de>;' =>
        [ '"Manuel Blechschmidt" <mblechschmidt@epages.de>' ],
        'mblechschmidt@epages.de;m.mustermann@epages.de' =>
        [ 'mblechschmidt@epages.de','m.mustermann@epages.de' ],
        '"Manuel Blechschmidt" <mblechschmidt@epages.de>; "Max; Mustermann" <m.mustermann@epages.de>' =>
        [ '"Manuel Blechschmidt" <mblechschmidt@epages.de>','"Max; Mustermann" <m.mustermann@epages.de>' ],
        '"O\'Neil\'s shop" <oneil@epages.de>; O\'Neil\'s shop <oneil@epages.de>' =>
        [ '"O\'Neil\'s shop" <oneil@epages.de>','O\'Neil\'s shop <oneil@epages.de>' ],
        '"Manuel Blechschmidt" <mblechschmidt@epages.de>; "John; \\"Mller" <j.doe@epages.de>' =>
        [ '"Manuel Blechschmidt" <mblechschmidt@epages.de>','"John; \\"Mller" <j.doe@epages.de>' ],
        '"<oneil@epages.de> O Neil\'s shop"; O Neil\'s shop <oneil@epages.de>' =>
        [ '"<oneil@epages.de> O Neil\'s shop"','O Neil\'s shop <oneil@epages.de>' ],
        ' oneil@epages.de;O N"eils" shop <mhoehn@epages.de>' =>
        [ 'oneil@epages.de','O N"eils" shop <mhoehn@epages.de>' ],
        );

    ok( $AddrListTests{$_} ~~ SplitEMailAddresses($_),  $_ ) foreach sort keys %AddrListTests;

    ok( Slice('This is a string to long to show in a navigation element', 15) eq 'This is a st...', "slice string with positive length and default postfix");
    ok( Slice('This is a string to long to show in a navigation element', 15, '***') eq 'This is a st***', "slice string with positive length and '***' as postfix");
    ok( Slice('This is a string to long to show in a navigation element') eq 'This is a string to long to...', "slice string with default length and default postfix");
    ok( Slice('This is a string to long to show in a navigation element', -15) eq '...tion element', "slice string with negative length and default postfix");

    ok( Substr('This is a string to long to show in a navigation element', 0, 15) eq 'This is a strin', "substring from first position to length");
    ok( Substr('This is a string to long to show in a navigation element', 0, -15) eq 'This is a string to long to show in a nav', "substring from first position to (length og string - 15)");
    ok( Substr('This is a string to long to show in a navigation element', 15) eq 'g to long to show in a navigation element', "substring with startposition 15");
    ok( Substr('This is a string to long to show in a navigation element', -15) eq 'igation element', "substring with startposition (length(String) -15)");
    ok( Substr('This is a string to long to show in a navigation element', 15, 15) eq 'g to long to sh', "substring with startposition (length(String) -15)");

    my @TestList = (
        {
            'Name'  => 'Wirkstoff: Aloe Vera, Swisa Beauty',
            'Title' => 'Wirkstoff: Aloe Vera, Swisa Beauty',
            'Ratio' => 1.0,
        },
        {
            'Name'  => 'Wirkstoff: Aloe Vera, Swisa Beauty',
            'Title' => 'Swisa Beauty - Aloe Vera - hautfreundliche Produkte',
            'Ratio' => 0.67,
        },
        {
            'Name'  => 'Wirkstoff: Aloe Vera, Swisa Beauty',
            'Title' => 'wirkstoffe: Aloe2 verAA, blah Swisas Beautye',
            'Ratio' => 0.83,
        },
        {
           'Name'  => 'Wirkstoff: Aloe Vera, Swisa Beauty',
           'Title' => 'Swisa Beauty - Aloe Vera - toll super kostenlos bestellen!',
           'Ratio' => 0.5,
        },
        {
            'Name'  => 'Wirkstoff: Aloe Vera, Swisa Beauty',
            'Title' => 'Sqisa Beetly - Quark Quatsch - toll super kostenlos bestellen!',
            'Ratio' => 0.12,
        },
        {
            'Name'  => 'yes',
            'Title' => 'not',
            'Ratio' => 0,
        },
        {
            'Name'  => 'Not empty',
            'Title' => '',
            'Ratio' => 0,
        },
        {
            'Name'  => '',
            'Title' => '',
            'Ratio' => 1,
        },
    );
    foreach my $Item (@TestList) {
        my $Ratio = CompareStrings($Item->{'Title'}, $Item->{'Name'});
        my $RoundedRatio = Math::BigFloat->new($Ratio)->ffround(-2);
        cmp_ok( $RoundedRatio, '==', $Item->{'Ratio'}, 'CompareStrings: check if strings are similar. Returned ratio is ' . $RoundedRatio . '.' );
    }

    # coverage parameter
    my $Ratio = CompareStrings(
        'Wirkstoff: Aloe Vera, Swisa Beauty',
        'Sqisa Beetly - Quark Quatsch - toll super kostenlos bestellen!',
        0.5
    );
    my $RoundedRatio = Math::BigFloat->new($Ratio)->ffround(-2);
    cmp_ok($RoundedRatio, '==', 0.4, 'Run CompareStrings with coverage parameter. Returned ratio is ' . $RoundedRatio . '.');

    is (CompareStrings('aaa', 'aaa', 1), 1, "Compare identical strings");

    is(sortCSVLine('a,c,b'),'a,b,c','sortCSVLine');
}

RunScript(Sub=>\&Main);
