#========================================================================================
# package      DE_EPAGES::Database::API::Connection
# state        public
# description  This package provides an interface to make, get and close connections.
#========================================================================================
package DE_EPAGES::Database::API::Connection;
use base qw( Exporter );

use strict;
our @EXPORT_OK = qw(
    GetCurrentDBHandle
    ConnectToSite
    ConnectSite
    ConnectTestSite
    GetTestSite
    DisconnectSite
    Connect
    RunOnStore
    RunOnSystem
    Transaction
    EncryptDBPassword
    StoreConnectionDefaults
    CreateStoreConf
    GetCartridgeName
    GetCartridgeStore
    GetStoreConfDir
);

use Readonly;
use Config::IniFiles;
use if $] < 5.038, 'DES::EDE3CBC'; ##no critic qw(ProhibitMagicNumbers)
use if $] >= 5.038, 'Crypt::CBC'; ##no critic qw(ProhibitMagicNumbers)

use DE_EPAGES::Core::API::Constants         qw( PERL538 );
use DE_EPAGES::Core::API::Error             qw (Error ExistsError GetError);
use DE_EPAGES::Core::API::File              qw (ExistsFile CorrectPathSlashes);
use DE_EPAGES::Core::API::Log               qw (GetLog);
use DE_EPAGES::Core::API::PerlTools         qw (Str2Hex UsePackage isValid trim);
use DE_EPAGES::Core::API::TestParameter     qw (TestString TestHash);
use DE_EPAGES::Core::API::Warning           qw (Warning);
use DE_EPAGES::DataCache::API::SessionCache qw (UpdateAll);
use DE_EPAGES::Database::API::Config        qw (DatabaseConfig
                                                DatabaseConfigFileName
                                                DatabaseConnectionData
                                                DatabaseBackupData
                                                GetRootConnections);

our $CURRENT_DBI;
our %CONNECTIONS = ();

Readonly::Scalar my $AES_KEYLENGTH => 16;
sub ErrorDefinition {
    return {
        'DatabaseNotActive' => 'Database connection #Store is not active',
        'DatabaseDefined' => 'Database defined in datasource conflicts with Database parameter!',
        'FILE_OPEN_FAILED' => 'couldn\'t open file (#FileName) because #Code.',
        'FILE_WRITE_FAILED' => 'couldn\'t write file (#FileName) because #Code.',
        'NO_ROOT_CONNECTION' => 'no connection exists with login=root in Database.d/#Conf.conf for creating #Store',
        'USERNAME_TOO_LONG' => 'Username must not exceed 15 chars!',
    }
}

#========================================================================================
# function     GetCurrentDBHandle
# state        public
#----------------------------------------------------------------------------------------
# syntax       $dbh = GetCurrentDBHandle();
#----------------------------------------------------------------------------------------
# description  Returns the current database handle.
#----------------------------------------------------------------------------------------
# return       $dbh | database handle | object
#========================================================================================
sub GetCurrentDBHandle () {
    return $CURRENT_DBI;
}

#========================================================================================
# function     SetCurrentDBHandle
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetCurrentDBHandle($dbh);
#----------------------------------------------------------------------------------------
# description  Sets the current database handle.
#----------------------------------------------------------------------------------------
# input        $dbh | database handle | object
#========================================================================================
sub SetCurrentDBHandle ($) { ## no critic qw(RequireCommentForExportedFunctions)
    $CURRENT_DBI = shift;
    UpdateAll();
}

#========================================================================================
# function     Connect
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $dbh = Connect($DataSource, $Login, $Password, $CacheKey)
# example      my $dbh = Connect('dbi:mysql:admindb', 'admin', 'geheim', 'admindb')
#----------------------------------------------------------------------------------------
# description  Connects to any database via the connection string (defined at DBI).
#               Username and password are required. The correct DBI interface will be
#               selected to make connection.
#               Use <function SetCurrentDBHandle> to make the database handle accessible
#               from DAL functions.
#----------------------------------------------------------------------------------------
# input        $DataSource | DBI connection string, see "perldoc DBI" for details | string
# input        $Login | database user name | string
# input        $Password | database password (optional) | string
# input        $CacheKey | key to identify the connection | string
# return       $dbh | database handle | object
#========================================================================================
sub Connect ($$$$) {
    my ($DataSource, $Login, $Password, $CacheKey) = @_;

    my ($driver) = $DataSource =~ /^dbi:(\w+)/;
    my $DBI = __PACKAGE__;
    $DBI =~ s/Connection$/DBI$driver/;
    UsePackage($DBI);
    return $DBI->new($DataSource, $Login, $Password, $CacheKey);
}

#========================================================================================
# function     ConnectToSite
# state        public
#----------------------------------------------------------------------------------------
# syntax       ConnectToSite();
# syntax       ConnectToSite($Password);
#----------------------------------------------------------------------------------------
# description  Connects to site database (section [Site.Default] at Database.d/Site.conf).
#               The password is not required if it is set in the config file.
#----------------------------------------------------------------------------------------
# input        $Password | database password (optional) | string
#========================================================================================
sub ConnectToSite (;$) {
    my @Params = @_;
    ConnectSite('Site', @Params);
}

#========================================================================================
# function     ConnectTestSite
# state        public
#----------------------------------------------------------------------------------------
# syntax       ConnectTestSite();
#----------------------------------------------------------------------------------------
# description  Connects to the test database (see <function GetTestSite>).
#               The password must be set in Database.d/$TestSite.conf.
#========================================================================================
sub ConnectTestSite (;$) {

    ConnectSite( GetTestSite() );
}

#========================================================================================
# function     GetTestSite
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $StoreName = GetTestSite();
# example      RunOnStore( 'Store'=> GetTestSite(), 'Sub' => sub {
#                   print LoadRootObject()->get('StoreName');
#               });
#----------------------------------------------------------------------------------------
# description  Returns the connection name of the test database.
#               The name can be specified by the EPAGES_TESTSITE environment variable.
#               The default value is calculated. Therefor the database.d folder
#               is taken into account. Every conf inside is checked to determine on which
#               store the cartridge is installed.
#               If this fails the default will be Site.
#----------------------------------------------------------------------------------------
# return       $StoreName | database connection name | string
#========================================================================================
sub GetTestSite {
    eval {
        ## no critic (RequireLocalizedPunctuationVars)
        $ENV{'EPAGES_TESTSITE'} = GetCartridgeStore(GetCartridgeName($0),GetStoreConfDir())
            unless defined $ENV{'EPAGES_TESTSITE'};
    };
    if (ExistsError()) {
        # do nothing
    }

    return $ENV{'EPAGES_TESTSITE'} || 'Site';
}

#========================================================================================
# function     GetStoreConfDir
# state        public
#----------------------------------------------------------------------------------------
# syntax       my storeConfDir = GetStoreConfDir();
#----------------------------------------------------------------------------------------
# description  Returns the Conf path of the stores conf.
#----------------------------------------------------------------------------------------
# return       storeConfPath | location of store configs | string
#========================================================================================
sub GetStoreConfDir {
    return CorrectPathSlashes( join ('/',($ENV{'EPAGES_CONFIG'}, 'Cartridges.d')));
}

#========================================================================================
# function     GetCartridgeName
# state        public
#----------------------------------------------------------------------------------------
# syntax       my CartridgeName = GetCartridgeName($fileName);
# example      my CartridgeName = GetCartridgeName('Database.t')
#----------------------------------------------------------------------------------------
# description  Returns the name of the cartridge in which the caller function of GetTestSite
#               method is placed.
#----------------------------------------------------------------------------------------
# input        $fileName       | file name of the caller file | string
# return       $cartrigeName   | name of the cartridge where caller function is located | string
#========================================================================================
sub GetCartridgeName {
    my ($fileName) = @_;

    # try to get the cartridge name
    my $absoluteCallerPath  = File::Spec->rel2abs($fileName);

    my (undef, $directories,undef) = File::Spec->splitpath( $absoluteCallerPath );
    my @Components = File::Spec->splitdir( $directories );

    my $count = 0;

    # search for cartridges and take the folder after the next folder
    foreach my $component (@Components) {
        if ($component eq 'Cartridges') {
            $count = $count + 2;
            last;
        }
        $count++;
    }

    return $Components[$count];
}

#========================================================================================
# function     GetCartridgeStore
# state        public
#----------------------------------------------------------------------------------------
# syntax       my CartridgeName = GetCartridgeStore($cartridgeName,$storeConfDir);
# example      my CartridgeName = GetCartridgeStore('Database','/srv/epages/eproot/Shared/Config/Cartridges.d')
#----------------------------------------------------------------------------------------
# description  Returns the name of the first Store (in lexicographical order)
#               where the cartridge is installed.
#----------------------------------------------------------------------------------------
# input        $cartridgeName  | name of the cartridge | string
# input        storeConfDir    | location of store configs| string
# return       $storeName      | name of the store where cartridge is installed | string
#========================================================================================
sub GetCartridgeStore {
    my ($cartridgeName,$storeConfDir) = @_;

    return unless isValid($cartridgeName);
    return unless isValid($storeConfDir);

    # No windows - feel free to improve.
    my $storeName = `grep -lr $cartridgeName= $storeConfDir 2>/dev/null | xargs -L 1 basename 2>/dev/null | sed 's/.conf//g' | sort | head -n 1`;

    my @storeName = trim(($storeName));

    return $storeName[0];
}

#========================================================================================
# function     EncryptDBPassword
#----------------------------------------------------------------------------------------
# syntax       $EncPassword = EncryptDBPassword($Password);
#----------------------------------------------------------------------------------------
# description  Crypt the plain password to save it at Database.d/$Store.conf. Enter the
#               crypted password at the Database.d/$Store.conf so that passwords are hidden
#               for users which have file system access.
#----------------------------------------------------------------------------------------
# input        $Password | database password (ini keyword is passwd) | string
# return       $EncPassword | crypted database password (ini keyword is encpasswd) | string
#========================================================================================
sub EncryptDBPassword ($) {
    my ($Password) = @_;
    my $EncPassword;

    my @DES_KEYS = ('Go09rOia', 'u377ureu', 'i1.6nanm', 'd.1Genb!');
    if ($] < PERL538) {
        $EncPassword = des_3ede_cbc_encrypt( $Password, @DES_KEYS, 1 );
    } else {
        my $DES_KEY = join('', @DES_KEYS[0..2]);
        my $cipher = Crypt::CBC->new( -cipher => 'DES_EDE3', -key => $DES_KEY,
                                      -iv => $DES_KEYS[3], -header => 'none',
                                      -literal_key => 1, 'padding' => 'none' );
        $EncPassword = $cipher->encrypt($Password);
    }

    return Str2Hex($EncPassword);
}

#========================================================================================
# function     ConnectSite
#----------------------------------------------------------------------------------------
# syntax       ConnectSite($Store);
# syntax       ConnectSite($Store, $Password);
# example      ConnectSite('Site');
#----------------------------------------------------------------------------------------
# description  Connects to any database with config file Database.d/$Store.conf.
#               The password is not required if it is defined in the config file.
#               Throws an error if the connection is not active.
#               This function sets the current database handle to make the database handle
#               accessable from DAL functions.
#----------------------------------------------------------------------------------------
# input        $Store | store name with database ini file $Store.conf | string
# input        $Password | database password (optional, default read from $Store.conf) | string
#========================================================================================
sub ConnectSite($;$) {
    my ($Store, $Password) = @_;
    TestString('Store', $Store);

    my $hSection = DatabaseConnectionData( $Store );
    Error('DatabaseNotActive' => { Store => $Store }) unless $hSection->{active};

    my $dbi;
    my $CacheConnection = $hSection->{'cache'};
    my $DataSource = $hSection->{'datasource'};
    GetLog->debug("Using datasource $DataSource");

    $Password = $Password ? $Password : $hSection->{'passwd'};
    $CacheConnection = 1 unless defined $CacheConnection; # cache connections by default (EPG-3656)
    if( $CacheConnection ) {
        my $key = (join '|', $DataSource , defined $hSection->{'login'} ? $hSection->{'login'} : '',$Password, $Store);
        my $ConnectionInfo = $CONNECTIONS{$key};
        if (defined $ConnectionInfo) {
            # is connection in use then use a new connection
            if ($ConnectionInfo->{'IsInUse'}) {
                $dbi = Connect($DataSource , $hSection->{'login'}, $Password, $Store);
                $dbi->{'OPENCONNECTIONKEY'} = undef;
            } else {
                $dbi = $ConnectionInfo->{'DBI'};
                my $pingOK = eval { $dbi->ping };
                Warning( GetError() ) if ExistsError();
                if ($pingOK) {
                    $ConnectionInfo->{'IsInUse'} = 1;
                } else {
                    $dbi->disconnect;
                    $dbi = undef;
                    delete $CONNECTIONS{$key};
                }
            }
        }
        if (not defined $dbi) {
            $dbi = Connect($DataSource , $hSection->{'login'}, $Password, $Store);
            $dbi->{'OPENCONNECTIONKEY'} = $key;
            $CONNECTIONS{$key} = {
                'IsInUse' => 1,
                'DBI'     => $dbi,
            };
        }
    }
    else {
        $dbi = Connect($DataSource, $hSection->{'login'}, $Password, $Store);
    }

    # store the AES key for data encryption in the dbi data (EPG-4374)
    if( defined $hSection->{'key'} ) {
        $dbi->{'key'} = $hSection->{'key'};
        Warning("AES key should have at least $AES_KEYLENGTH bytes!")
            if length $dbi->{'key'} > 0 && length $dbi->{'key'} < $AES_KEYLENGTH;
    }

    SetCurrentDBHandle($dbi);

    return;
}

#========================================================================================
# function     DisconnectSite
# state        public
#----------------------------------------------------------------------------------------
# syntax       DisconnectSite();
#----------------------------------------------------------------------------------------
# description  Disconnect current site connection and reset global current database handle
#               site alias.
#========================================================================================
sub DisconnectSite() {
    my $dbi = GetCurrentDBHandle();

    return unless defined $dbi;

    my $key = $dbi->{'OPENCONNECTIONKEY'};
    if( defined $key ) {
        if( $dbi->{'MANUAL_PING_LOSS'} ) {
            delete $CONNECTIONS{$key};
        }
        else {
            $CONNECTIONS{$key}{'IsInUse'} = 0;
        }
    }
    else {
        eval { $dbi->disconnect; };
        Warning( GetError() ) if ExistsError();
    }

    SetCurrentDBHandle(undef);
}

#========================================================================================
# function     RunOnStore
# state        public
#----------------------------------------------------------------------------------------
# syntax       RunOnStore( %Options );
# example      RunOnStore(
#                   'Store'=> $Store,
#                   'DBPassword'=> $DBPassword,
#                   'Sub' => $Sub,
#               );
#----------------------------------------------------------------------------------------
# description  Runs the main function in the context of the Error module and enables
#               detailed error information.
#----------------------------------------------------------------------------------------
# input        %Options | parameter
#               <ul>
#                   <li>Sub - a code reference of the main function - code ref
#                   <li>Store - store name - string
#                   <li>DBPassword - store password (default read from Database.d/$Store.conf) - string
#               </ul>
#               | hash
#========================================================================================
sub RunOnStore {
    my( %Options ) = @_;
    TestString('Store', $Options{'Store'});

    my $dbi = GetCurrentDBHandle();
    SetCurrentDBHandle(undef); # makes sure that disconnect is not on current handle
    eval {
        ConnectSite($Options{'Store'}, $Options{'DBPassword'});
        GetLog->debug("Switched to store $Options{'Store'}");
        # run the main function
        $Options{'Sub'}->();
    };
    my $Error = ExistsError() ?  GetError() : undef;
    DisconnectSite();
    SetCurrentDBHandle($dbi);
    if (defined $Error) {
        $Error->addVars({'Store' => $Options{'Store'}, 'Connection' => $Options{'Connection'}});
        Error($Error);
    }
    return;
}

#========================================================================================
# function     RunOnSystem
# state        public
#----------------------------------------------------------------------------------------
# syntax       RunOnSystem($cSub);
# example      RunOnSystem(sub { ... });
#----------------------------------------------------------------------------------------
# description  Runs the given sub on the system db. DB password is read from the conf.
#----------------------------------------------------------------------------------------
# input        $cSub | sub reference (closure) | ref.code
#========================================================================================
sub RunOnSystem {
    my ($cSub) = @_;
    RunOnStore('Store' => 'System', 'Sub' => $cSub);
}

#========================================================================================
# function     Transaction
# state        public
#----------------------------------------------------------------------------------------
# syntax       Transaction( $cCode, $cRollback);
# example      Transaction( sub {
#                   ProcessSql( ... );
#                   ProcessSql( ... );
#               } );
#----------------------------------------------------------------------------------------
# description  Runs a code block in a transaction using the current database handle. In
#               this transaction is no cache reset involved.
#----------------------------------------------------------------------------------------
# input        $cCode | code block | code reference
# input        $cRollback | roll back code block (optional) | code reference
#========================================================================================
sub Transaction {
    my ($cCode, $cRollback) = @_;
    GetCurrentDBHandle()->transaction( $cCode, $cRollback );
    return;
}

#========================================================================================
# function     StoreConnectionDefaults
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hDefault = StoreConnectionDefaults($hSettings, $hFlags);
#----------------------------------------------------------------------------------------
# description  returns a hash of default values for section [Connection] in
#               Database.d/Store.conf depending on $hSettings and
#               (for $hDefault->{datasource}) envvar EPAGES_MYSQL_CONNECTION.
#----------------------------------------------------------------------------------------
# input        $hSettings | settings
#                   <ul>
#                       <li>Store    - business unit name (required)
#                       <li>Database - store name
#                       <li>Backup   - connection for creating the database
#                   </ul>
#               | ref.hash
# input        $hFlags | flags
#                   <ul>
#                       <li>connect  - if set, dies unless can connect to $hSettings->{Backup}
#                   </ul>
#               | ref.hash
# return       $hDefault | settings (defaults in parenthesis)
#                   <ul>
#                       <li>Database   - store name (\L$hSettings->{Store}\Q)
#                       <li>active     - is database active? (1)
#                       <li>login      - user alias (usr_\L$hSettings->{Store}\Q)
#                       <li>passwd     - user password (epages)
#                       <li>Backup     - connection for creating the database (Backup)
#                       <li>datasource - datasource (derived from Backup.conf)
#                   </ul>
#               | ref.hash
#========================================================================================
sub StoreConnectionDefaults (;$$) {
    my ($hSettings, $hFlags) = @_;

    TestHash('hSettings', $hSettings, {'Store' => 'String;NotNull'});

    my %hDefault;

    my $MySQLDB = lc($hSettings->{Store});
    $hDefault{Database} = $MySQLDB;
    $hDefault{Database} .= 'db' unless $MySQLDB =~ /db$/;
    $hDefault{active} = 1;
    $hDefault{login} = 'usr_'. ($hSettings->{Database} // $hDefault{Database});
    $hDefault{passwd} = 'epages';

    # Backup

    my $RootConnections = GetRootConnections();
    $RootConnections = [$hSettings->{Backup}] if $hSettings->{Backup};

    foreach my $Store (@$RootConnections) {
        $hDefault{Backup} = $Store unless $hFlags->{connect};
        last if $hDefault{Backup};
        eval { ConnectSite($Store); };
        $hDefault{Backup} = $Store unless ExistsError();
        DisconnectSite();
    }
    my $Conf = scalar @$RootConnections == 1 ? $RootConnections->[0] : '*';
    Error('NO_ROOT_CONNECTION', { 'Store' => $hSettings->{Store}, 'Conf' => $Conf})
        unless $hDefault{Backup};

    # datasource

    $hDefault{datasource} = 'dbi:mysql';
    my @Parameters;
    if ($ENV{EPAGES_MYSQL_CONNECTION} and $ENV{EPAGES_MYSQL_CONNECTION} =~ /\S/) {
        if ($ENV{EPAGES_MYSQL_CONNECTION} =~ /\//) {
            my $Socket = $ENV{EPAGES_MYSQL_CONNECTION};
            $Socket =~ s,^.+=,,;
            push @Parameters, "mysql_socket=$Socket";
        } else {
            my ($Host, $Port) = split(/;/,$ENV{EPAGES_MYSQL_CONNECTION});
            if ($Host) {
                $Host =~ s,^.+=,,;
                push @Parameters, "host=$Host";
            }
            if ($Port) {
                $Port =~ s,^.+=,,;
                push @Parameters, "port=$Port";
            }
        }
    } else {
        my $DataSource = DatabaseBackupData(undef,
                                            $hDefault{Backup})->{datasource};
        if ($DataSource =~ s/^dbi:mysql://) {
            foreach (split(/;/,$DataSource)) {
                # Parameters must not contain the database name
                next if /^database=/;
                push @Parameters, $_;
            }
        }
    }
    $hDefault{datasource} .= ':'. join(';', @Parameters) if scalar @Parameters;

    return \%hDefault;
}

#========================================================================================
# function     CreateStoreConf
# state        public
#----------------------------------------------------------------------------------------
# syntax       CreateStoreConf($hSettings, $hFlags);
#----------------------------------------------------------------------------------------
# description  creates (unless exists) Shared/Config/Database.d/$hSettings->{Store}.conf
#----------------------------------------------------------------------------------------
# input        $hSettings | settings (defaults in parenthesis)
#                   <ul>
#                       <li>Store      - business unit name (required)
#                       <li>Database   - store name (\L$hSettings->{Store}\Q)
#                       <li>active     - is database active? (1)
#                       <li>login      - user alias (usr_\L$hSettings->{Store}\Q)
#                       <li>passwd     - user password (epages)
#                       <li>encpasswd  - encrypted user password ()
#                       <li>datasource - datasource (derived from Backup.conf)
#                       <li>Backup     - connection for creating the database (Backup)
#                       <li>BackupHost - backup database on BackupHost ()
#                       <li>DumpDir    - dumps can be found in DumpDir ()
#                       <li>StoreType  - store type ()
#                       <li>Key        - store key ()
#                   </ul>
#               | ref.hash
# input        $hFlags | flags
#                   <ul>
#                       <li>connect  - if set, dies unless can connect to $hSettings->{Backup}
#                   </ul>
#               | ref.hash
#========================================================================================
sub CreateStoreConf (;$$) {
    my ($hSettings, $hFlags) = @_;

    TestHash('hSettings', $hSettings, {'Store' => 'String;NotNull'});

    my $StoreConfName = DatabaseConfigFileName($hSettings->{Store});

    if (ExistsFile($StoreConfName)) {
        GetLog->info("Won't change existing file: $StoreConfName");
        return;
    }

    # get default values
    my $hConnection = StoreConnectionDefaults($hSettings, $hFlags);

    # overwrite default values with custom values
    # (skip Backup, it was already checked by StoreConnectionDefaults)
    foreach (qw(Database StoreType Key BackupHost DumpDir
                active login passwd encpasswd datasource)) {
        $hConnection->{$_} = $hSettings->{$_} if $hSettings->{$_};
    }

    # strip user name
    Readonly::Scalar my $MaxUserNameLength => 14;
    if (length($hConnection->{login}) >= $MaxUserNameLength) {
        $hConnection->{login} = substr($hConnection->{login},
                                       0, $MaxUserNameLength);
        Warning('USERNAME_TOO_LONG');
    }

    # prefer encypted password
    delete $hConnection->{passwd} if $hConnection->{encpasswd};

    # write Store.conf
    my $IniConfig = DatabaseConfig($hSettings->{Store});
    $IniConfig->section('Connection', $hConnection);
    $IniConfig->rewrite();
}

1;
