use strict;
use Readonly;
use Test::More;

use DES::EDE3CBC;
use Crypt::CBC;
use DE_EPAGES::Core::API::Constants qw( PERL538 );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw(EncryptDBPassword);
use DE_EPAGES::Database::API::Config;
use DE_EPAGES::Presentation::API::Session;

use constant DES_INIT_VECTOR => 'x.4ZeBj!';

sub Main {
    my @keys = qw(476F3039724F6961 7533373775726575 69312E366E616E6D 642E3147656E6221);
    my @plainKeys = map { hex2string($_) } @keys;
    my $plain = 'epages';
    my ($Expected, $Found, $plainHex, $encryptedHex, $encrypted, $decrypted);

    $Expected = 'Go09rOiau377ureui1.6nanmd.1Genb!';
    is(join('', @plainKeys), $Expected, 'hex2string(keys)');

    $Expected = '657061676573';
    $plainHex = string2hex($plain);
    is($plainHex, $Expected, 'plainHex = string2hex(plain)');

    $Expected = '20DB727D4F0A697B';
    $encryptedHex = encrypt($plain, @keys);
    is($encryptedHex, $Expected, 'encryptedHex = encrypt(plain)');

    ## no critic qw(ProhibitNoStrict ProhibitStringyEval ProhibitProlongedStrictureOverride)
    no strict;
    eval "\@key = ('Go09rOia', 'u377ureu', 'i1.6nanm', 'd.1Genb!'); \$enc = des_3ede_cbc_encrypt('$plain', \@key, 1);";
    $encryptedHex = string2hex($enc);
    eval "\@key = ('Go09rOia', 'u377ureu', 'i1.6nanm', 'd.1Genb!'); \$dec = des_3ede_cbc_encrypt('$enc', \@key, 0);";
    $Found = (split /\0/, $dec)[0];
    use strict;
    ## use critic

    is($encryptedHex, $Expected, 'des_3ede_cbc_encrypt(1)');
    is($Found, $plain, 'des_3ede_cbc_encrypt(0)');

    if ($] >= PERL538) {
        my $cipher = Crypt::CBC->new( -cipher => 'DES_EDE3', -key => 'Go09rOiau377ureui1.6nanm', -iv => 'd.1Genb!',
                                      -header => 'none', -literal_key => 1, 'padding' => 'none' );
        $encrypted = $cipher->encrypt($plain);
        is(string2hex($encrypted), $Expected, 'Crypt::CBC->encrypt');
        $Found = (split /\0/, $cipher->decrypt($encrypted))[0];
        is($Found, $plain, 'Crypt::CBC->decrypt');
        $Found = (split /\0/, $cipher->decrypt_hex(string2hex($encrypted)))[0];
        is($Found, $plain, 'Crypt::CBC->decrypt_hex');
    }

    is(EncryptDBPassword($plain), $Expected, 'EncryptDBPassword(plain)');
    is(EncryptDBPassword($plain), $Expected, 'EncryptDBPassword(plain) [again]');

    $Expected = " r}O\ni{";
    $encrypted = hex2string($encryptedHex);
    is($encrypted, $Expected, 'encrypted = hex2string(encryptedHex)');

    $decrypted = decrypt($encryptedHex, @keys);
    is($decrypted, $plain, 'decrypted = decrypt(encrypted)');

    $Found = DE_EPAGES::Database::API::Config::dECRYPTdbpASSWORD($encryptedHex);
    is($Found, $plain, 'dECRYPTdbpASSWORD(encryptedHex)');
    $Found = DE_EPAGES::Database::API::Config::dECRYPTdbpASSWORD($encryptedHex);
    is($Found, $plain, 'dECRYPTdbpASSWORD(encryptedHex) [again]');

    my $SessionID = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ012345';
    my $GUID = '50946F1D-0C28-4E0B-D6FD-D91A30FAA7A0';
    $Expected = '9755821CFCD7F55C278E17C4C552A673537E5CBE93A1FFA5619ECDDEF9334E4A';

    if ($] >= PERL538) {
        my $Key = (unpack('(A8)*', $GUID))[0];
        my $cipher = Crypt::CBC->new( -cipher => 'DES_EDE3', -key => $Key x 3, -iv => DES_INIT_VECTOR,
                                      -header => 'none', -literal_key => 1, 'padding' => 'none' );
        $encrypted = $cipher->encrypt($SessionID);
        is(string2hex($encrypted), $Expected, 'Crypt::CBC->encrypt(SessionID)');
        $Found = (split /\0/, $cipher->decrypt($encrypted))[0];
        is($Found, $SessionID, 'Crypt::CBC->decrypt(SessionID)');
        $Found = (split /\0/, $cipher->decrypt_hex(string2hex($encrypted)))[0];
        is($Found, $SessionID, 'Crypt::CBC->decrypt_hex(SessionID)');
    }

    ## no critic qw(ProtectPrivateSubs)
    $encrypted = DE_EPAGES::Presentation::API::Session::_CryptDESString($SessionID, $GUID);
    is(string2hex($encrypted), $Expected, '_CryptDESString(SessionID, GUID)');
    $encrypted = DE_EPAGES::Presentation::API::Session::_CryptDESString($SessionID, $GUID);
    is(string2hex($encrypted), $Expected, '_CryptDESString(SessionID, GUID) [again]');
    $decrypted = DE_EPAGES::Presentation::API::Session::_DeCryptDESString($encrypted, $GUID);
    is($decrypted, $SessionID, '_DeCryptDESString(encrypted, GUID)');
    $decrypted = DE_EPAGES::Presentation::API::Session::_DeCryptDESString($encrypted, $GUID);
    is($decrypted, $SessionID, '_DeCryptDESString(encrypted, GUID) [again]');
    ## use critic qw(ProtectPrivateSubs)

    done_testing();
}

RunScript( Sub => \&Main );

#############################################################################
#
#   verschlsseln einer Nachricht
#   Parameter:
#               $message    :   zu verschl. Nachricht
#               $KEY1,2,3   :   Schlssel als Folge von Hexadezimalzahlen,
#                               z.B. $KEY1='7968AB76EC51DC89';
#               $IVEC       :   Initialisierungsvektor als Folge von Hexdezimalzahlen
#
#   Ergebnis:
#               codierte Nachricht als Folge von Hexdezimalzahlen
#############################################################################
sub encrypt {
    my ($message,$KEY1,$KEY2,$KEY3,$IVEC) = @_;
    my ($coded_message, $k1, $k2, $k3, $ivec);

    $k1 = hex2string($KEY1);
    $k2 = hex2string($KEY2);
    $k3 = hex2string($KEY3);
    $ivec = hex2string($IVEC);

    $coded_message = des_3ede_cbc_encrypt($message,$k1,$k2,$k3,$ivec,1);
    return string2hex($coded_message);
}

#############################################################################
#
#   verschlsseln einer Nachricht
#   Parameter:
#               $coded_message  :   zu verschl. Nachricht als Folge von Hexadezimalzahlen
#               $KEY1,2,3       :   Schlssel als Folge von Hexadezimalzahlen,
#                                   z.B. $KEY1='7968AB76EC51DC89';
#               $IVEC           :   Initialisierungsvektor als Folge von Hexdezimalzahlen
#
#   Ergebnis:
#               codierte Nachricht als Folge von Hexdezimalzahlen
#############################################################################
sub decrypt {
    my ($coded_message,$KEY1,$KEY2,$KEY3,$IVEC) = @_;
    my ($message, $k1, $k2, $k3, $ivec);

    $k1=hex2string($KEY1);
    $k2=hex2string($KEY2);
    $k3=hex2string($KEY3);
    $ivec=hex2string($IVEC);

    $coded_message = hex2string($coded_message);
    $message = des_3ede_cbc_encrypt($coded_message,$k1,$k2,$k3,$ivec,0);
    $message = (split /\0/, $message)[0];
    return $message;
}


sub hex2string {
    my ($hex_string)=@_;
    my $string = '';
    my $i;

    for ($i = 0; $i<length($hex_string); $i+=2) {
        $string .= chr(hex(substr($hex_string,$i,2)));
    }
    return $string;
}

sub string2hex {
    my ($string)=@_;
    my $hex = '';
    my $i;

    for ($i = 0; $i<length($string); $i++) {
        $hex .= sprintf('%02X', ord(substr($string,$i,1)));
    }
    return $hex;
}

1;
