#========================================================================================
# package      DE_EPAGES::Dictionary::API::Localizer
# state        public
# example      use DE_EPAGES::Dictionary::API::Localizer qw ( Replace );
#               use DE_EPAGES::Dictionary::API::XML::DictionaryImportHandler qw ( ImportDictionary );
#               my $Dictionary = {};
#               ImportDictionary($_, $Dictionary) foreach @DictionaryFiles;
#               if (defined $Dictionary->{$LanguageCode}) {
#                   $Template = Replace($Template, $Dictionary->{$LanguageCode})
#               }
#----------------------------------------------------------------------------------------
# description  This package replaces dictionary entries in a content with real translations.
#========================================================================================
package DE_EPAGES::Dictionary::API::Localizer;
use base qw (Exporter);
use strict;
use English;

our @EXPORT_OK = qw(
    Replace
);

use DE_EPAGES::Core::API::Error qw( Error );
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::TLE::API::SpaceFormatter;
use DE_EPAGES::TLE::API::CaseFormatter;
use DE_EPAGES::TLE::API::HtmlFormatter;

sub ErrorDefinition {
    return {
        'NOT_REPLACED'      => 'tag #Name[#Format] was not replaced.',
        'FORMAT_UNKNOWN'    => 'format unknown #Format on #TLEValue.',
        'ENCODING_UNKNOWN'  => 'encoding unknown #Encoding on #TLEValue.',
    };
}

#========================================================================================
# function      Replace
# state         public
#----------------------------------------------------------------------------------------
# syntax        $Template = Replace( $MasterTemplate, $hStrings );
#----------------------------------------------------------------------------------------
# description   Exported Function to use this localizer easier. This function registers <ul>
#                   <li><package DE_EPAGES::TLE::API::SpaceFormatter>
#                   <li><package DE_EPAGES::TLE::API::CaseFormatter>
#                   <li><package DE_EPAGES::TLE::API::HtmlFormatter>
#                </ul> to the localizer and starts the localisation process with <function localize>.
#----------------------------------------------------------------------------------------
# input         $MasterTemplate | a string e.g.: template-content | string
# input         $hStrings | keyword => value pairs | hash ref
# return        $Template | the processed template content | string
#========================================================================================
sub Replace {
    my ($MasterTemplate, $hStrings) = @_;

    my $Localizer = __PACKAGE__->new( $hStrings );
    DE_EPAGES::TLE::API::SpaceFormatter->new->register($Localizer);
    DE_EPAGES::TLE::API::CaseFormatter->new->register($Localizer);
    DE_EPAGES::TLE::API::HtmlFormatter->new->register($Localizer);

    return $Localizer->localize( $MasterTemplate );
}

#========================================================================================
# function      new
# state         private
#----------------------------------------------------------------------------------------
# syntax        $Localizer = DE_EPAGES::Dictionary::API::Localizer->new();
#----------------------------------------------------------------------------------------
# description   Returns a new localizer object.
#----------------------------------------------------------------------------------------
# input         $hStrings | keyword => value pairs | hash ref
# return        $Localizer | localizer object | object
#========================================================================================
sub new {
    my $class = shift;
    my( $hStrings ) = @_;

    return bless {
        'EncodingHandler'       => [],
        'FormatHandler'         => [],

        'NamedEncodingHandler'       => {},
        'NamedFormatHandler'         => {},

        'Strings' => $hStrings,
    }, $class;
}

#========================================================================================
# function      localize
# state         private
#----------------------------------------------------------------------------------------
# syntax        $Template = $Localizer->localize( $MasterTemplate );
#----------------------------------------------------------------------------------------
# description   Replaces barwords in the given string with the corresponding value inside of
#                an global hash (%locales) depending on the given language. It will replace
#                only barwords encapsulate in '{' and '}'. The barword should only
#                contain (a-zA-Z_0-9).
#                The length and alignment may be specified as follows:
#                {TXT_quantity[-12]} right aligned
#                {TXT_quantity[12]}  left aligned
#----------------------------------------------------------------------------------------
# input         $MasterTemplate | a string e.g.: template-content | string
# return        $Template | the processed template content | string
#========================================================================================
sub localize {
    my $self = shift;
    my ($MasterTemplate) = @_;

    return '' unless defined $MasterTemplate;
    return '' if $MasterTemplate eq '';

    my @Parts = split /<script/, $MasterTemplate;
    my $BeforeScript = shift @Parts;

    $self->_localizePart( $BeforeScript );
    foreach my $Part (@Parts) {
        my ($Script, $AfterScript) = split /<\/script>/, $Part;

        my ($Tag, $Code);
        if ($Script =~ /^([^>]*?>)(.*)$/isx) {
            $Tag = $1;
            $Code = $2;
        }

        # Don't handle jquery template as javascript block
        if ($Tag =~ /type\s*=\s*(["']).*?jquery.*?\1/i) { #"
            $self->_localizePart( $Script );
        }
        # Javascript block
        else {
            $self->_localizePart( $Tag );
            $self->_localizePart( $Code, 'js' );

            $Script = $Tag.$Code;
        }

        $self->_localizePart( $AfterScript );
        $Part = $Script . '</script>' . $AfterScript;
    }

    return join '<script', $BeforeScript, @Parts;
}

#========================================================================================
# function     _localizePart
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Localizer->_localizePart( $TemplatePart, $DefaultFormat );
# example      $Localizer->_localizePart( $TemplatePart, 'js' );
#----------------------------------------------------------------------------------------
# description  Replaces localization tags in-place inside a template. Uses the
#               formatter $DefaultFormat for tags without format specification.
#========================================================================================
sub _localizePart { ## no critic qw( RequireArgUnpacking )
    my $self = shift;
    my $DefaultFormat = $_[1] // '';

    return unless defined $_[0];

    my $Keyword = '\w*[a-zA-Z]\w*'; # Keywords contain only word characters, but one of them must be a letter A-Z

    if ( $_[0] =~ /\{$Keyword\}/ ) {
        $_[0] =~ s/
            \{
                ($Keyword) # keyword -> $1
            \}
        /$self->_replace( $1, $DefaultFormat, $PREMATCH, $MATCH, $POSTMATCH )/egx;
    }
    if ($_[0] =~ /\{$Keyword\[[^\]]+\]\}/) {
        $_[0] =~ s/
            \{
                ($Keyword) # keyword -> $1
                (?:
                    \[
                        ([^\]]+) # format string -> $2
                    \]
                )
            \}
        /$self->_replace( $1, $2, $PREMATCH, $MATCH, $POSTMATCH )/egx;
    }
    return;
}

#========================================================================================
# function      _replace
# state         private
#----------------------------------------------------------------------------------------
# syntax        $Replace = $Localizer->_replace($Name, $Format, $PreMatch, $Match, $PostMatch);
#----------------------------------------------------------------------------------------
# description   Returns the localized  of hashkey $Name of the hash referenced by $hStrings.
#                If the key does not exist, the value is equal to the key encapsulated in
#                '{' and '}' and a message is printed.
#----------------------------------------------------------------------------------------
# input         $Name | key which should be localized | string
# input         $Format | (optional) format string | string
# input         $PreMatch | (optional) everything before keyword | string
# input         $Match | (optional) matched keyword | string
# input         $PostMatch | (optional) everything after keyword | string
# return        $Replace | replacement string | string
#========================================================================================
sub _replace {
    my $self = shift;
    my ($Name, $Format, $PreMatch, $Match, $PostMatch) = @_;

    $PreMatch //= '';
    $PostMatch //= '';
    return $Match if $PreMatch =~ /\$$/;                        # e.g. ${var}
    return $Match if $PreMatch =~ /\{$/ && $PostMatch =~ /^\}/; # e.g. {{each}}

    my $hStrings = $self->{'Strings'};
    if( not exists $hStrings->{$Name} ) {
        Warning('NOT_REPLACED', { 'Name' => $Name, 'Format'=>defined $Format ? $Format : '' });
        return $Match;
    }
    my $Value = $hStrings->{$Name};

    # SAME AS PART OF DE_EPAGES::TLE::API::Processor::tle
    # format value (like money, float, date, time)
    if (defined $Format and $Format =~ /\S/) {
        foreach my $FormatPart (split /,/, $Format) {
            my ($FormatString, @FormatParameter) = split /:/,$FormatPart;
            my $FormattedValue = undef;
            if (defined $self->{'NamedFormatHandler'}{$FormatString}) {
                $FormattedValue = $self->{'NamedFormatHandler'}{$FormatString}->Format($self, $FormatString, $Value, $Name, \@FormatParameter);
            } else {
                # get from unnamed format handler
                foreach my $Handler (@{$self->{'FormatHandler'}}) {
                    $FormattedValue = $Handler->Format($self, $FormatString, $Value, $Name, \@FormatParameter);
                    last if defined $FormattedValue;
                }
            }
            # get unknonwn format
            Error('FORMAT_UNKNOWN', {'TLEValue'=>$Name, 'Format'=>$FormatString} ) unless defined $FormattedValue;
            $Value = $FormattedValue;
        }
    }

    # encode formatted value (like html, xml, js)
    my $Encoding = $self->{'Context'}{'Encoding'};
    if (defined $Encoding and $Encoding =~ /\S/) {
        my $Handler = $self->{'NamedEncodingHandler'}{$Encoding};
        my $FormattedValue = undef;
        if (defined $Handler) {
            $FormattedValue = $Handler->Encode($self, $Encoding, $Value, $Name);
        } else {
            foreach my $Handler (@{$self->{'EncodingHandler'}}) {
                $FormattedValue = $Handler->Encode($self, $Encoding, $Value, $Name);
                last if defined $FormattedValue;
            }
        }
        Error('ENCODING_UNKNOWN', {'TLEValue'=>$Name, 'Encoding'=>$Encoding} ) unless defined $FormattedValue;
        $Value = $FormattedValue;
    }
    return $Value;
}

#========================================================================================
# function     registerHandler
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ReturnValue = $Localizer->registerHandler( $HandlerType, $Handler, $Name);
# example      $TLEProcessor->registerHandler('FormatHandler', DE_EPAGES::TLE::API::SpaceFormatter->new);
#----------------------------------------------------------------------------------------
# description  Formatter uses this function to add more functionality to this TLE localizer. You can
#               add different types:<ul>
#                  <li>'FormatHandler' - functions to format each TLE value ( [] in template)
#                  <li>'EncodingHandler' - functions to encode each TLE value (in template)
#               </ul>
#----------------------------------------------------------------------------------------
# input        $HandlerType | name of handler | string
# input        $Handler  | the handler object | ref.*
# input        $Name | name of tle or function | string
# return       $ReturnValue  | always true | boolean
#========================================================================================
sub registerHandler {
    my $self = shift;
    my ($HandlerType, $Handler, $Name) = @_;
    if (defined $Name) {
        $self->{'Named'.$HandlerType}{$Name} = $Handler;
    } else {
        push @{$self->{$HandlerType}}, $Handler;
    }
    return 1;
}

1;
