#========================================================================================
# package      DE_EPAGES::MolliePayments::API::MolliePaymentsUrlBuilder
# state        public
#----------------------------------------------------------------------------------------
# description  Build return URL for Backlink
#========================================================================================
package DE_EPAGES::MolliePayments::API::MolliePaymentsUrlBuilder;

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Core::API::Url qw ( ParseQueryString );
use URI;
use URI::URL;
use JSON;
use DE_EPAGES::Object::API::Factory qw(LoadRootObject);
use DE_EPAGES::MolliePayments::API::Constants qw ( APP_SCOPES CREATE_LINKS_API_V2_URL );
use DE_EPAGES::MolliePayments::API::Tools qw ( DoRequestCurl );
use strict;

use constant MOLLIE_V2_AUTHORIZE_CONNECT_URL => 'https://www.mollie.com/oauth2/authorize';

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( );
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::MolliePayments::API::MolliePaymentsUrlBuilder
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

#========================================================================================
# function     buildConnectAndReconnectUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       buildConnectAndReconnectUrl( $StateURL );
#----------------------------------------------------------------------------------------
# description  Build the connect URL used for on-boarding
#----------------------------------------------------------------------------------------
# input        $StateURL | Current URL in the MBO  | string
# return       $URL->as_string() | Mollie connect url | string
#========================================================================================
sub buildConnectAndReconnectUrl {
    my $self = shift;
    my ($PaymentMethod, $StateURL) = @_;

    my $hQueryParameters = $self->_buildQueryParameters($StateURL);

    my $Url = $self->_buildEncodedUrl($PaymentMethod, $hQueryParameters);

    return $Url;
}

#========================================================================================
# function     buildRedirectToMboUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       buildRedirectToMboUrl( $hQueryParameters );
#----------------------------------------------------------------------------------------
# description  Build redirect to MBO URL using returned data from MolliePayments
#----------------------------------------------------------------------------------------
# input        $hQueryParameters | Hashref which is returned from MolliePayments | hashref
# return       $URL->as_string() | BacklinkUrl with AuthorizationCode or Error | string
#========================================================================================
sub buildRedirectToMboUrl {
    my $self = shift;
    my ($hQueryParameters) = @_;
    my $aBacklinkUrl = $hQueryParameters->{'state'};

    if ( (!isValid($aBacklinkUrl)) or (!isValid($aBacklinkUrl->[0])) ) {
        Error('NoURLFound');
    }

    return $self->_urlBuilder($hQueryParameters);
}

#----------------------------------------------------------------------------------------
sub _buildEncodedUrl {
    my $self = shift;
    my ($PaymentMethod, $hQueryParameters) = @_;

    my $Res;
    if ( $PaymentMethod ) {
      my $Shop = $PaymentMethod->getSite();
      my $Address = $Shop->get('Address');
      my $System = LoadRootObject();
      my $MolliePayments_OrganisationToken_ClientLinks = $System->get('MolliePayments_OrganisationToken_ClientLinks');

      #-- If we have the necessary fields to use the API, then let's do it! ^.^
      if ( $Address && $MolliePayments_OrganisationToken_ClientLinks ) {
        my $Email = $Address->get('EMail');
        my $FirstName = $Address->get('FirstName');
        my $LastName = $Address->get('LastName');
        my $Company = $Address->get('Company') ? $Address->get('Company') : $Shop->get('Name', $Shop->get('LanguageID'));
        my $Country = $Address->get('Country') ? $Address->get('Country')->{'Code2'} : undef;
        if ( $Email && $FirstName && $LastName && $Company && $Country ) {
          my $Content = {
            'owner' => {
              'email' => $Email,
              'givenName' => $FirstName,
              'familyName' => $LastName,
            },
            'address' => {
              'streetAndNumber' => $Address->get('Street'),
              'postalCode' => $Address->get('Zipcode'),
              'city' => $Address->get('City'),
              'country' => $Country,
            },
            'name' => $Company,
          };

          $Res = DoRequestCurl('POST', CREATE_LINKS_API_V2_URL, encode_json($Content), {'Authorization' => "Bearer $MolliePayments_OrganisationToken_ClientLinks", 'Content-Type' => 'application/json'}, {'LogName' => 'create-links'});
        }
      }
    }

    #-- If we got a successful response from the create-links API, use it, otherwise, fallback to the original static link
    my $ConnectUrl = $Res && $Res->is_success ? decode_json($Res->decoded_content)->{'_links'}->{'clientLink'}->{'href'} : MOLLIE_V2_AUTHORIZE_CONNECT_URL;

    my $Url = URI::URL->new($ConnectUrl);

    $Url->query_form($hQueryParameters);
    my $hBacklinkUrlParameters = ParseQueryString($Url->equery(), 'utf-8');
    $Url->query_form( $hBacklinkUrlParameters );

    return $Url;
}
#----------------------------------------------------------------------------------------
sub _buildQueryParameters {
    my $self = shift;
    my ($StateURL) = @_;

    my $System = LoadRootObject();
    my $ClientID = $System->get('MolliePayments_AppClientID');
    my $RedirectUrl = $System->get('MolliePayments_AppRedirectURL');


    return {
        'response_type'   => 'code',
        'client_id'       => $ClientID,
        'scope'           => join(' ', APP_SCOPES),
        'redirect_uri'    => $RedirectUrl,
        'state'           => $StateURL,
        'approval_prompt' => 'force',
    };
}

#----------------------------------------------------------------------------------------
sub _urlBuilder {
    my $self = shift;
    my ($hQueryParameters) = @_;

    # Original Change action must be removed
    delete($hQueryParameters->{'ChangeAction'});

    # State contains the URL which is backlink to MBO. No need to send it with QueryString again
    my $BackLinkUrl = delete($hQueryParameters->{'state'});

    my $URL = URI->new($BackLinkUrl->[0]);
    my $hBacklinkUrlParameters = ParseQueryString($URL->query(), 'utf-8');
    my $hQueryParam = {%{$hBacklinkUrlParameters}, %{$hQueryParameters}};
    $URL->query_form($hQueryParam);

    return $URL->as_string();
}

1;
