#========================================================================================
# package      DE_EPAGES::MolliePayments::API::Tools
# state        public
#----------------------------------------------------------------------------------------
# description  Tooling for specific tasks like log filtering
#========================================================================================
package DE_EPAGES::MolliePayments::API::Tools;
use base qw( Exporter );
use DE_EPAGES::Core::API::PerlTools qw( isValid GenerateShortGUID );

use DE_EPAGES::Core::API::Object::NumberFormat;
use DE_EPAGES::MolliePayments::API::Constants qw ( SUPPORTED_LOCALES DEFAULT_LOCALE PAYMENT_TYPE_MOLLIE_PAYMENTS );
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError);
use DE_EPAGES::Core::API::Log qw ( LogError );
use DE_EPAGES::ExternalPayment::API::Log qw ( LogPayment );
use Net::Curl::Easy qw(:constants);
use HTTP::Response;

use strict;

our @EXPORT_OK = qw (
    DoRequestCurl
    GetSupportedLocale
    FormatFloatForMollie
    FilterTokensResponse
    FilterOnboardingReturnQuery
    ParseErrorStringToHash
);

#========================================================================================
# function     DoRequestCurl
# state        public
#----------------------------------------------------------------------------------------
# syntax       DoRequestCurl($Method, $Url, $Content, $hHeaders);
#----------------------------------------------------------------------------------------
# description  Perform a HTTP request with the given parameters, using libcurl
#----------------------------------------------------------------------------------------
# input        $Method | HTTP method e.g. GET, POST | string
# input        $Url | URL to connect to | string
# input        $Content | Request body content | string
# input        $hHeaders | Headers to add to the request | ref.hash
# input        $hAdditionalInfo | Additional information to log | ref.hash
#----------------------------------------------------------------------------------------
# return       $Res | HTTP Response object | HTTP::Response
#========================================================================================
sub DoRequestCurl {
    my ( $Method, $Url, $Content, $hHeaders, $hAdditionalInfo ) = @_;

    #-- https://gms.tf/when-curl-sends-100-continue.html
    $hHeaders->{'Expect'} = '';

    #-- Data structures (would appear to) get mangled by Data Dumper, so make a copy for the logging side and keep the original as it was
    my $hLogHeaders = defined $hHeaders ? {%$hHeaders} : undef;

    #-- Remove the authorization header from the log
    $hLogHeaders->{'Authorization'} = '--- removed ---' if exists $hLogHeaders->{'Authorization'};

    my $LogName = PAYMENT_TYPE_MOLLIE_PAYMENTS;
    $LogName .= ' ' . $hAdditionalInfo->{'LogName'} if $hAdditionalInfo->{'LogName'};
    my $LogSuffix = GenerateShortGUID();
    delete $hAdditionalInfo->{'LogName'};

    my $hReqLog = {
      Method          => $Method,
      Url             => $Url,
      Content         => $Content,
      hHeaders        => $hLogHeaders,
      hAdditionalInfo => $hAdditionalInfo,
    };

    my $Easy = Net::Curl::Easy->new();
    $Easy->setopt( CURLOPT_URL, $Url );
    $Easy->setopt( CURLOPT_CUSTOMREQUEST, $Method );

    my $aHeaders;
    push @$aHeaders, $_ . ':' . $hHeaders->{$_} foreach keys %$hHeaders;
    $Easy->setopt( CURLOPT_HTTPHEADER, $aHeaders );

    if ( $Content ) {
      #$Content = encode( 'utf-8', $Content );
      $Easy->setopt( CURLOPT_POSTFIELDS, $Content );
    }

    $Easy->setopt( CURLOPT_WRITEHEADER(), \$Easy->{Headers} );
    $Easy->setopt( CURLOPT_FILE(),        \$Easy->{Body} );

    LogPayment(PAYMENT_TYPE_MOLLIE_PAYMENTS, "Request $LogName $LogSuffix", $hReqLog, undef, undef);
    eval {
        $Easy->perform();
    };

    if ( ExistsError() ) {
        my $Error = GetError();
        LogError($Error);

        return undef;
    } else {

        my $ResponseCode = $Easy->getinfo(CURLINFO_RESPONSE_CODE);

        #-- We tried to log a 'nice' variant of the response by decoding the json, but it flattens the structure after about 3 levels, so ...

        my $hResLog = {
          Method         => $Method,
          Url            => $Url,
          ResponseCode   => $ResponseCode,
          Headers        => $Easy->{Headers},
          Body           => $Easy->{Body},
        };
        LogPayment(PAYMENT_TYPE_MOLLIE_PAYMENTS, "Response $LogName $LogSuffix", $hResLog, undef, undef);

        my $Res = HTTP::Response->parse( $Easy->{Headers} );
        $Res->content( $Easy->{Body} );

        return $Res;
    }
}

#========================================================================================
# function     FormatFloatForMollie
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FormattedFloatNumber = FormatFloatForMollie($FloatNumber);
#----------------------------------------------------------------------------------------
# description  It formats float number according to Mollie API V2
#----------------------------------------------------------------------------------------
# input        $FloatNumber | float number | float
# return       $FormattedFloatNumber | formatted float number string | string
#========================================================================================
sub FormatFloatForMollie {
    my ($FloatNumber) = @_;

    my $Formatter = DE_EPAGES::Core::API::Object::NumberFormat->new(
        'decimal_digits' => 2, 'decimal_fill' => 2, decimal_point  => '.', thousands_sep => ''
    );
    return $Formatter->format_number($FloatNumber);
}

#========================================================================================
# function     GetSupportedLocale
# state        public
#----------------------------------------------------------------------------------------
# syntax       $SupportedLocale = GetSupportedLocale($Locale);
#----------------------------------------------------------------------------------------
# description  It checks if provided locale is supported and returns it if so.
#               If not supported, then the most appropriated one is provided.
#----------------------------------------------------------------------------------------
# input        $Locale | locale string | string
# return       $SupportedLocale | locale string | string
#========================================================================================
sub GetSupportedLocale {
    my ($Locale) = @_;

    if ($Locale ~~ SUPPORTED_LOCALES) {
        return $Locale;
    }

    my ($Language, $Country) = split(/_/, $Locale);
    my $LanguageLocale = $Language . '_' . uc($Language);
    if ($LanguageLocale ~~ SUPPORTED_LOCALES) {
        return $LanguageLocale;
    }

    my $CountryLocale = lc($Country) . '_' . $Country;
    if ($CountryLocale ~~ SUPPORTED_LOCALES) {
        return $CountryLocale;
    }

    return DEFAULT_LOCALE;
}

#========================================================================================
# function     FilterTokensResponse
# state        public
#----------------------------------------------------------------------------------------
# syntax       FilterTokensResponse($LogString);
#----------------------------------------------------------------------------------------
# description  It filters token codes (to be used as paramter into LogPayment calls)
#----------------------------------------------------------------------------------------
# input        $LogString | string to filter | string
#========================================================================================
sub FilterTokensResponse {
    my $LogString = shift;

    $LogString =~ s/"clientId":"\w+"/"clientId":"****"/gix;
    $LogString =~ s/"clientSecret":"\w+"/"clientSecret":"****"/gix;
    $LogString =~ s/"code":"\w+"/"code":"****"/gix;
}

#========================================================================================
# function     FilterOnboardingReturnQuery
# state        public
#----------------------------------------------------------------------------------------
# syntax       FilterOnboardingReturnQuery($LogString);
#----------------------------------------------------------------------------------------
# description  It filters authorizatio code from URL query (to be used as paramter into LogPayment calls)
#----------------------------------------------------------------------------------------
# input        $LogString | string to filter | string
#========================================================================================
sub FilterOnboardingReturnQuery {
    my $LogString = shift;

    $LogString =~ s/'auth_\w+'/'auth_---replaced---'/gix;
}

#========================================================================================
# function     ParseErrorStringToHash
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hErrorResponse = ParseErrorStringToHash($ErrorString, $StatusCode);
#----------------------------------------------------------------------------------------
# description  Convert mollie error response string to a hash
#               String should contain  "MollieErrorResponse"
#               Split the error message in 2 parts: exception name, fields
#               Trim both ends [ ] of the fields part
#               Split fields by ' | '
#               Split key value pairs by =
#----------------------------------------------------------------------------------------
# input        $ErrorString | Mollie error string | string
#               $StatusCode | HTTP status code | string
# output       $hMollieResponse |                 | hash
# return       $hMollieResponse | Mollie response | hash
#========================================================================================
sub ParseErrorStringToHash {
    my ($ErrorString, $StatusCode) = @_;

    return unless isValid($ErrorString);
    return unless (index($ErrorString, "MollieErrorResponse") != -1);

    my @Words = split(/\s/, $ErrorString, 2);
    my $ExceptionName = $Words[0];
    my $ContentString = $Words[1];
    $ContentString =~ s/^\[+|\]+$//g;
    my @ContentWords = split(/\s\|\s/, $ContentString);

    my %hContent =  map { split( /=/, $_, 2 ) } @ContentWords;
    my $hErrorResponse = {
        'exception' => $ExceptionName,
        'status'    => $StatusCode,
        'content'   => \%hContent
    };

    return $hErrorResponse;
}

1;
