#========================================================================================
# package      DE_EPAGES::Object::API::XMLStructureImportHandler
# state        public
# description  Handles the import of XML files.
#========================================================================================
package DE_EPAGES::Object::API::XMLStructureImportHandler;
use base qw (DE_EPAGES::XML::API::BaseImportHandler);

use strict;
use DE_EPAGES::Object::API::Factory qw ( SplitObjectPath JoinObjectPath );
use DE_EPAGES::Object::API::Object::XMLStructure;
use DE_EPAGES::Core::API::Log qw (GetLog);

#========================================================================================
# function     new()
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Handler = DE_EPAGES::XML::API::XMLStructureImportHandler->new();
#----------------------------------------------------------------------------------------
# description  Creates a new XMLStructureImportHandler object
#               cunstructor uses additionally the constructor of super class
#----------------------------------------------------------------------------------------
# return       $Handler | XMLStructureImportHandler | object
#========================================================================================
sub new {
    my $class = shift;
    my $self = $class->SUPER::new();

    return $self;
}

#========================================================================================
# function     _importXML
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Package->_importXML( $FileName, $hOptions );
#----------------------------------------------------------------------------------------
# example      $Handler = $Package->new;
#               $Handler->_importXML( $FileName, {'register_notrans' => 1, 'distinguish_multiple_appearence' => 0} );
#----------------------------------------------------------------------------------------
# description  Imports classes and attributes from an XML file.
#----------------------------------------------------------------------------------------
# input        $FileName | XML file name | string
# input        $hOptions  | option flags
#               <ul>
#                   <li>register_notrans - notrans tags often have no effect on import into db, default is '0' - bool
#                   <li>distinguish_multiple_appearence - import doesn't care about how often a tag appears;
#                       default value is set to '0', whereas setting this value to '1' enlarges the element path by ':appearence(..)' - bool
#                   </li>
#               </ul>
#========================================================================================
sub _importXML {
    my $self = shift;
    my ($FileName, $hOptions) = @_;


    #stack to remember the object pathes while passing tags
    $self->{'_XMLStructStack-ObjectPaths'} = {};
    #stack to remember the names of packages, while passing cartridge tags
    $self->{'_XMLStructStack-CartridgeTags'} = {};
    #stack for language tags in translation files
    $self->{'_XMLStructStack-TranslationLanguageTags'} = {};

    #placeholder for attributes of <File> tag
    $self->{'_XMLStructPlaceholder-FileAttributes'} = {};

    #placeholder used for error message
    $self->{'FileName'} = $FileName;

    #placeholder is used to count how many tags already passed
    $self->{'PositionCount'} = 0;

    #control variables decided what kind of xml should be imported
    #already there are only two kind normal and translation xml files
    #they differ only a little bit, but not patch.xx.xml
    if (exists $hOptions->{'XML-Type'}){
        $self->{'XML-Type'} = $hOptions->{'XML-Type'};
    }
    elsif (
        ($FileName =~ /\.(\w\w|\w\w-\w\w)\.xml$/x or
        $FileName =~ /Translation.+\_(\w\w|\w\w-\w\w)\.xml$/x)
        and not $FileName =~ /patch\.(\w\w|\w\w-\w\w)\.xml$/x
    ){
        $self->{'XML-Type'} = 'epages_translation';
    }
    elsif($FileName =~ /Cartridges[\_\w]*\.xml/){
        $self->{'XML-Type'} = 'epages_cartridges_xml';
    }
    elsif($FileName =~ /ShopChildDependencies\.xml/){
        $self->{'XML-Type'} = 'epages_shop_child_dependencies';
    }
    else{
        $self->{'XML-Type'} = 'epages_standard';
    }

    #result sematic structure object
    $self->{'_XMLStruct-XMLStructure'} = DE_EPAGES::Object::API::Object::XMLStructure->new(
        $self->{'XML-Type'}
    );

    #set value to check if registering notrans tags is necessary
    $self->{'register_notrans'} = 0;
    $self->{'register_notrans'} = 1 if $hOptions->{'register_notrans'};

    #set value to check if distingtion for multiple appearence of tags is necessary
    $self->{'distinguish_multiple_appearence'} = 0;
    $self->{'distinguish_multiple_appearence'} = 1 if $hOptions->{'distinguish_multiple_appearence'};

    $self->SUPER::importXML($FileName);

    return;
}

#========================================================================================
# function     getXMLStructure
# state        public
#----------------------------------------------------------------------------------------
# syntax       $XMLStructure = $self->getXMLStructure($FileName, $hOptions);
#----------------------------------------------------------------------------------------
# example      $Handler = $Package->new;
#               $Handler->_importXML( $FileName, {'register_notrans' => 1, 'distinguish_multiple_appearence' => 0} );
#----------------------------------------------------------------------------------------
# description  Returns the xml structure object of an xml file.
#----------------------------------------------------------------------------------------
# input        $FileName | XML file name | string
# input        $hOptions  | option flags
#               <ul>
#                   <li>register_notrans - notrans tags often have no effect on import into db, default is '0' - bool</li>
#                   <li>distinguish_multiple_appearence - import doesn't care about how often a tag appears;
#                       default value is set to '0', whereas setting this value to '1' enlarges the element path by ':appearence(..)' - bool
#                   </li>
#               </ul>
#               | ref.hash
# return       $XMLStructure | xml structure object | object
#========================================================================================
sub getXMLStructure{
    my $self = shift;
    my ($FileName, $hOptions) = @_;

    $self->_importXML($FileName, $hOptions);

    return $self->{'_XMLStruct-XMLStructure'};
}

##no critic qw(ProhibitCascadingIfElse,ProhibitExcessComplexity)
sub start_element {#function is execute when a tag is started
    my $self = shift;
    my ($hElement) = @_;
    my $TagName = $hElement->{LocalName};
    my $hAttributes = $self->getAttributes( $hElement );

    if($TagName eq 'System' or (defined $hAttributes->{'Path'} and $hAttributes->{'Path'} eq '/')){
        $self->addSystemToStructure($hAttributes);
    }
    elsif (defined $hAttributes->{'Path'} and defined $hAttributes->{'Alias'}){#bad case
        my $Path = $hAttributes->{'Path'};
        $Path =~ s/[\\\/]\w*$//;
        $Path = $Path . '/' . $hAttributes->{'Alias'};
        delete $hAttributes->{'Alias'};
        $self->addObjectByPathToStructure($TagName,$hAttributes->{'Path'},$hAttributes);
    }
    elsif (defined $hAttributes->{'Path'}){#add object tag by path
        $self->addObjectByPathToStructure($TagName,$hAttributes->{'Path'},$hAttributes);
    }
    elsif(defined $hAttributes->{'Alias'}){#add object tag by alias
        $self->addObjectToStructure($TagName, $hAttributes->{'Alias'}, $hAttributes);
    }
    elsif($TagName eq 'AttributeValue'){#add attribute value tag
        $self->addAttributeToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'Permission'){#add permission tag
        $self->addPermissionToStructure( $TagName, $hAttributes )
    }
    elsif($TagName eq 'AutoCrossSelling'){#add autocrossselling tag
        $self->addElementToStructure($TagName, 'Product', $hAttributes);
    }
    elsif($TagName eq 'GroupMember'){#add groupmember tag
        $self->addObjectReferenceToStructure( $TagName, 'Trustee', $hAttributes )
    }
    elsif($TagName eq 'MemberOf'){#add memberof tag
        $self->addObjectReferenceToStructure($TagName, 'Group', $hAttributes);
    }
    elsif($TagName eq 'TaxRate'){#add taxrate tag
        $self->addTaxRateToStructure($TagName, $hAttributes);
    }
    elsif($TagName eq 'TaxRateDigital'){#add taxrate digital tag
        $self->addTaxRateDigitalToStructure($TagName, $hAttributes);
    }
    elsif($TagName eq 'AttributeLanguage'){#add localized attribute tag
        $self->addLocalizedAttributeToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'ViewAction'){#add view action tag
        $self->addElementToStructure( $TagName, 'URLAction', $hAttributes );
    }
    elsif($TagName eq 'Menu'){#ad menu tag
        $self->addMenuToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'Template'){#add template tag
        $self->addTemplateToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'FeatureCollectionFeature'){#add feature collection feature tag
        $self->addElementToStructure( $TagName, 'FeatureRef', $hAttributes );
    }
    elsif($TagName eq 'ShopCurrency'){#add shop currency tag
        $self->addObjectToStructure($TagName,$hAttributes->{'CurrencyID'},$hAttributes);
    }
    elsif($TagName eq 'RoleAction'){#add role action tag
        $self->addRoleActionToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'AttributePermission'){#add attributepermission tag
        $self->addAttributePermissionToStructure($TagName, $hAttributes);
    }
    elsif($TagName eq 'WSRoleMethod'){#add web service role method tag
        $self->addWSRoleMethodToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'HookFunction'){#add hook function tag
        $self->addHookFunctionToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'Hook'){#add hook tag
        $self->addElementToStructure($TagName, 'Name', $hAttributes);
    }
    elsif($TagName eq 'epages'){#add epages tag
        $self->addEpagesToStructure($hAttributes);
    }
    elsif($TagName eq 'Attribute' and
        not defined $hAttributes->{'Path'} and
        not defined $hAttributes->{'Alias'} and
        $self->{'XML-Type'} eq 'epages_translation'
    ){#add attribute tag, from translation xml file, by alias
        $self->addTranslationAttributeToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'Language' and
        not defined $hAttributes->{'Path'} and
        not defined $hAttributes->{'Alias'} and
        $self->{'XML-Type'} eq 'epages_translation'
    ){#add language tag, from translation xml file
        $self->addTranslationLanguageToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'Language'){
        $self->addElementToStructure($TagName,'Code3', $hAttributes);
    }
    elsif($TagName eq 'Cartridge'){#add cartridge tag
        #check for xml type is epages_cartridges_xml
        if (
            $self->{'XML-Type'} eq 'epages_cartridges_xml' or
            $self->{'XML-Type'} eq 'epages_shop_child_dependencies'
        ){
            $self->addElementToStructure($TagName, 'Package', $hAttributes);
        }
        else {
            $self->addCartridgeToStructure($TagName,$hAttributes);
        }
    }
    elsif($TagName eq 'Form'){#add form tag
        $self->addElementToStructure($TagName, 'Name', $hAttributes);
    }
    elsif($TagName eq 'FormLoop'){#add form loop tag
        $self->addElementToStructure($TagName, 'Name', $hAttributes);
    }
    elsif($TagName eq 'FormField'){#add form field tag
        $self->addElementToStructure( $TagName, 'Name', $hAttributes );
    }
    elsif($TagName eq 'Country'){
        $self->addElementToStructure( $TagName, 'CountryID', $hAttributes )
    }
    elsif($TagName eq 'CustomerUser'){
        $self->addElementToStructure( $TagName, 'User', $hAttributes )
    }
    elsif($TagName eq 'CountrySubdivision'){
        $self->addElementToStructure( $TagName, 'Code', $hAttributes );
    }
    elsif($TagName eq 'Locale'){
        $self->addElementToStructure( $TagName, 'LocaleID', $hAttributes )
    }
    elsif($TagName eq 'Currency'){
        $self->addElementToStructure( $TagName, 'CurrencyID', $hAttributes )
    }
    elsif($TagName eq 'SearchSort'){#add search sort tag
        $self->addElementToStructure( $TagName, 'Parameter', $hAttributes )
    }
    elsif($TagName eq 'SearchParameter'){
        $self->addElementToStructure( $TagName, 'Position', $hAttributes )
    }
    elsif($TagName eq 'SearchDefinition'){#add search definition tag
        $self->addSearchDefinitionToStructure($TagName,$hAttributes);
    }
    elsif($TagName eq 'SearchFrom'){#add search from tag
        $self->addElementToStructure( $TagName, 'TableName', $hAttributes )
    }
    elsif($TagName eq 'StyleGroupMap'){
        $self->addElementToStructure( $TagName, 'StyleTemplate', $hAttributes )
    }
    elsif($TagName eq 'StyleTemplateVariation'){
        $self->addElementToStructure( $TagName, 'StyleTemplate', $hAttributes )
    }
    elsif($TagName eq 'ContentChildClass'){
        $self->addElementToStructure( $TagName, 'ChildClass', $hAttributes )
    }
    elsif($TagName eq 'ShopFeature'){
        $self->addElementToStructure( $TagName, 'Feature', $hAttributes )
    }
    elsif($TagName eq 'ShopLanguage'){
        $self->addElementToStructure( $TagName, 'Language', $hAttributes )
    }
    elsif($TagName eq 'ShopLocale'){
        $self->addElementToStructure( $TagName, 'LocaleID', $hAttributes )
    }
    elsif($TagName eq 'ShopCountry'){
        $self->addElementToStructure( $TagName, 'CountryID', $hAttributes )
    }
    elsif($TagName eq 'ProductPrice'){
        $self->addElementToStructure( $TagName, 'CurrencyID', $hAttributes )
    }
    elsif($TagName eq 'ShippingLevel'){
        # this works only for fix-price methods
        $self->addShippingLevelToStructure( $TagName, $hAttributes )
    }
    elsif($TagName eq 'PaymentLevel'){
        $self->addElementToStructure( $TagName, 'CurrencyID', $hAttributes )
    }
    elsif($TagName eq 'EnableShipping'){
        $self->addObjectReferenceToStructure( $TagName, 'ShippingMethod', $hAttributes )
    }
    elsif($TagName eq 'StoreCartridge'){
        $self->addElementToStructure( $TagName, 'Cartridge', $hAttributes )
    }
    elsif($TagName eq 'StoreFeature'){
        $self->addElementToStructure( $TagName, 'FeatureRef', $hAttributes )
    }
    elsif($TagName eq 'ProviderStore'){
        $self->addElementToStructure( $TagName, 'Store', $hAttributes )
    }
    elsif($TagName eq 'ShopTypeFeaturePack'){
        $self->addElementToStructure( $TagName, 'FeaturePack', $hAttributes )
    }
    elsif($TagName eq 'Dependency'){
        $self->addElementToStructure( $TagName, 'Package', $hAttributes )
    }
    elsif($TagName eq 'TaxMatrixValue'){
        $self->addElementToStructure( $TagName, 'Name', $hAttributes )
    }
    elsif($TagName eq 'TaxMatrixDigitalValue'){
        $self->addElementToStructure( $TagName, 'Name', $hAttributes )
    }
    elsif($TagName eq 'notrans' and $self->{'XML-Type'} eq 'epages_translation'){
        #register no trans tag in tag text if it is forced
        $self->{'Characters'} .= '<notrans>' if $self->{'register_notrans'};
    }
    elsif($TagName eq 'File'){
        $self->addFileToStructure($hAttributes);
    }
    elsif($TagName eq 'PatchFunction'){
        $self->addElementToStructure( $TagName, 'Method', $hAttributes )
    }
    elsif(
        $TagName eq 'Patch' or
        $TagName eq 'PatchCartridgeDir' or
        $TagName eq 'PatchCartridgeDirPriorPart' or
        $TagName eq 'PatchCartridgeDirFinalPart' or
        $TagName eq 'PatchDBStructure' or
        $TagName eq 'PatchStoreTypeDir'
    ){
        $self->addPatchComponentToStructure( $TagName )
    }
    elsif($TagName eq 'PatchCartridge'){
        $self->addPatchCartridgeToStructure( $TagName, $hAttributes );
    }
    elsif($TagName eq 'PageTypeHelpTopic' or $TagName eq 'PageTypeHelpTopicExt'){
        $self->addElementToStructure( $TagName, 'HelpTopic', $hAttributes )
    }
    elsif($TagName eq 'TaxAreaCountry'){
        #distinction of cases, taxareacountry or taxareacountry region
        if(defined $hAttributes->{'ZipcodeFrom'} and defined $hAttributes->{'ZipcodeFrom'}){
            $self->addTaxAreaCountryRegionToStructure( $TagName, $hAttributes );
        }
        else{
            $self->addElementToStructure( $TagName, 'CountryID', $hAttributes );
        }
    }
    elsif($TagName eq 'TaxAreaDigitalCountry'){
        $self->addElementToStructure( $TagName, 'CountryID', $hAttributes );
    }

    elsif($TagName eq 'StyleHeaderImageMap'){
        $self->addElementToStructure( $TagName, 'StyleHeaderImage', $hAttributes )
    }
    elsif($TagName eq 'StyleBodyImageMap'){
        $self->addElementToStructure( $TagName, 'StyleBodyImage', $hAttributes )
    }
    elsif($TagName eq 'EbayCallType'){
        $self->addElementToStructure( $TagName, 'Name', $hAttributes )
    }
    elsif($TagName eq 'Translation'){
        $self->addTranslationToStructure( $TagName, $hAttributes )
    }
    elsif($TagName eq 'DownloadProductMap'){
        $self->addElementToStructure( $TagName, 'TargetUrl', $hAttributes )
    }
    elsif($TagName eq 'BundleProductMap'){
        $self->addElementToStructure( $TagName, 'BundledProduct', $hAttributes )
    }
    elsif($TagName eq 'TemplateMap'){
        $self->addElementToStructure( $TagName, 'TemplateType', $hAttributes )
    }
    elsif($self->{'XML-Type'} eq 'epages_cartridges_xml' and $TagName eq 'DatabaseType'){
        $self->addElementToStructure( $TagName, 'Name', $hAttributes )
    }
    elsif($TagName eq 'MailTypeTemplateTLEMap'){#add MailTypeTemplateTLEMap tag
        $self->addElementToStructure($TagName, 'MailTypeTemplateTLE', $hAttributes);
    }
    else {#unknown tag
        my $Keys = join(', ', sort keys %$hAttributes);
        die "Unknown Tag: $TagName (keys: $Keys) under path " . $self->_actualPath() . ' in file ' . $self->{'FileName'} . ' complete XMLStructureImportHandler';
    }

    #clean id from tag attributes if exists
    my $ObjectPaths = $self->_actualPath();
    if (defined $ObjectPaths and
        $ObjectPaths ne '' and
        $ObjectPaths ne '_XMLStruct-IgnoreID' and
        not ($ObjectPaths =~ /^\Q_XMLStruct-AttributeCollectCharacter-\E/x)){
        my %ID = %{$self->{'_XMLStruct-XMLStructure'}->getID($ObjectPaths)};

        #delete id from attributes
        foreach my $IDName (keys %ID){
            $self->{'_XMLStruct-XMLStructure'}->deleteTagAttribute(
                    $ObjectPaths,
                    $IDName
            );
        }
    }
    #compute import position
    $self->{'PositionCount'} = $self->{'PositionCount'} + 1;
}

#function ist execute when tag is ending, calls finish_Element()
sub end_element {
    my $self = shift;
    my ($hElement) = @_;
    my $TagName = $hElement->{LocalName};

#=============================================================================
#=notice attributes have a special handling because they have no main entry in
#=the xml structure they where positionate under _XMLStructAttributes
#=in entry of parent tag
#==============================================================================

    if ($TagName eq 'notrans'){
        #register no trans tag in tag text if it is forced
        $self->{'Characters'} .= '</notrans>' if $self->{'register_notrans'};
    }
    else{#normal tags
        #cartridge package name no longer need because tag is closed
        $self->popID('_XMLStructStack-CartridgeTags') if (
            ($TagName eq 'Cartridge') and (
                ($self->{'XML-Type'} ne 'epages_cartridges_xml') and
                ($self->{'XML-Type'} ne 'epages_shop_child_dependencies')
            )
        );
        $self->popID('_XMLStructStack-TranslationLanguageTags') if ($TagName eq 'Language' and $self->{'XML-Type'} eq 'epages_translation');

        #check for text between tag
        if ($self->collectCharacters){#text is between tag
            if ($self->_actualPath() =~ /^\Q_XMLStruct-AttributeCollectCharacter-\E/x){
                #attribute value tag with text inside, infomation is find on stack for object paths

                #collect attribute value from text between tag
                my $AttributeName = $self->_actualPath();
                $AttributeName =~ s/^\Q_XMLStruct-AttributeCollectCharacter-\E//x;
                #remove attribute name from stack for object paths
                $self->popID('_XMLStructStack-ObjectPaths');

                my $AttributeValue = $self->getCharacters;
                $self->collectCharacters(0);

                $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                    $self->_actualPath(),
                    $AttributeName,
                    $AttributeValue
                );

                #have to push an ignore-id because "pop" at the end of function
                $self->pushID('_XMLStructStack-ObjectPaths', '_XMLStruct-IgnoreID' );
            }
            elsif($self->{'_XMLStruct-XMLStructure'}->getClass($self->_actualPath()) eq 'TranslationAttribute') {
                #Attribute tag from TranslationFile
                $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                    $self->_actualPath(),
                    '_Tag-Text',
                    $self->getCharacters
                );
                $self->collectCharacters(0);
            }
            elsif($self->{'_XMLStruct-XMLStructure'}->getClass($self->_actualPath()) eq 'Translation') {
                #Translation tag from TranslationFile/Dictionary File
                $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                    $self->_actualPath(),
                    '_Tag-Text',
                    $self->getCharacters
                );
                $self->collectCharacters(0);
            }
            elsif($TagName eq 'File'){#File is a special tag <File>xyz.jpg</File> identifier is only known at the closing tag
                my $FakeAlias = $self->getCharacters;
                $self->collectCharacters(0);
                my $hFileAttributes = $self->{'_XMLStructPlaceholder-FileAttributes'};
                $hFileAttributes->{'_Tag-Text'} = $FakeAlias;

                $self->addElementToStructure('File', '_Tag-Text', $hFileAttributes);
            }
            else{#not known
                $self->collectCharacters(0);
                die 'dont know what to do with text contained in tag: ' . $self->{'_XMLStruct-XMLStructure'}->getClass($self->_actualPath());
            }
        };

        #delete ignore tag from xml-structure
        my $ObjectPaths = $self->_actualPath();
        if (defined $ObjectPaths && $ObjectPaths eq '_XMLStruct-IgnoreID'){
            $self->{'_XMLStruct-XMLStructure'}->deleteElement('_XMLStruct-IgnoreID');
        }

        #dont call popID if tag has no entry in '_XMLStructStack-ObjectPaths'
        return if ($TagName eq 'epages');
        return if (($TagName eq 'Language') and ($self->{'XML-Type'} eq 'epages_translation'));
        return if (
            ($TagName eq 'Cartridge') and (
                ($self->{'XML-Type'} ne 'epages_cartridges_xml') and
                ($self->{'XML-Type'} ne 'epages_shop_child_dependencies')
            )
        );

        #processing of actual tag is completed, delete related element from stack
        $self->popID('_XMLStructStack-ObjectPaths');
    }
}

#========================================================================================
# function     addSystemToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addSystemToStructure($hAttributes);
#----------------------------------------------------------------------------------------
# description  add system object tag to xml structure
#----------------------------------------------------------------------------------------
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addSystemToStructure{
    my $self = shift;
    my ($hAttributes) = @_;

    #register path
    $self->_registerPath('');

    $self->{'_XMLStruct-XMLStructure'}->addElement(
        '',
        {'Path' => $hAttributes->{'Path'}},
        {
            'Class' => 'System',
            'TagName' => 'System',
            'Position' => $self->{'PositionCount'},
            'NonEffectiveForSubPathes' => '1',

        },
        $hAttributes
    );

    #all object's addressed by path in Translation.*.xx.xml have to referenced
    $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                $self->_actualPath(),
                'reference',
                '1'
    ) if ($self->{'XML-Type'} eq 'epages_translation');
}

#========================================================================================
# function     addObjectToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addObjectToStructure($TagName, $Alias,$hAttributes);
#----------------------------------------------------------------------------------------
# description  add object tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $Alias | value of tag attribute Alias | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addObjectToStructure{
    #add an Object to structure
    #object path is created by having a look on actual object
    my $self = shift;
    my ($TagName, $Alias,$hAttributes) = @_;
    my $Class = $TagName;
    $Alias = $self->_quoteAlias($Alias);

    if($TagName eq 'Feature' or $TagName eq 'TemplateType'){#have to add cartridge in attributes
        if (not defined $hAttributes->{'Cartridge'} and
            ($self->topID('_XMLStructStack-CartridgeTags') ne ''))
        {#take cartridge name from overlaying parent cartridge tag (cartridge stack)
            $hAttributes->{'Cartridge'} = $self->topID('_XMLStructStack-CartridgeTags');
        }
    }

    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();

    my $ObjectPath = $TopElement . '/' . $Alias;

    if (defined $hAttributes->{'Class'}){
        $Class = $hAttributes->{'Class'};
        $TagName = $hAttributes->{'Class'};
        delete $hAttributes->{'Class'};
    }

    #register path
    $self->_registerPath($ObjectPath);

    #transform relative path to base object
    if (defined $hAttributes->{'Base'}){
        $hAttributes->
            {'Base'} =
            $self->_getPathByRelativePath(
                $hAttributes->{'Base'},
                $self->_actualPath()
            );
    }

    #register id
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'Alias' => $Alias}
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,$Class,$hAttributes);

}

#========================================================================================
# function     addEpagesToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addEpagesToStructure($hAttributes);
#----------------------------------------------------------------------------------------
# description  add epages tag to xml structure
#----------------------------------------------------------------------------------------
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addEpagesToStructure{
    my $self = shift;
    my ($hAttributes) = @_;
    my $Class = 'epages';

    #register FakeObjectPath
    $self->_registerPath('epages' );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {}
    );

    #<epages> ... </epages> tag have no effect on object path of contained object tags
    $self->{'_XMLStruct-XMLStructure'}->setXMLStructAttribute(
        $self->_actualPath(),
        'NonEffectiveForSubPathes',
        '1'
    );

    #fill xml structure
    $self->_finishXMLStructElement('epages',$Class,$hAttributes);

    $self->popID('_XMLStructStack-ObjectPaths');
}

#========================================================================================
# function     addCartridgeToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addCartridgeToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add cartridge tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addCartridgeToStructure{
    my $self = shift;
    my ($TagName,$hAttributes) = @_;

    #ignore attributes of tag because cartridge tags have no effect on object pathes
    $self->pushID('_XMLStructStack-CartridgeTags', $hAttributes->{'Package'} );
}

#========================================================================================
# function     addAttributeToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addAttributeToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add attribute tag to xml structure, differs between localized attributes
#               and normal attributes without language
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addAttributeToStructure{#add attribute to actual object in structure
    my $self = shift;
    my ($TagName,$hAttributes) = @_;

#==============================================================================
#=notice attributes have a special handling because they have no main entry in
#=the xml structure they where positionate under _TagAttributes
#=in entry of parent tag, so they need a key on stack that can be ignored by
#=function end_element
#=a special case is when the attribute contains text between tag, than name
#=the name of attribute have to remember, here the name will put to stack
#==============================================================================

    if (!exists $hAttributes->{'Language'}){#no localized attribute
        if (exists $hAttributes->{'Value'}){#normal attribute tag
            $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                $self->_actualPath(),
                $hAttributes->{'Name'},
                $hAttributes->{'Value'}
            );

            #have to push an ignore-id because end_element function do "pop" everytime
            $self->pushID('_XMLStructStack-ObjectPaths', '_XMLStruct-IgnoreID' );
        }
        else{#special case, text between attribute tag

            #have to remember for function end_element that here is a attributetag that have the Value as text between tag
            $self->pushID('_XMLStructStack-ObjectPaths', '_XMLStruct-AttributeCollectCharacter-' . $hAttributes->{'Name'} );
            $self->collectCharacters(1);
        }
    }
    else{#Attribute is localized
        #create fake Alias
        my $FakeAlias = 'AttributeValue{Language->' . $hAttributes->{'Language'} . ';Name->' . $hAttributes->{'Name'} . '}';

        my $TopElement = $self->_actualPath();


        $TopElement='' unless defined $self->_actualPath();
        my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
        $self->_registerPath($ObjectPath );

        #fill all attributes into xml-structure
        foreach my $Key (keys %$hAttributes){
            #dont set reference if delete is already settet, this occures when a defined object is referenced in same file
            next if ($Key eq 'reference' and $self->{'_XMLStruct-XMLStructure'}->existsTagAttribute(
                $self->_actualPath(),'delete')
            );
            $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                $self->_actualPath(),
                $Key,
                $hAttributes->{$Key}
            );
        }

        #register ID
        $self->{'_XMLStruct-XMLStructure'}->setID(
            $self->_actualPath(),
            {'Language' => $hAttributes->{'Language'}, 'Name' => $hAttributes->{'Name'}}
        );

        #register fake class
        $self->{'_XMLStruct-XMLStructure'}->setXMLStructAttribute(
            $self->_actualPath(),
            'Class',
            'AttributeValue'
        );

        #register TagName
        $self->{'_XMLStruct-XMLStructure'}->setXMLStructAttribute(
            $self->_actualPath(),
            'TagName',
            $TagName
        );

        #register position of tag via import order
        $self->{'_XMLStruct-XMLStructure'}->setXMLStructAttribute(
            $self->_actualPath(),
            'Position',
            $self->{'PositionCount'}
        );

    }
}

#========================================================================================
# function     addLocalizedAttributeToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addLocalizedAttributeToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add attribue value tag to xml structure, attribute are localized
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addLocalizedAttributeToStructure{#add localized attribute to actual object in structure
    my $self = shift;
    my ($TagName,$hAttributes) = @_;

    #transform into a attributevalue structure, in reason of uniquess
    my %TempAttributes = %$hAttributes;
    delete $TempAttributes{'Language'};
    my @Keys = keys %TempAttributes;
    my $Count = 0;

    #attribute language
    foreach my $Key (sort @Keys){

        $Count += 1;

        my $Value = $TempAttributes{$Key};

        #create fake Alias
        my $FakeAlias = 'AttributeValue{Language->' . $hAttributes->{'Language'} . ';Name->' . $Key . '}';

        my $TopElement = $self->_actualPath();
        $TopElement='' unless defined $self->_actualPath();

        my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
        $self->_registerPath($ObjectPath );

        $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                $self->_actualPath(),
                'Name',
                $Key
        );
        $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                $self->_actualPath(),
                'Value',
                $Value
        );
        $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                $self->_actualPath(),
                'Language',
                $hAttributes->{'Language'}
        );

        #register ID
        $self->{'_XMLStruct-XMLStructure'}->setID(
            $self->_actualPath(),
            {'Language' => $hAttributes->{'Language'}, 'Name' => $Key}
        );

        #fill xml structure
        $self->_finishXMLStructElement('AttributeValue','AttributeValue',{});

        #remove the current path but not on the last attribute tag, because the end_element function will also call a pop
        $self->popID('_XMLStructStack-ObjectPaths') if ($Count < scalar @Keys);
    }
}

#========================================================================================
# function     addTranslationLanguageToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addTranslationLanguageToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add language tag to xml structure, only used in translation xml files
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addTranslationLanguageToStructure{
    my $self = shift;
    my ($TagName,$hAttributes) = @_;
    my $Class = 'Language';

    $self->_registerPath('Language' );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'Language' => $hAttributes->{'Language'}}
    );

    #<Language> ... </Language> tag have no effect on object path of contained object tags
    $self->{'_XMLStruct-XMLStructure'}->setXMLStructAttribute(
        $self->_actualPath(),
        'NonEffectiveForSubPathes',
        '1'
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,$Class,$hAttributes);

    #remember actual language
    $self->pushID('_XMLStructStack-TranslationLanguageTags', $hAttributes->{'Language'});

    #tag has no influence on object paths
    $self->popID('_XMLStructStack-ObjectPaths');
}

#========================================================================================
# function     addTranslationAttributeToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addTranslationAttributeToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add translation attribute tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addTranslationAttributeToStructure{#add attribute to actual object in structure
    my $self = shift;
    my ($TagName,$hAttributes) = @_;

    #create fake Alias
    my $FakeAlias = 'AttributeValue{Language->' . $self->topID('_XMLStructStack-TranslationLanguageTags') . ';Name->' . $hAttributes->{'Name'} . '}';

    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID, get one part of id from language stack, because in translatio
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {
            'Language' => $self->topID('_XMLStructStack-TranslationLanguageTags'),
            'Name' => $hAttributes->{'Name'}
        }
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,'TranslationAttribute',$hAttributes);

    #collect characters between tag
    $self->collectCharacters(1);
}

#========================================================================================
# function     addShippingLevelToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addShippingLevelToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add shipping level tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addShippingLevelToStructure {
    my $self = shift;
    my ($TagName,$hAttributes) = @_;

    #create fake Alias
    my $FakeAlias = 'ShippingLevel{CurrencyID->' . $hAttributes->{'CurrencyID'} . ';LowerBound->' . $hAttributes->{'LowerBound'} . '}';

    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID, get one part of id from language stack, because in translatio
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {
            'CurrencyID' => $hAttributes->{'CurrencyID'},
            'LowerBound' => $hAttributes->{'LowerBound'}
        }
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,'ShippingLevel',$hAttributes);
}

#========================================================================================
# function     addTaxAreaCountryRegionToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addTaxAreaCountryRegionToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add add tax area country tag, with region info to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addTaxAreaCountryRegionToStructure {
    my $self = shift;
    my ($TagName,$hAttributes) = @_;

    #create fake Alias
    my $FakeAlias = 'TaxAreaCountryRegion{' .
        'CountryID->'    . $hAttributes->{'CountryID'} .
        ';ZipcodeFrom->' . $hAttributes->{'ZipcodeFrom'} .
        ';ZipcodeTo->'   . $hAttributes->{'ZipcodeTo'} .
    '}';

    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID, get one part of id from language stack, because in translatio
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {
            'CountryID' => $hAttributes->{'CountryID'},
            'ZipcodeFrom' => $hAttributes->{'ZipcodeFrom'},
            'ZipcodeTo' => $hAttributes->{'ZipcodeTo'}
        }
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,'TaxAreaCountryRegion',$hAttributes);
}

#========================================================================================
# function     addTranslationToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addTranslationToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add translation tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addTranslationToStructure{
    my $self = shift;
    my ($TagName,$hAttributes) = @_;

    #create fake Alias
    my $FakeAlias = 'Translation{Keyword->' . $hAttributes->{'Keyword'} . '}';

    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID, get one part of id from language stack, because in translatio
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {
            'Keyword' => $hAttributes->{'Keyword'}
        }
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,'Translation',$hAttributes);

    #collect characters between tag
    $self->collectCharacters(1, {});
}

#========================================================================================
# function     addRoleActionToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addRoleActionToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add role action tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addRoleActionToStructure{#add RoleAction to actual object in structure
    my $self = shift;
    my ($TagName,$hAttributes) = @_;


    my $Action;

    #check if class is defined in tag attributes
    if (defined $hAttributes->{'Class'}){

        #get action from class and action
        $Action = '/Classes/' . $hAttributes->{'Class'} . '/Actions/' . $hAttributes->{'Action'};

        $hAttributes->{'Action'} = $Action;

        delete $hAttributes->{'Class'};
    }
    else{

        #take action as it is
        $Action = $hAttributes->{'Action'};
    }

    #create fake Alias
    my $FakeAlias = 'RoleAction{Action->' . $Action . '}';

    my $TopElement = $self->_actualPath();
    delete $hAttributes->{'Path'};
    $TopElement='' unless defined $self->_actualPath();
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);

    #register ObjectPath
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'Action' => $hAttributes->{'Action'}}
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,'RoleAction',$hAttributes);
}

#========================================================================================
# function     addWSRoleMethodToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addWSRoleMethodToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add web service role method tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addWSRoleMethodToStructure{#add WSRoleMethod to actual object in structure
    my $self = shift;
    my ($TagName,$hAttributes) = @_;

    #create ID
    my $FakeAlias = 'WSRoleMethod{Method->' . $hAttributes->{'Method'} . ';WebService->' . $hAttributes->{'WebService'} . '}';

    my $TopElement = $self->_actualPath();
    delete $hAttributes->{'Path'};
    $TopElement='' unless defined $self->_actualPath();
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);

    #register ObjectPath
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'Method' => $hAttributes->{'Method'}, 'WebService' => $hAttributes->{'WebService'}}
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,'WSRoleMethod',$hAttributes);
}

#========================================================================================
# function     addHookFunctionToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addHookFunctionToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add hook function tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addHookFunctionToStructure{
    my $self = shift;
    my ($TagName,$hAttributes) = @_;

    #create fake Alias
    my $FakeAlias = 'HookFunction{FunctionName->' . $hAttributes->{'FunctionName'} . '}';

    my $TopElement = $self->_actualPath();
    delete $hAttributes->{'Path'};
    $TopElement='' unless defined $self->_actualPath();
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);

    #register ObjectPath
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'FunctionName' => $hAttributes->{'FunctionName'}}
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,'HookFunction',$hAttributes);
}

#========================================================================================
# function     addTaxRateToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addTaxRateToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add tax rate tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addTaxRateToStructure{
    my $self = shift;
    my ($TagName, $hAttributes) = @_;
    my $Class = $TagName;
    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();

    my $ObjectPath;

    #<TaxArea Alias='Not European Union'>
    #  <TaxRate TaxClass='normal' Rate='4' />

    #create fake alias
    my $FakeAlias = 'TaxRate{TaxClass->' . $self->_quoteAlias($hAttributes->{'TaxClass'}) . '}';

    if(defined $hAttributes->{'TaxArea'}){
        $ObjectPath = $TopElement . '/' . $self->_quoteAlias($hAttributes->{'TaxArea'}) . '/' . $FakeAlias;
        delete $hAttributes->{'TaxArea'};
    }
    else{
        $ObjectPath = $TopElement . '/' . $FakeAlias;
    }

    $self->_registerPath($ObjectPath );

    #register id
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'TaxClass' => $hAttributes->{'TaxClass'}}
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,$Class,$hAttributes);
}

#========================================================================================
# function     addTaxRateDigitalToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addTaxRateDigitalToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add tax rate digital tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addTaxRateDigitalToStructure{
    my $self = shift;
    my ($TagName, $hAttributes) = @_;
    my $Class = $TagName;
    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();

    my $ObjectPath;

    #<TaxAreaDigital Alias='Germany>'>
    #  <TaxRateDigital TaxClassDigital='ebooks' Rate='4' />

    #create fake alias
    my $FakeAlias = 'TaxRateDigital{TaxClassDigital->' . $self->_quoteAlias($hAttributes->{'TaxClassDigital'}) . '}';

    if(defined $hAttributes->{'TaxAreaDigital'}){
        $ObjectPath = $TopElement . '/' . $self->_quoteAlias($hAttributes->{'TaxAreaDigital'}) . '/' . $FakeAlias;
        delete $hAttributes->{'TaxAreaDigital'};
    }
    else{
        $ObjectPath = $TopElement . '/' . $FakeAlias;
    }

    $self->_registerPath($ObjectPath );

    #register id
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'TaxClassDigital' => $hAttributes->{'TaxClassDigital'}}
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,$Class,$hAttributes);
}


#========================================================================================
# function     addObjectByPathToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addObjectByPathToStructure($TagName, $Path, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add object tag to xml structure, object is addressed by path
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $Path | value of tag attribute Alias | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addObjectByPathToStructure{#add an object to structure, using relative path to get object path
    my $self = shift;
    my ($TagName, $Path, $hAttributes) = @_;
    my $Class = $TagName;
    delete $hAttributes->{'Path'};

    #get real path from e.g. '../../Attributes/Test'
    my $ObjectPath = $self->_getPathByRelativePath($Path);
    my $QuotetObjectPath = JoinObjectPath(SplitObjectPath($ObjectPath));

    #check for valid object path
    if ($ObjectPath ne $QuotetObjectPath){
        my $File = $self->{'FileName'};
        GetLog()->info("error invalid path in xml detected, file '$File' path '$ObjectPath'");
    }

    #get alias, alias have to be added to attributes of the added object
    my $aPath = SplitObjectPath($ObjectPath);
    my @Path = @$aPath;
    my $Alias = $Path[(scalar @Path) - 1 ];

    if (defined $hAttributes->{'Class'}){
        $Class = $hAttributes->{'Class'};
        $TagName = $hAttributes->{'Class'};
        delete $hAttributes->{'Class'};
    }

    $self->_registerPath($QuotetObjectPath);

    #transform relative path to base object
    if (defined $hAttributes->{'Base'}){
        $hAttributes->
            {'Base'} =
            $self->_getPathByRelativePath(
                $hAttributes->{'Base'},
                $self->_actualPath()
            );
    }

    #all object's addressed by path in Translation.*.xx.xml have to referenced
    $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
                $self->_actualPath(),
                'reference',
                '1'
    ) if ($self->{'XML-Type'} eq 'epages_translation');

    #register id
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'Alias' => $Alias}
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,$Class,$hAttributes);
}

#========================================================================================
# function     addSearchDefinitionToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addSearchDefinitionToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add search definition tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addSearchDefinitionToStructure{
    my $self = shift;
    my ($TagName,$hAttributes) = @_;
    my $Class=$TagName;

    #create fake Alias
    my $FakeAlias = 'SearchDefinition{DBDriver->' . $hAttributes->{'DBDriver'} . ';SearchName->' . $hAttributes->{'SearchName'} . '}';

    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();

    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'DBDriver' => $hAttributes->{'DBDriver'}, 'SearchName' => $hAttributes->{'SearchName'}}
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,$Class,$hAttributes);
}

#========================================================================================
# function     addMenuToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addMenuToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add menu tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addMenuToStructure{
    my $self = shift;
    my ($TagName,$hAttributes) = @_;
    my $Class=$TagName;

    if (!exists $hAttributes->{'Name'}){
        $hAttributes->{'Name'} = $hAttributes->{'Template'};
    }

    #create fake Alias
    my $FakeAlias = 'Menu{Name->' . $hAttributes->{'Name'} . '}';

    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();

    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );


    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'Name' => $hAttributes->{'Name'}}
    );

    if (defined $hAttributes->{'URLAction'} and not $hAttributes->{'URLAction'} =~ m{\/Classes\/}x){
        #URLAction have to be transformed
        my $ActionClass = $hAttributes->{'Class'};

        #get class name of object where tag is defined
        unless($ActionClass){
            $ActionClass = $ObjectPath;
            $ActionClass =~ s{^.*Classes\/(\w+)\/.*$}{$1}x;
        }

        my $URLAction = $hAttributes->{'URLAction'};
        $URLAction = '/Classes/' . $ActionClass . '/Actions/' . $URLAction;
        $hAttributes->{'URLAction'} = $URLAction;
    }

    delete $hAttributes->{'Class'};

    #fill xml structure
    $self->_finishXMLStructElement($TagName,$Class,$hAttributes);
}

#========================================================================================
# function     addTemplateToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addTemplateToStructure($TagName, $hAttributes);
#----------------------------------------------------------------------------------------
# description  add template tag to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addTemplateToStructure{
    my $self = shift;
    my ($TagName,$hAttributes) = @_;
    my $Class=$TagName;

    if (not defined $hAttributes->{'Cartridge'} and ($self->topID('_XMLStructStack-CartridgeTags') ne '')){
        #take cartridge name from overlaying parent cartridge tag (cartridge stack)
        $hAttributes->{'Cartridge'} = $self->topID('_XMLStructStack-CartridgeTags');
    }

    #create fake Alias
    my $FakeAlias = 'Template{Name->' . $hAttributes->{'Name'} . '}';

    my $TopElement = $self->_actualPath();
    $TopElement='' unless defined $self->_actualPath();

    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {'Name' => $hAttributes->{'Name'}}
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName,$Class,$hAttributes);
}

#========================================================================================
# function     addPermissionToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addPermissionToStructure($TagName, $hAttributes);
# example      $self->addPermissionToStructure('Permission', $hAttributes);
#----------------------------------------------------------------------------------------
# description  add permission to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addPermissionToStructure {
    my $self = shift;
    my ($TagName, $hAttributes) = @_;

    if (
        (not exists $hAttributes->{'Trustee'}) or
        (not exists $hAttributes->{'MetaAction'})
    ){
        #get file name
        my $FileName = $self->{'FileName'};

        #tag is obsolete
        die "\nobsolete tag '<Permission>' in file $FileName\n" .
            'use correct notation <Permission Trustee=\'/Groups/Everyone\' MetaAction=\'/Classes/System/Roles/Everyone\' delete=\'1\' />';
    }

    # build the absolute path
    my $Trustee = $self->_getPathByRelativePath( $hAttributes->{'Trustee'} );
    my $MetaAction = $self->_getPathByRelativePath( $hAttributes->{'MetaAction'} );

    #create fake Alias
    my $FakeAlias = $TagName.'{Trustee->'.$Trustee.',MetaAction->'.$MetaAction.'}';
    my $TopElement = $self->_actualPath() // '';
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID( $ObjectPath, {
        'Trustee' => $Trustee,
        'MetaAction' => $MetaAction,
    });

    #fill xml structure
    $self->_finishXMLStructElement($TagName, $TagName, $hAttributes);
}

sub addAttributePermissionToStructure {
    my $self = shift;
    my ($TagName, $hAttributes) = @_;

    #create fake Alias
    my $FakeAlias = 'AttributePermission{' .
        'Class->'       . $hAttributes->{'Class'} .
        ';Attribute->'  . $hAttributes->{'Attribute'} .
        ';Permission->' . $hAttributes->{'Permission'} .
    '}';

    my $TopElement = $self->_actualPath();
    delete $hAttributes->{'Path'};
    $TopElement = '' unless defined $self->_actualPath();
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);

    #register ObjectPath
    $self->_registerPath($ObjectPath);

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {
            'Class'      => $hAttributes->{'Class'},
            'Attribute'  => $hAttributes->{'Attribute'},
            'Permission' => $hAttributes->{'Permission'},
        }
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName, 'AttributePermission', $hAttributes);
}

#========================================================================================
# function     addObjectReferenceToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addObjectReferenceToStructure($TagName, $hAttributes);
# example      $self->addObjectReferenceToStructure('EnableShipping', 'ShippingMethod', $hAttributes);
#----------------------------------------------------------------------------------------
# description  Adds an element where the key is an object to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $KeyAttribute | name of the key attribute, the value is an
#                               absolute or relative object path | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addObjectReferenceToStructure {
    my $self = shift;
    my ($TagName, $KeyAttribute, $hAttributes) = @_;

    #create fake Alias
    my $KeyValue = $hAttributes->{$KeyAttribute};
    unless( defined $KeyValue ) {
        my $Keys = join(', ', sort keys %$hAttributes);
        die "Missing value for attribute $KeyAttribute on tag $TagName (keys: $Keys)";
    }

    # build the absolute path
    $KeyValue = $self->_getPathByRelativePath( $KeyValue );

    #create fake Alias
    my $FakeAlias = $TagName.'{'.$KeyAttribute.'->'.$KeyValue.'}';
    my $TopElement = $self->_actualPath() // '';
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID( $ObjectPath, {
        $KeyAttribute => $KeyValue,
    });

    #fill xml structure
    $self->_finishXMLStructElement($TagName, $TagName, $hAttributes);
}

#========================================================================================
# function     addElementToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addElementToStructure($TagName, $KeyAttribute, $hAttributes);
# example      $self->addElementToStructure('ShopCountry', 'CountryID', $hAttributes);
#----------------------------------------------------------------------------------------
# description  add a simple element to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $KeyAttribute | name of the key attribute | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addElementToStructure {
    my $self = shift;
    my ($TagName, $KeyAttribute, $hAttributes) = @_;

    #create fake Alias
    my $KeyValue = $hAttributes->{$KeyAttribute};
    unless( defined $KeyValue ) {
        my $Keys = join(', ', sort keys %$hAttributes);
        die "Missing value for attribute $KeyAttribute on tag $TagName (keys: $Keys)";
    }

    my $FakeAlias = $TagName.'{'.$KeyAttribute.'->'.$KeyValue.'}';
    my $TopElement = $self->_actualPath() // '';
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID(
        $self->_actualPath(),
        {
            $KeyAttribute => $KeyValue
        }
    );

    #fill xml structure
    $self->_finishXMLStructElement($TagName, $TagName, $hAttributes);
}

#========================================================================================
# function     addPatchCartridgeToStructure
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->addPatchCartridgeToStructure($TagName, $hAttributes);
# example      $self->addPatchCartridgeToStructure('PatchCartridge', $hAttributes);
#----------------------------------------------------------------------------------------
# description  add a patch cartridge element to xml structure
#----------------------------------------------------------------------------------------
# input        $TagName | name of tag | string
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub addPatchCartridgeToStructure {
    my $self = shift;
    my ($TagName, $hAttributes) = @_;

    my $FakeAlias = $TagName.'{OldRelease->'.$hAttributes->{'OldRelease'}.'}';
    my $TopElement = $self->_actualPath() // '';
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID( $ObjectPath, {
        'OldRelease' => $hAttributes->{'OldRelease'}
    });

    #fill xml structure
    $self->_finishXMLStructElement($TagName, $TagName, $hAttributes);
}

sub addPatchComponentToStructure {
    my $self = shift;
    my ($TagName, $hAttributes) = @_;

    my $FakeAlias = $TagName . '{' . $TagName . '}';
    my $TopElement = $self->_actualPath() // '';
    my $ObjectPath = $TopElement . '/' . $self->_quoteAlias($FakeAlias);
    $self->_registerPath($ObjectPath );

    #register ID
    $self->{'_XMLStruct-XMLStructure'}->setID( $ObjectPath, {
        '_XMLStruct-ExportIgnore' => $TagName
    });

    #fill xml structure
    $self->_finishXMLStructElement($TagName, $TagName, {});
}

sub addFileToStructure {
    my $self = shift;
    my ($hAttributes) = @_;

    $self->{'_XMLStructPlaceholder-FileAttributes'} = $hAttributes;
    #collect characters between tag
    $self->collectCharacters(1);
}

sub _registerPath{
    my $self = shift;
    my ($ObjectPath) = @_;

    #check if multiple appearence of tag should be distinguished
    if ($self->{'distinguish_multiple_appearence'}){

        #check if tag is called first time
        if (not $self->{'_XMLStruct-XMLStructure'}->existsElement($ObjectPath)){
            #first call, object path can registered
            $self->pushID('_XMLStructStack-ObjectPaths', $ObjectPath);
        }
        else{
            #tag was called one or many times before, check for number of appearence before, build new object path
            my $Counter=1;
            my $MultipleAppearenceObjectPath = $ObjectPath . ':appearance(';
            do{
                $Counter++;
            }
            while ($self->{'_XMLStruct-XMLStructure'}->existsElement($MultipleAppearenceObjectPath . $Counter));
            $self->pushID('_XMLStructStack-ObjectPaths', $MultipleAppearenceObjectPath . $Counter . ')');
        }
    }
    else{
        #normal processing multiple tags where count as one element in the structure
        $self->pushID('_XMLStructStack-ObjectPaths', $ObjectPath);
    }
}

sub _actualPath{
    my $self = shift;

    return $self->topID('_XMLStructStack-ObjectPaths');
}
#========================================================================================
# function     _finishXMLStructElement
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_finishXMLStructElement($TagName, $Class, $hAttributes);
# example      $self->_finishXMLStructElement('Object', 'TemplateTypes', {'Alias'=> 'TemplateTypes', 'reference'=> '1'});
#----------------------------------------------------------------------------------------
# description  Helper function, fill the attributes of actual element in stack, also the class and tagname
#               Additionally it registers the position of imported order of the element.
#----------------------------------------------------------------------------------------
# input        $TagName | tag name | string
# input        $Class | class object | object
# input        $hAttributes | hash with 'attribute name' => 'value' pairs | ref.hash
#========================================================================================
sub _finishXMLStructElement{
    my $self=shift;
    my ($TagName,$Class,$hAttributes) = @_;

    #fill all attributes into xml-structure
    foreach my $Key (keys %$hAttributes){
        #dont set reference if delete is already settet, this occures when a defined object is referenced in same file
        next if ($Key eq 'reference' and $self->{'_XMLStruct-XMLStructure'}->existsTagAttribute(
            $self->_actualPath(),'delete')
        );
        $self->{'_XMLStruct-XMLStructure'}->setTagAttribute(
            $self->_actualPath(),
            $Key,
            $hAttributes->{$Key}
        );
    }

    #register class/fake class
    $self->{'_XMLStruct-XMLStructure'}->setXMLStructAttribute(
        $self->_actualPath(),
        'Class',
        $Class
    );

    #register TagName
    $self->{'_XMLStruct-XMLStructure'}->setXMLStructAttribute(
        $self->_actualPath(),
        'TagName',
        $TagName
    );

    #register position of tag via import order
    $self->{'_XMLStruct-XMLStructure'}->setXMLStructAttribute(
        $self->_actualPath(),
        'Position',
        $self->{'PositionCount'}
    );
}

#========================================================================================
# function     _getPathByRelativePath
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ActualPath = $self->_getPathByRelativePath($RelativePath, $TempObjectPath);
# example      $ActualPath = $self->_getPathByRelativePath('../../SubObject2Level1','SuperObject/SubObject1Level1/SubObjectLevel2');
#----------------------------------------------------------------------------------------
# description  Generates the objectpath by a relative and absolute path.
#               Needs e.g. when an object in specific tag adressed by path
#               or relative path and not by alias.
#----------------------------------------------------------------------------------------
# input        $RelativePath | relative path to an object | string
# input        $TempObjectPath | temporary path to an object | string
# return       $ActualPath | Actual path of object that adressed by a relative path | string
#========================================================================================
sub _getPathByRelativePath{
    my $self = shift;
    my ($RelativePath, $TempObjectPath) = @_;

    if ($RelativePath eq '.'){
        return $self->_actualPath();
    }
    elsif ($RelativePath =~ /\.\.(\/|\\)/x){
        my $aActualPath;
        my $aRelativePath = SplitObjectPath($RelativePath);

        if(defined $TempObjectPath){
            $aActualPath = SplitObjectPath($TempObjectPath);
        }
        else{
            $aActualPath = SplitObjectPath($self->_actualPath());
        }

        my @RelativeTempPath = @$aRelativePath;

        my $RelativePosition = 0;
        for (my $i=$RelativePosition; $i<scalar @$aRelativePath; $i++){
            if ($RelativeTempPath[$i] eq '..'){
                pop(@$aActualPath);
            }
            else {
                push(@$aActualPath, $RelativeTempPath[$i]);
            }
        }
        return JoinObjectPath($aActualPath);
    }
    else {
        return $RelativePath;
    }
}

sub _quoteAlias{
    my $self = shift;
    my ($Alias) = @_;

    if (
        ($Alias =~ /\\/                 #slash in alias
        or $Alias =~ /\//               #backslash in alias
        or $Alias =~ /\s/)              #alias contains spaces
        and not $Alias =~ /^\".+\"$/    #alias is not already quoted "..."
    ){
        my $QuoteAlias = '"' . $Alias . '"';
        return $QuoteAlias;
    }
    else{
        return $Alias;
    }
}
1;
