# perl -W
use strict;
use Test::More tests => 22;
use DE_EPAGES::Core::API::Constants qw( PERL538 );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Object::API::Factory qw( ExistsObjectByPath LoadObjectByPath LoadRootObject
                                        LoadClassByAlias ExistsClassByAlias );
use DE_EPAGES::Database::API::Connection qw( GetTestSite RunOnStore );

my $ATTRIBUTE_PACKAGE = 'DE_EPAGES::Object::API::Attributes::JSONAttribute';
my $ROOTPATH = '/Test';

##no critic qw(ProhibitInterpolationOfLiterals)

sub Main {

    my $TestObject = LoadObjectByPath( $ROOTPATH );
    my $Object = LoadClassByAlias('C1')->insertObject({
        Parent => $TestObject,
        Alias => 'C1O1'
    } );

    is_deeply( $Object->get('GoogleMap'), undef, 'no value set yet' );

    my $SollData = {
        'center' => '12345',
        'height' => '220',
        'width'  => '200',
        'key'    => {
            'subkey' => [
                'bla',
                'bli',
                'blubb',
            ]
        },
        'null'   => 'null',
        'undef'  => undef,
        'empty'  => '',
        'float'  => 123.456,
    };

    $Object->set({'GoogleMap'=> $SollData});

    my $IsData = $Object->get('GoogleMap');
    foreach my $Key ('center','height','widht','null','undef','empty','float') {
        is($IsData->{$Key}, $SollData->{$Key}, "key: $Key");
    }

    my $IsValue = $IsData->{'key'}{'subkey'}[2];
    my $SollValue = $SollData->{'key'}{'subkey'}[2];
    is($IsValue, $SollValue, "Data->{key}{subkey}[2]: $IsValue eq $SollValue");

    $Object->set({ 'GoogleMap' => ['a','b','c'] });
    is( $Object->get('GoogleMap')->[2], 'c', 'array ref' );

    $Object->set({'GoogleMap' => undef});
    is( $Object->get('GoogleMap'), undef, 'undef' );

    if ($] < PERL538) {
        eval { $Object->set({'GoogleMap' => '[string]'}); };
        like( $@, qr/hash- or arrayref expected/, 'string' );
    } else {
        $Object->set({'GoogleMap' => '[string]'});
        is( $Object->get('GoogleMap'), '[string]', 'string' );
    }

    # DefaultValue
    is_deeply( $Object->get('J1'), { Test => "TestValue", Undef => undef, Int => 1 }, "DefaultValue is hash" );
    is_deeply( $Object->get('J2'), ["A", "B", "C"], "DefaultValue is array" );
    is( $Object->get('J3'), undef, "DefaultValue is undef" );

    $Object->set({ 'J1' => [1,2,3] });
    is_deeply( $Object->get('J1'), [1,2,3], "set array" );
    $Object->set({ 'J1' => undef });
    is_deeply( $Object->get('J1'), { Test => "TestValue", Undef => undef, Int => 1 }, "set to undef restores the default value" );
    $Object->set({ 'J1' => [1,2,3] });
    $Object->set({ 'J2' => { A => 1, B => 2 } });
    is_deeply( $Object->get('J2'), { A => 1, B => 2 }, "set hash" );

    my $Clone = $Object->clone( {
        Alias => 'Clone',
        J3 => [4,5]
    });

    is_deeply( $Clone->get('J1'), [1,2,3], "clone array" );
    is_deeply( $Clone->get('J2'), { A => 1, B => 2 }, "clone hash" );
    is_deeply( $Clone->get('J3'), [4,5], "clone, set array" );
    is( $Clone->get('J4'), undef, "clone undef" );
}

sub Cleanup {
    if( ExistsClassByAlias( 'C1' ) ) {
        LoadClassByAlias( 'C1' )->delete;
    }
    if( ExistsObjectByPath( $ROOTPATH ) ) {
        LoadObjectByPath( $ROOTPATH )->delete;
    }
    return 1;
}

sub Init {
    Cleanup();
    my $ClassObject = LoadClassByAlias( 'Object' );
    my $Class1 = $ClassObject->insertSubClass({ Alias => 'C1' });
    $Class1->insertAttribute({ Alias => 'GoogleMap', Type => 'Hash', Package => $ATTRIBUTE_PACKAGE });
    $Class1->insertAttribute({ Alias => 'J1', Type => 'JSON', IsExportable=> 1, ExportLevel => 1,
        Package => $ATTRIBUTE_PACKAGE,
        DefaultValue => '{"Test":"TestValue","Undef":null,"Int":1}',
    });
    $Class1->insertAttribute({ Alias => 'J2', Type => 'JSON', IsExportable=> 1, ExportLevel => 1,
        Package => $ATTRIBUTE_PACKAGE,
        DefaultValue => '["A", "B", "C"]',
    });
    $Class1->insertAttribute({ Alias => 'J3', Type => 'JSON', IsExportable=> 1, ExportLevel => 1,
        Package => $ATTRIBUTE_PACKAGE,
    });
    $Class1->insertAttribute({ Alias => 'J4', Type => 'JSON', IsExportable=> 1, ExportLevel => 1,
        Package => $ATTRIBUTE_PACKAGE,
    });

    $ClassObject->insertObject( { Alias => 'Test', Parent => LoadRootObject() } );
    return 1;
}

RunScript( Debug => 1, Sub => sub {
    RunOnStore( Store => GetTestSite(), Sub => sub {
        Init();
        Main();
        Cleanup();
    });
});
