#========================================================================================
# package      DE_EPAGES::Presentation::API::Session
# state        public
#----------------------------------------------------------------------------------------
# description  This package provides the interface for sessions, session ids and session
#               attributes.
#               The TLE variable #Session implemented by
#               <package DE_EPAGES::Presentation::API::TLE::SessionHandler> provides
#               access to session attributes from the TLE language. Note that #Session
#               is undefined if the user is not logged on and does not have a basket yet.
#               The following list shows all pre-defined session attributes.
#               <ul>
#                 <li>User - user identified by the session (class:User, can be undefind) - object</li>
#                 <li>UserID - id of the user identified by the session (can be undefind) - integer</li>
#                 <li>IsClosed - true if the session is no longer valid, either because
#                               the user has signed our or because the session has timed
#                               out - boolean</li>
#                 <li>IsCookieIdentified - true if the session was associated with the user
#                               based on a cookie. For security reasons it may be necessary
#                               to request confirmation of the users identity by asking for
#                               the password - integer</li>
#                 <li>CreationDate - date and time when the session has started - datetime</li>
#                 <li>RemoteIP - IP address of the client - string</li>
#                 <li>RemoteHost - host name or IP address of the client (can be undefined)
#                               - string</li>
#                 <li>RemoteUser - user name if the user was identified via HTTP
#                               authentication (can be undefind) - string</li>
#                 <li>RefererURL - URL of the previously visited page before the session
#                               has started - string</li>
#               </ul>
#               Additional session attributes can be created with the function
#               <function SetSessionAttribute>.
#========================================================================================
package DE_EPAGES::Presentation::API::Session;
use base qw( Exporter );

use strict;
our @EXPORT_OK = qw(
    GenerateSessionID
    InsertSession

    ExistsSessionAttribute
    GetSessionAttribute
    SetSessionAttribute
    GetSessionAttributes

    IsSessionTimedOut
    GetSessionAge

    TestAdminSFSession
    CreateAdminSFSessionID
    GetAdminSFSessionCookie

);

use if $] < 5.038, 'DES::EDE3CBC'; ##no strict qw(ProhibitMagicNumbers)
use if $] >= 5.038, 'Crypt::CBC'; ##no strict qw(ProhibitMagicNumbers)

use DE_EPAGES::Core::API::Constants qw( PERL538 );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::PerlTools qw (GetIPAddress GetMD5Digest Ip2Hex Str2Hex Hex2Str);
use DE_EPAGES::Core::API::TestParameter qw ( TestInteger TestIpAddress );
use DE_EPAGES::Database::API::Connection qw (  GetCurrentDBHandle );
use DE_EPAGES::Object::API::Factory qw (LoadObject LoadRootObject);
use DE_EPAGES::Presentation::API::Table::Session qw (UpdateSession GetInfoSession ExistsSessionBySessionID GetInfoSessionBySessionID);
use DE_EPAGES::Presentation::API::Table::SessionAttribute qw ( GetPKeySessionAttributesByName InsertSessionAttribute GetAllInfoSessionAttributes);
use DE_EPAGES::Presentation::API::Table::SessionParameter qw (
    ExistsSessionParameter
    GetInfoSessionParameter
    UpdateSessionParameter
    InsertSessionParameter
    DeleteSessionParameter
);
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook);
use DE_EPAGES::WebInterface::API::ApplicationServer qw( GetAppServer );
use DE_EPAGES::WebInterface::API::Cookie qw( GetCookies );

my $COUNTER = 0;
my %DIRECT_SESSION_ATTRIBUTES = map { $_ => undef }
    qw (User IsClosed IsCookieIdentified CreationDate RemoteIP RemoteHost RemoteUser RefererURL LastActivity);
use constant MAX_REFERRER_LENGTH => 255;
use constant DES_INIT_VECTOR     => 'x.4ZeBj!';

#========================================================================================
# function     GenerateSessionID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $SessionID = GenerateSessionID()
# syntax       $SessionID = GenerateSessionID( $Address, $Port )
# example      $SessionID = GenerateSessionID( '12.34.56.78', 10047 )
#----------------------------------------------------------------------------------------
# description  Generates a new, unique session id. The session id is not automatically
#               registered in the database. See also <function InsertSession>.
#----------------------------------------------------------------------------------------
# input        $Address | (optional) application server IP address or '0.0.0.0' | string
# input        $Port    | (optional) application server port | int
# return       $SessionID | a new session id | string
#========================================================================================
sub GenerateSessionID (;$$) {
    my ($Address, $Port) = @_;
    if (not defined $Address) {
        my $AppServer = GetAppServer();
        if ($AppServer) {
            $Address = GetIPAddress($AppServer->address);
            $Port    = $AppServer->port;
        } else {  # script
            $Address = GetIPAddress();
            $Port    = 0;
        }
    }
    $Port = 0 unless defined $Port;
    TestInteger( 'Port', $Port );
    TestIpAddress( 'Address', $Address );

    ## no critic qw(ProhibitMagicNumbers)
    my $Time = time() & 0xFFFFFFFF;
    my $Rand = int(rand(65536));
    $COUNTER = ( $COUNTER + 1 ) & 0xFFFF;
    ## use critic qw(ProhibitMagicNumbers)

    my $SessionID = sprintf '%08X%04X%04X%04X%s',
        $Time,                  #--- byte 0, 1, 2, 3: current time
        $COUNTER,               #--- byte 4, 5: counter
        $Rand,                  #--- byte 6, 7: random value
        $Port,                  #--- byte 8, 9: port number
        Ip2Hex($Address,8); #--- byte 10, 11, 12, 13: ip address

    # add checksum
    my $CheckSum = 0;
    $CheckSum += unpack '%16C*', $_ foreach unpack( 'C*', pack( 'H*', $SessionID ));
    $SessionID .= sprintf '%04X', $CheckSum;

    # add 100 random chars
    my @VECT = ('A'..'Z','a'..'z','0'..'9');
    $SessionID .= $VECT[int(rand(62))] foreach (1..100); ## no critic qw(ProhibitMagicNumbers)

    # build md5 hash
    return GetMD5Digest(\$SessionID);
}

#========================================================================================
# function     _GenerateToken
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Token = _GenerateToken();
#----------------------------------------------------------------------------------------
# return       $Token | a new security token | string
#----------------------------------------------------------------------------------------
# description  Generates a new random number, that is not guaranteed to be unique.
#               The number is hexadecimal with 10 digits.
#========================================================================================
sub _GenerateToken() {

    ## no critic qw(ProhibitMagicNumbers)
    return sprintf '%04x%04x%02x', int(rand(65536)), int(rand(65536)), int(rand(256));
    ## use critic qw(ProhibitMagicNumbers)
}

#========================================================================================
# function     InsertSession
# state        public
#----------------------------------------------------------------------------------------
# syntax       $SessionID = InsertSession($User, $PerCookie);
#----------------------------------------------------------------------------------------
# description  Generates a new session id for a user and saves it in the database.
#               If you have access to a servlet object, use
#               $Servlet->registerSession($User, $PerCookie) instead of InsertSession().
#               See: <function DE_EPAGES::Presentation::Servlet::Permission::registerSession>
#----------------------------------------------------------------------------------------
# input        $User          | user                      (optional)| object
# input        $PerCookie     | user identified by cookie (optional)| bit
# return       $SessionID     | a new session id | string
#========================================================================================
sub InsertSession ($$) {
    my ($User, $PerCookie) = @_;

    my $SessionID = GenerateSessionID();
    my $serverVariables = GetAppServer->serverVariables;
    my $RefererURL = $serverVariables->{'HTTP_REFERER'};
    $RefererURL = substr($RefererURL, 0, MAX_REFERRER_LENGTH) if defined $RefererURL && length $RefererURL > MAX_REFERRER_LENGTH;
    DE_EPAGES::Presentation::API::Table::Session::InsertSession({
        'SessionID' => $SessionID,
        'UserID'    => defined $User ? $User->id : undef,
        'IsClosed'  => 0,
        'IsCookieIdentified' => $PerCookie ? 1 : 0,
        'RemoteIP'    => $serverVariables->{'REMOTE_ADDR'},
        'RemoteHost'  => $serverVariables->{'REMOTE_HOST'},
        'RemoteUser'  => $serverVariables->{'REMOTE_USER'},
        'RefererURL'  => $RefererURL,
        'Token'       => _GenerateToken(),
        'LastActivity'=> GetCurrentDBHandle()->currentDateTime,
    });
    return $SessionID;
}

#========================================================================================
# function     ExistsSessionAttribute
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Exists = ExistsSessionAttribute($Name);
#----------------------------------------------------------------------------------------
# description  Returns true if a session attribute exists.
#----------------------------------------------------------------------------------------
# input        $Name | name of session attribute | string
# return       $Exists | exists attribute | boolean
#========================================================================================
sub ExistsSessionAttribute ($) {
    my ($Name) = @_;
    return 1 if exists $DIRECT_SESSION_ATTRIBUTES{$Name};
    my $AttributeIDs = GetPKeySessionAttributesByName($Name);
    return scalar @$AttributeIDs;
}

#========================================================================================
# function     _SetSessionAttribute
# state        private
#----------------------------------------------------------------------------------------
# syntax       $AttributeID = _SetSessionAttribute($Name);
#----------------------------------------------------------------------------------------
# input        $Name    | name of session attribute | string
# return       $AttributeID | attribute id  | int
#----------------------------------------------------------------------------------------
# description  Gets and if necessary creates session attribute.
#========================================================================================
sub _SetSessionAttribute ($) {
    my ($Name) = @_;
    my $AttributeIDs = GetPKeySessionAttributesByName($Name);
    return $AttributeIDs->[0] if scalar @$AttributeIDs;
    GetLog->info(sprintf('session attribute %s inserted.', $Name));
    return InsertSessionAttribute({'Name' => $Name});
}

#========================================================================================
# function     GetSessionAttribute
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Value = GetSessionAttribute( $IntSessionID, $Name )
#----------------------------------------------------------------------------------------
# description  Returns a session attribute value.
#----------------------------------------------------------------------------------------
# input        $IntSessionID | internal session id | int
# input        $Name    | name of session attribute | string
# return       $Value | session parameter | string
#========================================================================================
sub GetSessionAttribute ($$) {
    my ($IntSessionID, $Name) = @_;

    if ($Name eq 'User') {
        my $Info = GetInfoSession($IntSessionID);
        return undef unless defined $Info->{'UserID'};
        return LoadObject($Info->{'UserID'});
    }
    elsif ( exists $DIRECT_SESSION_ATTRIBUTES{$Name} ) {
        return GetInfoSession($IntSessionID)->{$Name};
    }

    my $AttributeID = _SetSessionAttribute($Name);
    return ExistsSessionParameter($IntSessionID, $AttributeID)
        ? GetInfoSessionParameter($IntSessionID, $AttributeID)->{'Value'}
        : undef;
}

#========================================================================================
# function     GetSessionAttributes
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aValues = GetSessionAttributes()
#----------------------------------------------------------------------------------------
# description  Returns all session attribute values.
#----------------------------------------------------------------------------------------
# return       $aValues | session attributes values | reference to array
#========================================================================================
sub GetSessionAttributes () {

    my @Names = keys(%DIRECT_SESSION_ATTRIBUTES), map {$_->{'Name'}} @{GetAllInfoSessionAttributes()};
    return \@Names;
}

#========================================================================================
# function     SetSessionAttribute
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetSessionAttribute( $IntSessionID, $Name, $Value )
#----------------------------------------------------------------------------------------
# description  Sets a session attribute value.
#----------------------------------------------------------------------------------------
# input        $IntSessionID | internal session id | int
# input        $Name    | name of session attribute | string
# input        $Value | session parameter | string
#========================================================================================
sub SetSessionAttribute ($$$){
    my ($IntSessionID, $Name, $Value) = @_;

    if ($Name eq 'User') {
        UpdateSession($IntSessionID, {'UserID'=> defined $Value ? $Value->id : undef});
        return;
    }
    elsif ( exists $DIRECT_SESSION_ATTRIBUTES{$Name} ) {
        UpdateSession($IntSessionID, {$Name => $Value});
        return;
    }

    my $AttributeID = _SetSessionAttribute($Name);
    if( ExistsSessionParameter($IntSessionID, $AttributeID) ) {
        if( defined $Value ) {
            UpdateSessionParameter($IntSessionID, $AttributeID, {'Value'=>$Value});
        }
        else {
            DeleteSessionParameter($IntSessionID, $AttributeID);
        }
    }
    elsif( defined $Value ) {
        InsertSessionParameter({
            'IntSessionID'       => $IntSessionID,
            'SessionAttributeID' => $AttributeID,
            'Value'              => $Value
        });
    }

    return;
}

#========================================================================================
# function     GetSessionAge
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetSessionAge($IntSessionID);
#----------------------------------------------------------------------------------------
# description  Get the number of minutes since creation of the session.
#----------------------------------------------------------------------------------------
# input        $IntSessionID | internal session id | int
# return       $minutes | age of the session in minutes | integer
#========================================================================================
sub GetSessionAge {
    my ($IntSessionID) = @_;

    return _GetSessionAge($IntSessionID, GetCurrentDBHandle()->currentDateTime);
}

# like GetSessionAge but with parameters that make testing easier
sub _GetSessionAge {
    my ($IntSessionID, $Now) = @_;

    my $CreationDate = GetInfoSession($IntSessionID)->{'CreationDate'};
    return _DiffMinutes($CreationDate, $Now);
}

#========================================================================================
# function     IsSessionTimedOut
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsTimeout = IsSessionTimedOut($IntSessionID)
# example      if (IsSessionTimedOut($IntSessionID)) { ... }
#----------------------------------------------------------------------------------------
# description  Check if a session has timed out, that is, at least as many minutes as
#               the session timeout have gone since session creation.
#----------------------------------------------------------------------------------------
# input        $IntSessionID | internal session id | int
# return       $IsTimeout | true if the session has timed out | boolean
#========================================================================================
sub IsSessionTimedOut {
    my ($IntSessionID) = @_;
    TestInteger( 'IntSessionID', $IntSessionID );

    my $hTimeouts = LoadRootObject()->get(['SessionTimeout', 'SessionIdleTimeout']);
    my $Now = GetCurrentDBHandle()->currentDateTime;
    return _IsSessionTimedOut($IntSessionID, $hTimeouts, $Now);
}

# like IsSessionTimedOut but with parameters that make testing easier
sub _IsSessionTimedOut {
    my ($IntSessionID, $hTimeouts, $Now) = @_;

    my ($SessionTimeout, $IdleTimeout) = @{$hTimeouts}{'SessionTimeout', 'SessionIdleTimeout'};

    return 0 unless (defined $SessionTimeout || defined $IdleTimeout); # not specified means never expires

    my ($CreationDate, $LastActivity) = @{GetInfoSession($IntSessionID)}{'CreationDate', 'LastActivity'};
    if (defined $SessionTimeout) {
        my $SessionAge = _DiffMinutes($CreationDate, $Now);
        return 1 if $SessionAge >= $SessionTimeout;
    }

    if (defined $IdleTimeout) {
        my $IdleTime = _DiffMinutes($LastActivity, $Now);
        return 1 if $IdleTime >= $IdleTimeout;

        # reset inactivity period once a minute
        SetSessionAttribute($IntSessionID, 'LastActivity', $Now) if $IdleTime > 0;
    }

    return 0;
}

sub _DiffMinutes {
    my ($DateTime1, $DateTime2) = @_;
    my $Duration = $DateTime2->delta_ms($DateTime1);
    return 60 * $Duration->hours + $Duration->minutes; ## no critic qw(ProhibitMagicNumbers)
}

#========================================================================================
# function     GetAdminSFSessionCookie
# state        public
#----------------------------------------------------------------------------------------
# syntax       $AdminSessionCookie = GetAdminSFSessionCookie()
#----------------------------------------------------------------------------------------
# description  Gets the cookie value for cookie MerchantSF from the current request
#----------------------------------------------------------------------------------------
# return       $AdminSessionCookie | cookie value | string
#========================================================================================
sub GetAdminSFSessionCookie {

    my $AppServer = GetAppServer();
    return undef unless defined $AppServer;
    my $Servlet = $AppServer->currentServlet;
    my $Request = $Servlet->request;
    my $AdminSessionCookie = GetCookies($Request)->{'MerchantSF'};

    return $AdminSessionCookie;
}

#========================================================================================
# function     TestAdminSFSession
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ExistsValidAdminSession = TestAdminSFSession($Site,$Cookie)
# example      if (TestAdminSFSession($Site,$Cookie) == -1) { ... }
#----------------------------------------------------------------------------------------
# description  Check if a valid admin session exists for usage in the closed storefront.
#----------------------------------------------------------------------------------------
# input        $Site | shop object | object
# input        $Cookie | cookie value | string
# return       $ExistsValidAdminSession | 1 if admin session for sf exists and all is valid,
#                                          0 if is not a admin sf session (not ClosedByMerchant ),
#                                         -1 if session is invalid (ClosedByMerchant and e.g. no cookie)
#                                        | integer
#========================================================================================
sub TestAdminSFSession {
    my ($Site, $Cookie) = @_;

    # test if site is closed by merchant
    # if so, we do not have a admin sf request
    my $hParams = {'Site'=>$Site};
    TriggerHook('API_TestAdminSFSession', $hParams);
    return 0 unless $hParams->{'ClosedByMerchant'};

    # we do not have a 'MerchantSF' cookie
    return -1 unless $Cookie;

    my $SiteGUID = $Site->get('GUID');
    my $CryptedStr = Hex2Str($Cookie);
    my $DeCryptedAdminSessionID = _DeCryptDESString($CryptedStr,$SiteGUID);

    return -1 unless $DeCryptedAdminSessionID;
    return -1 unless ExistsSessionBySessionID($DeCryptedAdminSessionID);

    # test if a admin session exists
    my $IntSessionID = GetInfoSessionBySessionID($DeCryptedAdminSessionID)->{'IntSessionID'};
    # is session closed?
    return -1 if GetSessionAttribute($IntSessionID,'IsClosed');
    #  test timeout
    return -1 if IsSessionTimedOut($IntSessionID);

    # test if we have a user who is merchant
    # and do a site permission check for user
    my $hVars = {
        'Site' => $Site,
        'IntSessionID' => $IntSessionID,
        'IsMerchant' => 0,
    };
    TriggerHook('API_TestAdminSFSession', $hVars);

    return -1 unless $hVars->{'IsMerchant'};

    # session is valid
    return 1;

}

#========================================================================================
# function     CreateAdminSFSessionID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $AdminSFSessionID = CreateAdminSFSessionID($Site,$SessionID)
#----------------------------------------------------------------------------------------
# description  Returns an admin session id for the usage in the closed storefront
#               or undef is session is invalid
#----------------------------------------------------------------------------------------
# input        $Site | site object | object
# input        $SessionID | session of user | string
# return       $AdminSFSessionID | des crypted admin session id or undef | string
#========================================================================================
sub CreateAdminSFSessionID{
    my($Site,$SessionID) = @_;

    return undef unless ExistsSessionBySessionID($SessionID);
    my $IntSessionID = GetInfoSessionBySessionID($SessionID)->{'IntSessionID'};

    # is session closed?
    return undef if GetSessionAttribute($IntSessionID,'IsClosed');
    # test timeout
    return undef if IsSessionTimedOut($IntSessionID);

    my $GUID = $Site->get('GUID');

    return Str2Hex(_CryptDESString($SessionID,$GUID));

}

#========================================================================================
# function     _CryptDESString
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Crypted = _CryptDESString($String,$Key);
#----------------------------------------------------------------------------------------
# description  crypts $String by using tripple DES with $Key
#----------------------------------------------------------------------------------------
# input        $String | string to crypt | string
# input        $Key | crypt key | string
# return       $Crypted | 3des crypted string  | string
#========================================================================================
sub _CryptDESString {
    my ($String,$Key) = @_;
    ## no critic qw(ProhibitMagicNumbers)
    $Key .= '0' x (8 - length($Key)) if length($Key) < 8;
    ## use critic qw(ProhibitMagicNumbers)

    # crypt value with key
    # use a fix value ivec
    my $Crypted;
    if ($] < PERL538) {
        my @DES_KEYS = ($Key, reverse($Key), $Key, DES_INIT_VECTOR);
        $Crypted = des_3ede_cbc_encrypt( $String, @DES_KEYS, 1 );
    } else {
        my $DES_KEY = (unpack('(A8)*', $Key))[0];
        my $cipher = Crypt::CBC->new( -cipher => 'DES_EDE3', -key => $DES_KEY x 3,
                                      -iv => DES_INIT_VECTOR, -header => 'none',
                                      -literal_key => 1, 'padding' => 'none' );
        $Crypted = $cipher->encrypt($String);
    }

    return $Crypted;
}

#========================================================================================
# function     _DeCryptDESString
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Decrypted = _DeCryptDESString($Crypted, $Key);
#----------------------------------------------------------------------------------------
# description  decrypts $Crypted by using tripple DES with $Key
#----------------------------------------------------------------------------------------
# input        $Crypted | string to decrypt | string
# input        $Key | decrypt key | string
# return       $Decrypted | 3des decrypted string  | string
#========================================================================================
sub _DeCryptDESString {
    my ($Crypted, $Key) = @_;
    ## no critic qw(ProhibitMagicNumbers)
    $Key .= '0' x (8 - length($Key)) if length($Key) < 8;
    ## use critic qw(ProhibitMagicNumbers)

    # decrypt value with key $Key
    my $Decrypted;
    if ($] < PERL538) {
        my @DES_KEYS = ($Key, reverse($Key), $Key, DES_INIT_VECTOR);
        $Decrypted = des_3ede_cbc_encrypt( $Crypted, @DES_KEYS, 0 );
    } else {
        my $DES_KEY = (unpack('(A8)*', $Key))[0];
        my $cipher = Crypt::CBC->new( -cipher => 'DES_EDE3', -key => $DES_KEY x 3,
                                      -iv => DES_INIT_VECTOR, -header => 'none',
                                      -literal_key => 1, 'padding' => 'none' );
        $Decrypted = $cipher->decrypt($Crypted);
    }

    $Decrypted = (split /\0/, $Decrypted)[0];
    return $Decrypted;
}



1;
