use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Error qw( GetError ExistsError );
use DE_EPAGES::Core::API::Log qw( LogError );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadObjectByPath );
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw (ImportObjects);
use DE_EPAGES::Shop::API::Cartridge;
use DE_EPAGES::Shop::API::System qw( GetUnityShopIDs );
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );

#========================================================================================
# function         InitializeMenuSwitch
# state            public
#----------------------------------------------------------------------------------------
# syntax           InitializeMenuSwitch();
#----------------------------------------------------------------------------------------
# description      Activates the notification about the new menu switch for all existing Unity shops
#========================================================================================
sub InitializeMenuSwitch {
    my ($ShopAlias, $Force) = @_;

    my $Cartridge = DE_EPAGES::Shop::API::Cartridge->new;

    if (defined $ShopAlias) {
        # single shop run
        my $Shop = LoadObjectByPath('/Shops/'.$ShopAlias);
        if ($Shop->isUnity()) {
            # Unity shop
            if (!$Force && $Shop->folder('ShopNotifications')->existsChild('MenuSwitch')) {
                print 'The shop notifications for the menu switch exists already for shop ' . $ShopAlias . ".\n"
                 . "Rerun with -force flag, to import nevertheless. \n";
            } else {
                # import shop notifications
                ImportObjects($Cartridge->xmlDirectory().'/ShopNotificationMenuSwitch.xml', $Shop->pathString)
                    unless (!$Force && $Shop->folder('ShopNotifications')->existsChild('MenuSwitch'));
            }
        } else {
            # non Unity shop
           print 'Shop ' . $ShopAlias . " is not a Unity shop.\n";
        }
    } else {
        print "Reading all Unity shops\n";
        my $aShopIDs = GetUnityShopIDs();

        foreach my $ShopID (@$aShopIDs) {
            eval {
                my $Shop = LoadObject($ShopID);
                # import shop notifications
                ImportObjects($Cartridge->xmlDirectory().'/ShopNotificationMenuSwitch.xml', $Shop->pathString)
                    unless (!$Force && $Shop->folder('ShopNotifications')->existsChild('MenuSwitch'));
            };
            warn('Patch failed in shop ' . $ShopID . " with error \"" . GetError() . "\".")
                if ExistsError();
        }
    }

    eval { SynchronizeCache(); };
    LogError(GetError()) if ExistsError();

    return;
}

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $ShopAlias;
    my $Force;
    my $Help;

    GetOptions(
        'help'        => \$Help,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'force'       => \$Force,
        'shopalias=s' => \$ShopAlias,
    );

    usage() if $Help;

    unless ( defined $StoreName ) {
        print STDERR "Missing parameter -storename.\n\n";
        usage();
    }
    my $time1 = time;

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub { InitializeMenuSwitch($ShopAlias, $Force); },
    );
    my $time2 = time;
    printf "need %.1f seconds\n", $time2 - $time1;
}

sub usage {
    print <<END_USAGE;
Description:
    Activates the notification for Menu Switch for all Unity shops.

Usage:
    perl $0 [options] [flags] datafiles

options:
    -passwd     database user password                          ()
    -storename  name of store                                   ()
    -shopalias  alias of a single shop                      (optional)
                (if not given, the script is preformed for all Unity shops)
flags:
    -help       show the command line options
    -force      delete relative Url, if it exists already

Example:
    perl $0 -storename Store
END_USAGE
    exit 2;
}

RunScript( 'Sub' => \&Main );