use Test::More;
use strict;
use Text::Diff qw(diff);
use DE_EPAGES::Core::API::Constants qw( PERL538 );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::String qw ( sortCSVLine );
use DE_EPAGES::Core::API::File  qw ( GetFileContent WriteFile DeleteFile DeleteDirectory );
use DE_EPAGES::TLE::API::TLE    qw ( ReplaceTLEVars GetByteCodeByText );
use DE_EPAGES::TLE::API::BaseProcessor;
use DE_EPAGES::TLE::t::GUIObjects;
use DE_EPAGES::TLE::t::tlevars;

our $TEST_ONE;

sub Main {
    Initialize();
    if ( defined $TEST_ONE ) {
        plan tests => 1;
        Test_Package($TEST_ONE );
    }
    else {
        my @Files = glob '*_t.txt';
        plan tests => 1 + scalar @Files;

        foreach my $File ( @Files ) {
            if($File =~ /^(.+)_t\.txt$/) {
                my $TestCase = $1;
                Test_Package($TestCase );
            }
            else {
                die "invalid file name $File";
            }
        }

        Test_ReplaceTLEVars( '05_cond' );
    }
    Cleanup();
}

local $| = 1;

RunScript(
    'Sub' => \&Main
);

sub ErrorDefinition {
    return {
        NO_EQUAL_LINE => 'Output is different at line #AtLine'
    };
}

sub Test_ReplaceTLEVars {
    my ($i) = @_;

    my $rTemplate = GetFileContent( $i.'_t.txt' );
    my $byteCode = GetByteCodeByText( $rTemplate );
    my %Vars = (
        %{ DE_EPAGES::TLE::t::tlevars::GetLoops($i) },
        %{ DE_EPAGES::TLE::t::tlevars::GetVars($i) }
    );
    my $isContent = ReplaceTLEVars($byteCode, \%Vars );
    my $exContent = ${ GetFileContent($i.'_e.txt') };


    ok( $isContent eq $exContent, "ReplaceTLEVars $i" )
        or Test::More::diag( diff \$isContent, \$exContent, { STYLE => 'OldStyle' } );
}

sub Test_Package {
    my ($i) = @_;

    my $Loops =  DE_EPAGES::TLE::t::tlevars::GetLoops($i);
    my $Vars  =  DE_EPAGES::TLE::t::tlevars::GetVars($i);

    if ( $i eq '37_script_invalid' ) {
        eval {
            ProcessFileTemplate($i.'_t.txt', $Vars , $Loops);
        };
        if ( $@ ) {
            my $Match = $@ =~ m/(Can not find closing \'\<script\>\' tag \(postion 0\) at line 1)/;
            ok( $Match, "Different error expected $i" );
        } else {
            ok( 0, "Expected a NativeDie $i" );
        }
        return;
    }

    local $DE_EPAGES::TLE::API::Execute::TEMPLATE_SOURCE = $i.'_t.txt';
    my $isContent = ProcessFileTemplate($i.'_t.txt', $Vars , $Loops);
    my $Expected = "${i}_e.txt";
    $Expected = "${i}_e-5.38.0.txt" if $] >= PERL538 and -f "${i}_e-5.38.0.txt";
    my $exContent = ${GetFileContent($Expected)};
    if ($i =~ /^3[24]/) {

        $isContent = sortCSVLine($isContent, "\n");
        $exContent = sortCSVLine($exContent, "\n");
    } elsif ($i =~ /^31/) {
        $isContent = sortJson($isContent);
        $exContent = sortJson($exContent);
    }
    my $equal = $isContent eq $exContent;
    ok( $equal, $i );
    if( $equal ) {
        DeleteFile($i.'_i.txt');
    }
    else {
        Test::More::diag( diff \$isContent, \$exContent, { STYLE => 'OldStyle' } );
        WriteFile($i.'_i.txt', \$isContent);
    }

    return;
}

sub ProcessFileTemplate {
    my ( $FileName, $hVars, $hLoops ) = @_;

    my $TLEProcessor = DE_EPAGES::TLE::API::BaseProcessor->new;
    DE_EPAGES::TLE::API::PrefixHandler->new->register( $TLEProcessor );

    my $GUIHandler = $TLEProcessor->{'NamedFunctionHandler'}->{'GUI'};
    $GUIHandler->registerGUIObject('DE_EPAGES::TLE::t::GUIObjects::Accordion', 'accordion');
    $GUIHandler->registerGUIObject('DE_EPAGES::TLE::t::GUIObjects::List', 'list');
    $GUIHandler->registerGUIObject('DE_EPAGES::TLE::t::GUIObjects::ListItem', 'listitem');

    my $rTemplate = GetFileContent( $FileName );

    my $byteCode = GetByteCodeByText( $rTemplate );
    my %TempVars = (%$hVars, %$hLoops);

    return $TLEProcessor->replaceTLEs($byteCode, \%TempVars);
}

sub Initialize {
    Cleanup();
}

sub Cleanup {
    DeleteDirectory( 'ctmpl', { recursive => 1 } );
}

sub sortJson {
    my ($String) = @_;
    my ($Open, $Close);
    if ($String =~ s,^(.*{),,s) {
        $Open = $1;
    }
    if ($String =~ s,(}.*)$,,s) {
        $Close = $1;
    }
    $String =~ s/^\s*//s;
    $String =~ s/\s*$//s;
    $String =~ s/\s*,\s*/,/sg;
    $String = sortCSVLine($String);
    $String = "$Open$String$Close";
}

1;
