#========================================================================================
# package      DE_EPAGES::WebInterface::API::ApplicationServer
# state        public
#----------------------------------------------------------------------------------------
# description  The application server is a server process that implements the business
#               loging of the epages application. The process waits for incoming requests
#               from the message center and distributes them to the appropriate modules,
#               aka. servlets for further processing.<br>
#               This module encapsulates the communication protocol between the web
#               adapter and the applictaion server. It uses the MessageCenter module
#               to communicate the status of the process and to exchange cache hints
#               with other application servers.
# example
#               my $AppServer = DE_EPAGES::WebInterface::API::ApplicationServer->new(
#                   Address     => '213.232.211.243',  Port   => 10048,
#                   MaxRequest  => 5000,
#                   BindAddress => inet_ntoa( INADDR_ANY )
#               );
#               $AppServer->run;
#========================================================================================
package DE_EPAGES::WebInterface::API::ApplicationServer;
use base Exporter;
use strict;

our @EXPORT_OK = qw( GetAppServer TestSocket  ProcessRequest );

use feature 'switch';
use Config;
use IO::Socket::INET6;
use IO::Select;
use HTTP::Response;
use HTTP::Status; # RC_OK, RC_NOT_FOUND etc.
use Time::HiRes qw (time);
use DateTime;

use DE_EPAGES::Core::API::Constants qw(PERL538 STATISTIC_POSITION_TIME);
use DE_EPAGES::Core::API::Error  qw( Error ExistsError GetError HandleError );
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::Core::API::Hook qw( TriggerCoreHook );
use DE_EPAGES::Core::API::Statistics qw(WithStatistics GetStatisticString);
use DE_EPAGES::Core::API::IniConfig;
use DE_EPAGES::Core::API::Log    qw( GetLog LogError LogDebug );
use DE_EPAGES::Core::API::String qw( QuoteTags );
use DE_EPAGES::Core::API::File qw( ExistsFile GetCachedFile );
use DE_EPAGES::Core::API::Process qw( GetProcessMemory GetCurrentProcessID );
use DE_EPAGES::Core::API::TestParameter qw( TestString );
use DE_EPAGES::Core::API::Url qw( ParseQueryString );
use DE_EPAGES::DataCache::API::Cache qw( GetRootCache );
use DE_EPAGES::WebInterface::API::Cookie qw(SetCookie GetCookies);
use DE_EPAGES::WebInterface::API::Constants qw ( LONG_REQUEST_TIME SEND_X_APPSERVER SEND_X_POWERED_BY SEND_X_EPAGES_SITE SEND_X_EPAGES_REQUESTID );
use DE_EPAGES::WebInterface::API::Form;
use DE_EPAGES::WebInterface::API::MessageCenter;
use DE_EPAGES::WebInterface::API::ServerConfig qw ( ServerConfig );
use DE_EPAGES::XML::API::DateTimeFormatXML;
use DE_EPAGES::Core::API::PerlTools qw(UsePackage UnloadPackage);
BEGIN {
    if ($] < PERL538) {
        # use old packages for old perl
        UsePackage('Socket');
        Socket->import(qw(AF_INET AF_INET6));
        UsePackage('Socket6');
        Socket6->import(qw(inet_ntop));
    } else {
        # use new packages for new perl
        UsePackage('Socket');
        Socket->import(qw(AF_INET AF_INET6 inet_ntop));
    }
}

# the application server handling the current request
our $APPLICATION_SERVER;
my $SERVLET_CONFIG = $ENV{'EPAGES_CONFIG'}.'/Servlet.conf';

# used by _resetInfoCaches
use constant MEMORY_TRESHOLD => 0.85;   # At what % of MaxMem _resetInfoCaches should start? (350 MByte -> 300 Mbyte)
use constant WAIT_NR_REQUESTS => 200;   # number of requests to wait before next _resetInfoCaches
use constant MEGA => 1048576;

sub ErrorDefinition {
    return {
        BindServerSocket => 'Can\'t bind server socket on #LocalAddr:#LocalPort (#OSError)',
        NoServletFound   => 'No servlet was found for the current request',
        BadRequest       => 'Bad request matching filter "#Filter" for servlet #Servlet',
        RegExprFailed    => 'regualar expression in config file (#FileName) section (#Servlet) key (#Key) has an error (#ErrorMessage)',
        SocketFailed     => 'Connecting socket (#Host:#Port) failed (#Tries tries): #Code',
    };
}

#========================================================================================
# function     GetAppServer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $AppServer = GetAppServer();
#----------------------------------------------------------------------------------------
# description  Returns the application server object that handles the current request.
#               Returns undef if the application no application server is running.
#----------------------------------------------------------------------------------------
# return       $AppServer | the applcation server object | object
#========================================================================================
sub GetAppServer() {
    return $APPLICATION_SERVER;
}

#========================================================================================
# function     TestSocket
# state        public
#----------------------------------------------------------------------------------------
# syntax       TestSocket($Host, $Port, $Timeout, $Tries);
#----------------------------------------------------------------------------------------
# description  Checks connection to socket.
#----------------------------------------------------------------------------------------
# input        $Host | host name or IP address | string
# input        $Port | port | integer
# input        $Timeout | connection timeout in seconds (optional) | integer
# input        $Tries | number of connection tries (optional) | integer
#========================================================================================
sub TestSocket ($$;$$) {
    my ($Host, $Port, $Timeout, $Tries) = @_;

    local($^W) = 0;
    $Tries   = 8 unless $Tries;
    my $Sleep = 2;
    my ( $Socket, $OsError);

    my $Family = $Host =~ /:/ ? AF_INET6 : undef;
    $Family = AF_INET if $Host =~ /^([0-9]+\.){3}[0-9]+$/x;

    my %Options = ( PeerAddr => $Host,
                    PeerPort => $Port,
                    Proto    => 'tcp',
                    Timeout  => $Timeout // 4,
#                    Domain   => $Family,
        );

    foreach my $Try (1 .. $Tries ) {
        my $time1 = time;
        if ($Socket = IO::Socket::INET6->new( %Options  )) {
            close($Socket) or die "Could not close Socket: $!\n";
            return;
        }
        $OsError = $!;
        my $time2 = time;
        last if $time2 - $time1 > $Timeout and $Try * $Timeout >= $Tries * $Sleep;
        sleep($Sleep) unless $Try == $Tries;
    }
    Error( 'SocketFailed', { 'Host' => $Host, 'Port' => $Port, 'Tries' => $Tries, 'Code' => $OsError});
    return;
}

#========================================================================================
# function     ProcessRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Response = ProcessRequest($Request, $Response);
#----------------------------------------------------------------------------------------
# description  Executes the request and fills the response structure.
#----------------------------------------------------------------------------------------
# input        $Request | request object | HTTP::Request
# output       $Response | response object | HTTP::Response
#========================================================================================
sub ProcessRequest {
    my ($Request, $Response) = @_;
    GetAppServer()->_processRequest($Request, $Response);
    return;
}

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $AppServer = DE_EPAGES::WebInterface::API::ApplicationServer->new( %Options );
# example
#    my $AppServer = DE_EPAGES::WebInterface::API::ApplicationServer->new(
#        Address     => '213.232.211.243',  Port   => 10048,
#        MaxRequest  => 5000,
#        MaxUptime   => 3600,
#        MaxMemory   => 150,
#        BindAddress => inet_ntoa( INADDR_ANY )
#    );
#----------------------------------------------------------------------------------------
# description  Creates a new application server object.
#----------------------------------------------------------------------------------------
# input        %Options | some named parameters
#                   <ul><li>Address - IP address of the application server</li>
#                       <li>Port - port number where the application server listens for requests</li>
#                       <li>MaxRequest - maximum number of requests handled by the application server (default: 10000)</li>
#                       <li>MaxUptime - maximum runtime of the application server in seconds (default: undef=unlimited)</li>
#                       <li>MaxMemory - maximum memory size of the application server in MBytes (default: undef=unlimited)</li>
#                       <li>FreeRequestPeriod - period time in seconds when the AS contacts the MC (default: 5)</li>
#                       <li>BindAddress - ip address used to listen for requests
#                           (in case of multiple network interfaces)</li>
#                       <li>WebAdapter - web adapter object to receive the request</li>
#                       <li>MessageCenter - (optional) message center object (set by FakeRequest)</li>
#                   </ul> | hash
# return       $AppServer | the applcation server object | object
#========================================================================================
sub new {
    my $class = shift;
    my %Options = @_;
    LogDebug('Options', \%Options);

    %Options = (
        'MaxRequest'      => 10000,
        'FreeRequestPeriod' => 5,
        %Options,
        'Sockets'         => IO::Select->new,
        'ServerSocket'    => undef,
        'TiedHandle'      => undef,
        'RequestNo'       => 0,
        'CurrentRequest'  => undef,
        'CurrentResponse' => undef,
        'ClientSocket'    => undef,
        'ServerVariables' => {},
    );

    my $self = bless \%Options, $class;
    $self->_loadConfiguration() unless $self->messageCenter;

    return $self;
}

#========================================================================================
# function     messageCenter
# state        public
#----------------------------------------------------------------------------------------
# syntax       $MessageCenter = $AppServer->messageCenter;
#----------------------------------------------------------------------------------------
# description  Returns the message center.
#----------------------------------------------------------------------------------------
# return       $MessageCenter | message center object, see also DE_EPAGES::WebInterface::API::MessageCenter | object
#========================================================================================
sub messageCenter {
    return shift->{MessageCenter};
}

#========================================================================================
# function     address
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Address = $AppServer->address;
#----------------------------------------------------------------------------------------
# description  Returns the application server IP address.
#----------------------------------------------------------------------------------------
# return       $Address | ip address | string
#========================================================================================
sub address {
    return shift->{Address};
}

#========================================================================================
# function     port
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Port = $AppServer->port;
#----------------------------------------------------------------------------------------
# description  Returns the application server port number.
#----------------------------------------------------------------------------------------
# return       $Port | port number | int
#========================================================================================
sub port {
    return shift->{Port};
}

#========================================================================================
# function     sockets
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hSockets = $AppServer->sockets;
#----------------------------------------------------------------------------------------
# description  Returns the selector of sockets that are used by the application server. These
#               sockets include the server socket that is used to wait for incoming
#               requests and client sockets that are used for the actual data transfer
#               between application server and web adapter.
#----------------------------------------------------------------------------------------
# return       $hSockets | an IO::Select object | object
#========================================================================================
sub sockets {
    return shift->{Sockets};
}

#========================================================================================
# function     serverSocket
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Socket = $AppServer->serverSocket;
#----------------------------------------------------------------------------------------
# description  Returns the server socket that are used to wait for incoming requests
#               from the web adapter.
#----------------------------------------------------------------------------------------
# return       $Socket | an IO::Socket object | object
#========================================================================================
sub serverSocket {
    return shift->{ServerSocket};
}

#========================================================================================
# function     currentServlet
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Request = $AppServer->currentServlet;
#----------------------------------------------------------------------------------------
# description  Returns the currently processed servlet.
#----------------------------------------------------------------------------------------
# return       $Servlet | servlet object | object
#========================================================================================
sub currentServlet {
    return shift->{CurrentServlet};
}

#========================================================================================
# function     currentRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Request = $AppServer->currentRequest;
#----------------------------------------------------------------------------------------
# description  Returns the currently processed request.
#----------------------------------------------------------------------------------------
# return       $Request | request object, see HTTP::Request | object
#========================================================================================
sub currentRequest {
    return shift->{CurrentRequest};
}

#========================================================================================
# function     requestNo
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Counter = $AppServer->requestNo;
#----------------------------------------------------------------------------------------
# description  Returns the current request counter.
#----------------------------------------------------------------------------------------
# return       $Counter | the request counter value | int
#========================================================================================
sub requestNo {
    return shift->{RequestNo};
}

#========================================================================================
# function     currentResponse
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Request = $AppServer->currentResponse;
#----------------------------------------------------------------------------------------
# description  Returns the response object for the currently processed request.
#----------------------------------------------------------------------------------------
# return       $Request | response object, see HTTP::Response | object
#========================================================================================
sub currentResponse {
    return shift->{CurrentResponse};
}

#========================================================================================
# function     webadapter
# state        public
#----------------------------------------------------------------------------------------
# syntax       $WebAdapter = $AppServer->webadapter;
#----------------------------------------------------------------------------------------
# description  Returns the the webadapter interface object.
#----------------------------------------------------------------------------------------
# return       $WebAdapter | web adapter | object
#========================================================================================
sub webadapter {
    return shift->{WebAdapter};
}

#========================================================================================
# function     serverVariables
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hVars = $AppServer->serverVariables;
#----------------------------------------------------------------------------------------
# description  Returns a hash of server variables that correspond to the CGI interface.
#               The values are only valid for the current request.
#----------------------------------------------------------------------------------------
# return       $hVars | server variables | hash ref
#========================================================================================
sub serverVariables {
    return shift->{ServerVariables};
}

#========================================================================================
# function     run
# state        public
#----------------------------------------------------------------------------------------
# syntax       $AppServer->run;
#----------------------------------------------------------------------------------------
# description  Waits for incoming requests from the web adapter, distributes them to
#               request handlers and sends the response back to the web apapter. This
#               loop continues until the number of MaxRequest is reached or a 'shutdown'
#               request is received.
#========================================================================================
sub run {
    my $self = shift;

    local $APPLICATION_SERVER = $self;

    $self->{RequestNo} = 0;
    $self->{StartTime} = time();

    # prepare for receiving requests
    $self->_createServerSocket;
    $self->messageCenter->sendInitialFreeRequest;

    # run the startup scripts
    my $time1 = time;
    TriggerCoreHook('AppServerStartup');
    my $time2 = time;
    GetLog()->debug( sprintf( "AppServerStartup takes %0.1f sec\n", ($time2-$time1) ) );

    # the AS will process the number of requests that was requested
    # on the command line
    my $activeSocket;
    while( 1 )
    {
        # wait for something to happen in our list of sockets
        ($activeSocket) = $self->{'Sockets'}->can_read( $self->{'FreeRequestPeriod'} );

        # the timeout was reached, we will send a periodical FREE to the MC
        if( !defined $activeSocket ) {
            last if $self->_doShutdown(1); # check for MaxUptime

            $self->messageCenter->sendIdleRequest;
            next;
        }

        # a client wants to connect to the server
        # we will create a new socket and add it to the list of sockets
        if( $activeSocket == $self->serverSocket ) {
            $self->_createClientSocket;
            next;
        }

        # a regular request came in
        # we will receive the request and handle it
        local $self->{ClientSocket} = $activeSocket;
        my ($Request, $ServerVariables) = eval { $self->_receiveRequest( $activeSocket ); };
        if( ExistsError ) {
            # if there was an error while receiving the request the socket
            # will be closed an removed from the socket list
            my $Error = GetError();
            $self->_removeClientSocket( $activeSocket );
            if ($Error->code eq 'SocketClosed') {
                next;
            } elsif ($Error->code eq 'Shutdown') {
                my $ExtraInfo = $Error->vars->{'EXTRAINFO'};
                if ($ExtraInfo eq 'shutdown') {
                    $self->_logShutdown( '"shutdown" request' );
                    last;
                } elsif ($ExtraInfo eq 'restart') {
                    $self->_logShutdown( '"restart" request' );
                    last;

                }
            } elsif($Error->code eq 'Reload') {
                GetLog()->info('"reload" request');
                $self->_loadConfiguration();
                next;
            } elsif($Error->code eq 'ReloadModules') {
                GetLog()->info('"reloadmodules" request');
                my $ModuleList = $Error->vars->{'MODULELIST'};
                $self->_reloadModules($ModuleList);
                next;
            } else {
                LogError($Error);
                next;
            }
        }

        # get the latest cache data from the MC
        $self->messageCenter->sendWorkingRequest;

        # check if any module timestamp has changed and reload these
        if ($self->{'AutoRecompile'}) {
            GetLog()->debug( 'AutoRecompile mode... refreshing %INC. Disable this feature in Webinterface.conf > AUTORECOMPILE=0' );
            UsePackage('Module::Refresh');
            Module::Refresh->refresh;
        }

        # now process the request inside of the AS
        local $self->{FlushedOutput} = 0;
        local $self->{ServerVariables} = $ServerVariables;
        my $Response = HTTP::Response->new( RC_OK );
        $self->_processRequest( $Request, $Response );

        my $DoShutdown = 0;
        eval {
            $DoShutdown = $self->_doShutdown;
            if (! $DoShutdown) {
                # tell the MC that this AS is idle again
                # Note: sendIdleRequest() must be called before flushOutput(), otherwise
                # a subsequent web service request might be served by a different AS with
                # outdated cache data
                $self->messageCenter->sendIdleRequest;
            }
        };
        HandleError( Finally => sub {
            # send the response to WA in spite of a previous MC error
            # then re-throw the original error
            eval {
                $self->flushOutput( $Response );
            };
            HandleError( OnError => sub {
                my ($Error) = @_;
                LogError( $Error );
                $self->_removeClientSocket( $activeSocket );
                return 'FIXED';
            });
        });

        last if $DoShutdown;
    }

    # tell the MC that this AS is shut down
    $self->messageCenter->sendFinishRequest;

    TriggerCoreHook('AppServerShutdown');

    return;
}

sub _reloadModules {
    my $self = shift;
    my ($ModuleList) = @_;

    return GetLog()->info('no modules to reload') unless $ModuleList;

    # refresh given modules
    UsePackage('Module::Refresh');
    foreach my $Module ( split(/,/,$ModuleList) ) {
        if($Module =~ /::/gi ){
            $Module =~ s/::/\//g;
            $Module .= '.pm';
        }

        if(not ExistsFile($ENV{'EPAGES_CARTRIDGES'}.'/'. $Module)){
            GetLog()->info("Can not reload module: $Module doesn't exists.");
            next;
        }

        GetLog()->info("reload single module: $Module");
        Module::Refresh->refresh_module($Module);
    }

    # we don't need this module anymore
    UnloadPackage('Module::Refresh');

    return;
}

#========================================================================================
# function     _doShutdown
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $DoShutdown = $self->_doShutdown($WithoutRequest);
#----------------------------------------------------------------------------------------
# description  We are going to shutdown because one of
#               the specified limits (requests, duration or memory) has been reached.
#----------------------------------------------------------------------------------------
# input        $WithoutRequest | server had no request (default:0) | boolean
# return       $DoShutdown | shut down or not | boolean
#========================================================================================
sub _doShutdown {
    my $self = shift;
    my ($WithoutRequest) = @_;
    if (not $WithoutRequest) {
        # check the limits
        if( defined $self->{MaxRequest} && $self->{RequestNo} >= $self->{MaxRequest} ) {
            $self->_logShutdown( "request limit exceeded ($self->{RequestNo})" );
            return 1;
        }

        if( defined $self->{MaxMemory} ) {
            my $CurrMemSize = int( GetProcessMemory( GetCurrentProcessID() ) / MEGA);

            # Delete some Caches near Memory Limit
            if ($self->{MaxMemory} * MEMORY_TRESHOLD < $CurrMemSize) {
                if( 0 == ($self->requestNo % WAIT_NR_REQUESTS) ) {
                    $self->_resetInfoCaches;
                    $CurrMemSize = int( GetProcessMemory( GetCurrentProcessID() ) / MEGA);
                }
            }

            if( $CurrMemSize >= $self->{MaxMemory} ) {
                $self->_logShutdown( "memory limit exceeded ($CurrMemSize MBytes)" );
                return 1;
            }
        }
    }

    if( defined $self->{MaxUptime} ) {
        my $Uptime = time() - $self->{StartTime};
        if( $Uptime >= $self->{MaxUptime} ) {
            $self->_logShutdown( "uptime limit exceeded ($Uptime seconds)" );
            return 1;
        }
    }

    return 0;
}

#========================================================================================
# function     _loadConfiguration
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_loadConfiguration();
#----------------------------------------------------------------------------------------
# description  Loads ServerConfig.xml and updates pool configuration of this application
#               server
#========================================================================================
sub _loadConfiguration {
    my $self = shift;

    $self->messageCenter->close if($self->messageCenter);
    my $hGroup = ServerConfig()->appServerGroup($self->address, $self->port);
    $self->{'MessageCenter'} = DE_EPAGES::WebInterface::API::MessageCenter->new(
        'AppServerAddress'  => $self->address,
        'AppServerPort' => $self->port,
        'PoolName' => $hGroup->{'PoolName'},
        'Priority' => $hGroup->{'Priority'}
    );

    # connect to MC, try multiple times if MC is down
    $self->messageCenter->connect(100); ## no critic qw(ProhibitMagicNumbers)
}

#========================================================================================
# function     _logShutdown
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_logShutdown( $Reason );
#----------------------------------------------------------------------------------------
# description  Writes a log entry when the application server is shut down.
#               Calls <function _logCacheStats>.
#               Enable in log4perl.conf:
#               log4perl.category.DE_EPAGES::WebInterface::API::ApplicationServer::_logShutdown = INFO
#----------------------------------------------------------------------------------------
# input        $Reason | description, why the process is shut down | string
#========================================================================================
sub _logShutdown {
    my $self = shift;
    my ($Reason) = @_;
    TestString( 'Reason', $Reason );

    GetLog()->info( 'shutting down because: ' . $Reason );

    # write addtional statistics
    $self->_logCacheStats;
    return;
}

#========================================================================================
# function     _logCacheStats
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_logCacheStats();
#----------------------------------------------------------------------------------------
# description  Writes additional statistical information about the application server
#               process we are going to shutdown.
#               Enable in log4perl.conf:
#               log4perl.category.DE_EPAGES::WebInterface::API::ApplicationServer::_logCacheStats = INFO
#               or
#               log4perl.category.DE_EPAGES::WebInterface::API::ApplicationServer::_logCacheStats = DEBUG
#========================================================================================
sub _logCacheStats {
    my $self = shift;

    my $Logger = GetLog();
    return unless $Logger->is_info;

    my $RootCache = GetRootCache();
    my $Parser = DE_EPAGES::XML::API::DateTimeFormatXML->new;

    my $result = '';
    $result .= sprintf "Total cache size: %.2f MBytes\n", $RootCache->totalBytes / MEGA;
    $result .= sprintf "Requests handled: %d\n", $self->{RequestNo};
    $result .= sprintf "Final memory size: %.2f MByte\n", int( GetProcessMemory( GetCurrentProcessID() ) / MEGA );
    $result .= sprintf "start time: %s\n", $Parser->format_datetime( DateTime->from_epoch(epoch => $self->{StartTime}) );
    $result .= sprintf "end time: %s\n", $Parser->format_datetime( DateTime->now );
    $result .= sprintf "uptime: %d sec\n", int(time() - $self->{StartTime});

    # detailed cache info only at debug logging stage
    if( $Logger->is_debug ) {
        $result .= "Detailed CacheInfo:\n";
        foreach my $hInfo ( @{$RootCache->infos} ) {
            $result .=
                sprintf "Name: %s\nCurrentCount: %s (MaxCount: %s)\nCurrentBytes: %s (MaxBytes %s)\nHits: %s\nMisses: %s\n\n",
                defined $hInfo->{'Name'} ? $hInfo->{'Name'} : 'undef',
                defined $hInfo->{'CurrentCount'} ? $hInfo->{'CurrentCount'} : 'undef',
                defined $hInfo->{'MaxCount'} ? $hInfo->{'MaxCount'} : 'undef',
                defined $hInfo->{'CurrentBytes'} ? $hInfo->{'CurrentBytes'} : 'undef',
                defined $hInfo->{'MaxBytes'} ? $hInfo->{'MaxBytes'} : 'undef',
                defined $hInfo->{'Hits'} ? $hInfo->{'Hits'} : 'undef',
                defined $hInfo->{'Misses'} ? $hInfo->{'Misses'} : 'undef';
        }
        $Logger->debug($result);
    }
    else {
        $Logger->info($result);
    }

    return;
}

#========================================================================================
# function     _resetInfoCaches
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_resetInfoCaches();
#----------------------------------------------------------------------------------------
# description  Reset all InfoCaches but InfoClass and InfoAttribute
#========================================================================================
sub _resetInfoCaches {
    my $self = shift;

    my $MBytes = int( GetProcessMemory( GetCurrentProcessID() ) / MEGA );
    GetLog()->debug( 'Memory Limit near. Clearing InfoCaches.' );
    GetLog()->debug( "Before cache update: $MBytes MBytes" );

    my $hAll = GetRootCache()->caches;
    foreach my $ConnectionKey (keys %$hAll) {
        my $hConnection = $hAll->{$ConnectionKey};
        foreach my $CacheName (keys %$hConnection) {
            if ($CacheName =~ m/^Info/ and
                $CacheName !~ m/^Info(?:Attribute|Class|Menu|Cartridge|Language|Locale)/ ) {
                my $Cache = $hConnection->{$CacheName};
                $Cache = $Cache->hash unless ref $Cache eq 'HASH';
                %$Cache = ();
            }
        }
    }

    $MBytes = int( GetProcessMemory( GetCurrentProcessID() ) / MEGA );
    GetLog()->debug( "After cache update $MBytes MBytes" );

    return;
}

#========================================================================================
# function     _createServerSocket
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_createServerSocket;
#----------------------------------------------------------------------------------------
# description  Creates the server socket that is used to listen for incoming connections.
#========================================================================================
sub _createServerSocket {
    my $self = shift;

    # the reuse under Unix must be set, to allow a fast restart
    # under NT must not be set, to prevent to Apps from
    # binding to the same socket twice
    my %options = (
        LocalAddr => $self->{BindAddress},
        LocalPort => $self->port,
        Proto     => 'tcp',
        Listen    => 5,
        Type      => SOCK_STREAM,
    );
    if( $Config{'osname'} ne 'MSWin32' ) {
        # SO_REUSEADDR will allow us to use the address while a previous connection
        # is still shutting down
        $options{'Reuse'} = SO_REUSEADDR;
    }

    my $ServerSocket = IO::Socket::INET6->new( %options )  or Error( BindServerSocket => {
        LocalAddr => $options{LocalAddr},
        LocalPort => $options{LocalPort},
        OSError   => $!
    } );

    # finally add the socket to the list
    $self->sockets->add($ServerSocket);

    GetLog->debug( sub {
        sprintf "Added server socket #%d at %s:%d to the socket list.\n".
                'The socket list now contains: %s',
                $ServerSocket->fileno(), $options{LocalAddr}, $options{LocalPort},
                join( ' ', map { $_->fileno() } $self->sockets->handles )
    });

    $self->{ServerSocket} = $ServerSocket;

    return;
}

#========================================================================================
# function     _removeClientSocket
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_removeClientSocket( $Socket ):
#----------------------------------------------------------------------------------------
# description  Removes a client socket from the socket list.
#----------------------------------------------------------------------------------------
# input        $ClientSocket | client socket that is to be removed | object
#========================================================================================
sub _removeClientSocket {
    my $self = shift;
    my( $socket ) = @_;

    $self->sockets->remove($socket);

    GetLog->debug( 'The socket #'. $socket->fileno() .' was closed and removed from the socket list' );

    return;
}

#========================================================================================
# function     _createClientSocket
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_createClientSocket;
#----------------------------------------------------------------------------------------
# description  This procedure has to be called whenever a client wishes to
#               connect to the server.
#               A new connection is established and the new socket is added to the socket
#               list.
#========================================================================================
sub _createClientSocket {
    my $self = shift;

    # accept the connection from the client
    my $socket     = $self->serverSocket->accept() or die 'Error creating new client socket';

    # add the newly created socket to the socket list
    $self->sockets->add($socket);

    GetLog->debug( sub {

        my $clientPort = $socket->peerport();
        my $family     = length($socket->peeraddr()) == 16 ? AF_INET6 : AF_INET; ## no critic qw(ProhibitMagicNumbers)
        my $clientIP   = inet_ntop($family, $socket->peeraddr());

        sprintf "Added client socket #%d at %s:%d to the socket list.\n".
            'The socket list now contains: %s',
            $socket->fileno(), $clientIP, $clientPort,
            join( ' ', map { $_->fileno() } $self->sockets->handles )
    });

    return;
}

#========================================================================================
# function     _receiveRequest
# state        private
#----------------------------------------------------------------------------------------
# syntax       ($Request, $ServerVariables) = $self->_receiveRequest( $ClientSocket );
#----------------------------------------------------------------------------------------
# description  Receives the HTTP request (headers and content) that was sent by the web
#               adapter through the client socket.
#               Stores the CGI environment variables in the $self->{'ServerVariables'}
#               attribute.
#----------------------------------------------------------------------------------------
# input        $ClientSocket | active client socket that is connected to a web adapter | object
# return       $Request | received request from web adapter | HTTP::Request
#========================================================================================
sub _receiveRequest {
    my $self = shift;
    my ($clientSocket) = @_;

    $self->webadapter->setSockedFD($clientSocket->fileno());
    return $self->webadapter->receive;
}

#========================================================================================
# function     _handleRequest
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_handleRequest( $Request, $Response );
#----------------------------------------------------------------------------------------
# description  Handles the request and produces the response.
#               Does not do error handling and profiling.
#----------------------------------------------------------------------------------------
# input        $Request  | received request from web adapter | HTTP::Request
# output       $Response | Response for the Webbrowser | HTTP::Response
#========================================================================================
sub _handleRequest {
    my $self = shift;
    my ($Request, $Response) = @_;

    my $log = GetLog;

    my $IniConfig = GetCachedFile( $SERVLET_CONFIG, sub {
        my ($FileName) = @_;
        GetLog->debug( "Loading config file $FileName" );
        return DE_EPAGES::Core::API::IniConfig->new( FileName => $FileName, CaseSensitive => 1 );
    });

    # get all servlet names ordered by key number
    my $hServlets = $IniConfig->section( 'Servlets' );
    my @Servlets = map { $hServlets->{$_} } sort { $a <=> $b } keys %$hServlets;

    my $hServerVariables = $self->serverVariables; # used by regex in Servlet.conf
    #init undefined variables to use in regex of Servlet.conf
    foreach (qw(HTTP_ACCEPT QUERY_STRING)) {
        $hServerVariables->{$_} = '' unless defined $hServerVariables->{$_}
    }

    my $Servlet;
    my $hSection;
    while( defined( $Servlet = shift @Servlets ) ) {
        $hSection = $IniConfig->section( $Servlet );
        my $match = eval $hSection->{'expression'}; ##no critic qw(ProhibitStringyEval)
        if( $@ ) {
            Warning('RegExprFailed', {'ErrorMessage' => $@, 'FileName'=> $SERVLET_CONFIG, 'Servlet'=> $Servlet, 'Key'=>'expression'});
            next;
        }

        if( $match ) {
            $log->debug( "Selected servlet $Servlet matches $hSection->{'expression'}" );
            last;
        }
    }
    Error( 'NoServletFound' ) unless defined $Servlet;

    # filter bad URLs
    if( defined $hSection->{'filter'} ) {
        my $match = eval $hSection->{'filter'}; ##no critic qw(ProhibitStringyEval)
        if( $@ ) {
            Warning('RegExprFailed', {'ErrorMessage' => $@, 'FileName'=> $SERVLET_CONFIG, 'Servlet'=> $Servlet, 'Key'=>'filter'});
        }

        if( $match ) {
            Error('BadRequest', { 'Servlet' => $Servlet, 'Filter' => $hSection->{'filter'} });
        }
    }

    # get the servlet package from the config file
    my $Package = $hSection->{'package'};
    UsePackage( $Package );

    # get parameters for the servlet
    my %Parameters;
    foreach my $Key ( grep { !/^expression|package$/ } keys %$hSection ) {
        my ($Value) = eval $hSection->{$Key}; ##no critic qw(ProhibitStringyEval)
        if( $@ ) {
            Warning('RegExprFailed', {'ErrorMessage' => $@, 'FileName'=> $SERVLET_CONFIG, 'Servlet'=> $Servlet, 'Key'=>$Key});
            next;
        }

        $log->debug("Servlet parameter $Key=". (defined $Value ? $Value : '(undef)'));
        $Parameters{ lc($Key) } = $Value;
    }

    my $Handler = $Package->new;
    local $self->{'CurrentServlet'} = $Handler;
    $Handler->init( \%Parameters ) if $Handler->can( 'init' );
    $Handler->handle( $Request, $Response );

    return;
}

sub _handleError {
    my $self = shift;
    my ($Error, $Request, $Response) = @_;

    # add some CGI variables to error message
    my $hCGIVars = $self->serverVariables;
    my %ErrorVars =
        map { ( $_ => $hCGIVars->{$_} ) }
        grep { exists $hCGIVars->{$_} }
        qw( SCRIPT_NAME PATH_INFO QUERY_STRING HTTP_HOST HTTP_REFERER HTTP_USER_AGENT REMOTE_ADDR );

    $Error->addVars( \%ErrorVars );
    $Error->addVars( { Request => $Request->as_string } );

    my $ResponseCode = RC_INTERNAL_SERVER_ERROR;

    # write error to log file
    eval {
        given ($Error->code) {
            when ('BadRequest') {
                GetLog('USERFAULT::NoServletFound')->warn($Error->longMessage);
            }
            when ('NoServletFound') {
                GetLog('USERFAULT::NoServletFound')->warn($Error->longMessage);
                $ResponseCode = RC_NOT_FOUND;
            }
            when ('NON_PRINTABLE_CHARS') {
                GetLog('USERFAULT::NON_PRINTABLE_CHARS')->warn($Error->longMessage('SkipStackTrace' => 1));
                $ResponseCode = RC_BAD_REQUEST;
            }
            default {
                LogError($Error);
            }
        }
    };
    my $LogError = ExistsError() ? GetError() : undef;

    # print error message on screen
    $Response->code($ResponseCode);
    $Response->content_type('text/html');
    $Response->header('Cache-Control', 'no-cache');

    my $Content;
    if (defined $LogError) { # the error could not be written to log file, so display it
        my $Message = QuoteTags($Error->longMessage) . "\n" . QuoteTags($LogError->longMessage);
        $Content = <<END_ERROR;
<html>
<head><title>Server Error</title></head>
<body><pre>$Message</pre></body>
</html>
END_ERROR
    }
    else {
        $Content = $self->_errorPage($Error);
    }

    $Response->content($Content);
    return;
}

#========================================================================================
# function     _processRequest
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_processRequest( $Request , $Response);
#----------------------------------------------------------------------------------------
# description  Distributes the request to handler modules, aka. servlets.
# input        $Request  | received request from web adapter | HTTP::Request
# output       $Response | Response for the Webbrowser | HTTP::Response
#========================================================================================
sub _processRequest {
    my $self = shift;
    my ($Request, $Response) = @_;

    eval {
        my $time0 = time();
        local $self->{CurrentRequest} = $Request;
        local $self->{CurrentResponse} = $Response;

        # increment the request counter
        $self->{RequestNo}++;
        GetLog()->debug( "Request $self->{RequestNo} : " . $self->currentRequest->uri );

        # do not use a form here because we need only one parameter
        my $hQuery = ParseQueryString( $Request->uri->query, 'utf-8' );
        my $paramPerfBar = $hQuery->{'PerformanceBar'}->[0];
        $hQuery = undef;

        # log statistics
        my $PerformanceLog = GetLog('PERFORMANCE::ApplicationServer');
        my $hCockies = GetCookies($Request);
        my $hStatistics = WithStatistics( sub {
            $self->_handleRequest( $Request, $Response )
        });

        my $Headers = $Response->headers;
        $Headers->header( 'X-EPAGES_SITE', $self->currentRequest->headers->header('X-EPAGES_SITE')) if SEND_X_EPAGES_SITE;
        $Headers->header( 'X-EPAGES-RequestId', $self->currentRequest->headers->header('X-EPAGES-RequestId')) if SEND_X_EPAGES_REQUESTID;
        $Headers->header( 'X-AppServer', $self->address .':'.$self->port ) if SEND_X_APPSERVER;
        $Headers->header( 'X-Powered-By', 'epages 6' ) if SEND_X_POWERED_BY;

        my $TimeDiff = $hStatistics->{'__MAIN__'}[STATISTIC_POSITION_TIME] //0;
        if( defined $paramPerfBar ){
            if($paramPerfBar){
                SetCookie($Response,'PerformanceBar', 'enabled', { 'Path' => '/epages/' });
            }
            else{
                SetCookie($Response,'PerformanceBar', undef, { 'Path' => '/epages/' });
            }
        }

        my $contentType = $Headers->header('Content-Type') //'';
        my $view = $contentType =~ m/^text\/(?:html|plain)/i   ? 1 : 0;

        if((defined ($hCockies->{'PerformanceBar'}) ||  $paramPerfBar) && $view){
            my $TLETime= $hStatistics->{'Template_Processing'}[STATISTIC_POSITION_TIME] //0;
            my $MysqlTime=$hStatistics->{'mysqlExecute'}[STATISTIC_POSITION_TIME] //0;
            my $Memory = int( GetProcessMemory( GetCurrentProcessID() ) / MEGA );
            SetCookie($Response,'RequestTime', $TimeDiff, { 'Path' => '/epages/' });
            SetCookie($Response,'TLETime', $TLETime, { 'Path' => '/epages/' });
            SetCookie($Response,'MysqlTime', $MysqlTime, { 'Path' => '/epages/' });
            SetCookie($Response,'AppServerMemory', $Memory, { 'Path' => '/epages/' });
        }

        if($PerformanceLog->is_info && $TimeDiff > LONG_REQUEST_TIME ){
            my $Statistics=GetStatisticString($hStatistics,{
                'referer' => $self->serverVariables->{HTTP_REFERER} // '',
                'uri'     => $Request->uri,
            });
            $PerformanceLog->info( $Statistics );
        }

        $Response->header( 'X-TimeAS', sprintf '%i', (time - $time0) * 1000 ); ## no critic(ProhibitMagicNumbers)
    };
    if( ExistsError ) {
        $self->_handleError( GetError(), $Request, $Response );
    }

    return;
}

#========================================================================================
# function     enterMonitorMode
# state        public
#----------------------------------------------------------------------------------------
# syntax       $AppServer->enterMonitorMode;
#----------------------------------------------------------------------------------------
# description  Sets the application server into monitor mode which means that the
#               application server will send no further output and the progress can
#               be monitored.
#========================================================================================
sub enterMonitorMode {
    my $self = shift;

    $self->messageCenter->sendMonitorRequest;

    return;
}

#========================================================================================
# function     flushOutput
# state        public
#----------------------------------------------------------------------------------------
# syntax       $AppServer->flushOutput( $Response );
#----------------------------------------------------------------------------------------
# description  Sends the buffered response data immediately to the client.
#----------------------------------------------------------------------------------------
# input        $Response | respnse object | HTTP::Response
#========================================================================================
sub flushOutput {
    my $self = shift;
    my ($Response) = @_;

    unless( $self->{FlushedOutput} ) {
        $self->webadapter->send($Response);
        $self->{FlushedOutput} = 1;
    }
    return;
}

#========================================================================================
# function     _errorPage
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Content = $AppServer->_errorPage( $Error );
#----------------------------------------------------------------------------------------
# description  Returns the HTML code of the last-chance error message.
#               The message is in English and German.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
# return       $Content | HTML content of errorpage | String
#========================================================================================
sub _errorPage {
    my $self = shift;
    my ($Error) = @_;

    my $ErrorCode = $Error->guid;
    my $ErrorHash = $Error->hash;
    my $ErrorBox;

    given ($Error->code) {
        when ('NON_PRINTABLE_CHARS') {
            $ErrorBox = <<HTML;
      <h1>An error occurred during processing.</h1>
      <p>The request contains an invalid control character.</p>
      <h1>Bei der Bearbeitung ist ein Fehler aufgetreten.</h1>
      <p>In der Anfrage ist ein nicht erlaubtes Steuerzeichen enthalten.</p>
HTML
        }
        default {
            $ErrorBox = <<HTML;
      <h1>An error occurred during processing.</h1>
      <p>Information about this error is available under the number listed below in the log file on the server.</p>
      <h1>Bei der Bearbeitung ist ein Fehler aufgetreten.</h1>
      <p>Die Details zu diesem Fehler wurden mit der unten genannten Nummer in der Protokolldatei auf dem Server hinterlegt.</p>
      <p>GUID: $ErrorCode</p>
      <p>Stacktrace hash: $ErrorHash</p>
HTML
        }
    }

    return <<HTML;
<!DOCTYPE html>
<html xmlns="http://www.w3.org/1999/xhtml" lang="de" xml:lang="de">
  <head>
    <title>Error</title>
    <style type="text/css">
        body{background:#F9F9F9;color:#333;padding:5em 0 3em;font-size:14px;line-height:1.5;text-align:center;font-family:sans-serif}h1{font-weight:400;color:#c00;font-size:1.5em}#ErrorBox{min-width:280px;max-width:50%;text-align:left;margin:0 auto}
    </style>
  </head>
  <body>
    <div id="ErrorBox">
        $ErrorBox
    </div>
  </body>
</html>
HTML
}

1;
