# BEGIN {print "start fast\n";} # uncomment this line if the AS starts slow

use warnings;
use strict;
use Getopt::Long;
use lib $ENV{EPAGES_CARTRIDGES};

use DE_EPAGES::Core::API::PerlTools qw( UsePackage GetIPAddress SetApplicationServerPort );
use DE_EPAGES::Core::API::Process qw( GetCurrentProcessID SetProcessPriority );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::WebInterface::API::ApplicationServer;
use DE_EPAGES::WebInterface::API::Config qw( WebInterfaceConfig );
use DE_EPAGES::WebInterface::API::Object::WebAdapter;
use DE_EPAGES::WebInterface::API::ServerConfig qw( ServerConfig );

use DE_EPAGES::Core::API::Constants qw(PERL538);
BEGIN {
    if ($] < PERL538) {
        # use old packages for old perl
        UsePackage('Socket');
        Socket->import(qw(INADDR_ANY AF_INET6 inet_ntoa));
        UsePackage('Socket6');
        Socket6->import(qw(inet_pton inet_ntop));
    } else {
        # use new packages for new perl
        UsePackage('Socket');
        Socket->import(qw(INADDR_ANY AF_INET6 inet_ntoa inet_pton inet_ntop));
    }
}

#  start the main function
RunScript( Sub => \&Main );

#========================================================================================
# function     Main
# state        public
#----------------------------------------------------------------------------------------
# syntax       Main()
#----------------------------------------------------------------------------------------
# description  Reads the command line parameters and creates an ApplicationServer
#               object
#========================================================================================
sub Main {

    # the following variables will be set from the command line
    my $help = 0;             # the help flag
    my $AppServerPort;        # the own server port
    my $HostName;             # this is the host name or IP of the AS that will be sent
                              # along with the protocol
    my $Bind = 0;             # will use binding if it is set

    if( !GetOptions ( 'help'         => \$help,
                      'bind'         => \$Bind,
                      'localip:s'    => \$HostName,
                      'port=i'       => \$AppServerPort,
                      ) ||
        !defined( $AppServerPort ) )
    {
        Usage();
        exit 1;
    }

    # print the usage screen if it was requested
    if( $help ) {
        Usage();
        return;
    }

    # if the HostName was not set, we will use the primary IP
    my $AppServerIP;
    if( defined $HostName ) {
        $AppServerIP = GetIPAddress($HostName);
        die "Can't resolve host name '$HostName'" unless defined $AppServerIP;
    }
    else {
        $AppServerIP = GetIPAddress();
        $HostName = $AppServerIP;
    }

    SetApplicationServerPort($AppServerPort);

    # $In6AddrAny is same as IN6ADDR_ANY (new Socket) or in6addr_any (old Socket6)
    my $In6AddrAny = inet_pton(AF_INET6, '::');
    my $InAddrAny = $AppServerIP =~ /:/ ? inet_ntop(AF_INET6, $In6AddrAny) : inet_ntoa(INADDR_ANY);

    my %Options = (
        Address     => $HostName,
        Port        => $AppServerPort,
        WebAdapter  => DE_EPAGES::WebInterface::API::Object::WebAdapter->new,
        BindAddress => $Bind ? $AppServerIP : $InAddrAny,
    );

    # Get pool and search for a max memory set for this application server.
    # If none found, just take the value from the WebInterface.conf file.
    my $ServerConfig = ServerConfig();
    my $hAppServer = $ServerConfig->appServerGroup( $HostName, $AppServerPort );
    $Options{'MaxMemory'} = $hAppServer->{'MaxMemory'} if defined $hAppServer->{'MaxMemory'};

    # read options from section GLOBAL in WebInterface.conf;
    my $IniConfig = WebInterfaceConfig();
    foreach my $Key (qw( MaxRequest MaxUptime MaxMemory FreeRequestPeriod AutoRecompile )) {
        next if defined $Options{$Key}; # do not overwrite set values
        my $Value = $IniConfig->get( 'GLOBAL', uc($Key) );
        $Options{$Key} = $Value if defined $Value;
    }

    # now start the application server
    my $AppServer = DE_EPAGES::WebInterface::API::ApplicationServer->new( %Options );

    # adjust the process priority
    my $Priority = $IniConfig->get( 'GLOBAL', 'PRIORITY' );
    SetProcessPriority( GetCurrentProcessID(), $Priority ) if defined $Priority;

    # wait for incoming requests
    $AppServer->run;
}

#========================================================================================
# function     Usage
# state        public
#----------------------------------------------------------------------------------------
# syntax       Usage()
#----------------------------------------------------------------------------------------
# description  Prints information about the command line options.
#========================================================================================
sub Usage {

    print STDERR <<END_HELP;
epages Application Server
usage: perl $0 [options]
  options are:

  -port       the port on which the application server should be listening
              (required)
  -bind       if this option is set then the AS will bind its server port to
              the requested IP (-localip) otherwise it will bind to INADDR_ANY
  -localip    is the IP that the AS should use for itself; the AS will
              use this IP for its own identification, otherwise it will
              use the primary ip address of the local machine
  -help       print this text

example: perl $0 -port 10045
END_HELP

    return;
}
