use base DE_EPAGES::Core::API::Test;
use strict;
use Test::More tests => 25;
use DE_EPAGES::WebInterface::API::ServerConfig qw( ServerConfig );
use DE_EPAGES::WebInterface::API::XML::ServerConfigImportHandler;
use DE_EPAGES::WebInterface::API::XML::ServerConfigExportDriver;
use Data::Dumper;

## no critic(ProhibitMagicNumbers)

sub testServerConfig {

    my $Config = ServerConfig();
    my $PoolName = 'DefaultPool';
    my ($Host, $Port) = $Config->messageCenter($PoolName);
    ok( defined $Host, 'messageCenter.Host' );
    ok( $Port > 0, 'messageCenter.Port' );
    ok( scalar@{ $Config->fallbackMessageCenters($PoolName) } >= 0, 'fallbackMessageCenters' );
}

sub testImportExport {
    my $self = shift;
    my $Importer = DE_EPAGES::WebInterface::API::XML::ServerConfigImportHandler->new;
    $Importer->importXML('ServerConfigImport.xml');

    my $ExportDriver = DE_EPAGES::WebInterface::API::XML::ServerConfigExportDriver->new($Importer);
    $ExportDriver->exportXML('ServerConfigExport.temp.xml');

    my $Config1 = DE_EPAGES::WebInterface::API::ServerConfig->new( 'ServerConfigExport.temp.xml' );
    my $Config2 = DE_EPAGES::WebInterface::API::ServerConfig->new( 'ServerConfigImport.xml' );

    is_deeply($Config1->{XMLConfig}, $Config2->{XMLConfig}, 'exported ServerConfig equals imported ServerConfig');

    unlink('ServerConfigExport.temp.xml');
}

sub test {
    my $self = shift;
    $self->testServerConfig;

    my $Config = DE_EPAGES::WebInterface::API::ServerConfig->new( 'ServerConfig.xml' );
    my $hhRequestRouters = $Config->requestRouters;
    is( scalar keys %$hhRequestRouters, 4, 'requestRouters' );
    is( $hhRequestRouters->{'Pool_1_Default'}->{'Name'}, 'Pool_1_Default', 'requestRouters value' );

    my $hRequestRouter = $Config->requestRouter('Pool_1_Fallback');
    is_deeply( $hRequestRouter, {
        Name => 'Pool_1_Fallback',
        Host => 'w2',
        Port => 10063,
        Bind => 0,
        MonitorPort => 10061,
        MessageCenterPort => 10062,
    }, 'requestRouter' );

    $hRequestRouter = $Config->requestRouter('Pool_2_Default');
    ok( !defined $hRequestRouter->{'MonitorPort'}, 'no monitor port' );

    my $hhPools = $Config->pools;
    is( scalar keys %$hhPools, 3, 'pools' );
    is( $hhPools->{'Pool_1'}->{'Name'}, 'Pool_1', 'pools value' );

    my $hPool = $Config->pool('Pool_2');
    my $aRequestRouters = $hPool->{'RequestRouters'};
    is( scalar @$aRequestRouters, 2, 'pool.RequestRouters' );
    is( $aRequestRouters->[1], 'Pool_2_Fallback', 'pool.RequestRouters value' );

    my $hAppServers = $hPool->{'AppServers'};
    is( scalar keys %$hAppServers, 3, 'pool.AppServers' );
    is( $hAppServers->{'a5_10045-10055'}->{'Name'}, 'a5_10045-10055', 'pool.AppServers value' );

    my $hGroup = $Config->appServerGroup('a2', 10045);
    is_deeply( $hGroup, {
        Name => 'a2_10045-10055',
        Host => 'a2',
        Ports => '10045-10055',
        Bind => 1,
        PoolName => 'Pool_1',
        Priority => 2,
        MaxMemory => 300,
    }, 'appServerGroup' );

    my ($Host, $Port) = $Config->messageCenter('Pool_1');
    is( $Host, 'w1', 'messageCenter.Host' );
    is( $Port, '10042', 'messageCenter.Port' );

    my $aMCs = $Config->fallbackMessageCenters('Pool_1');
    is( scalar @$aMCs, 1, 'fallbackMessageCenters' );
    is( $aMCs->[0]->[0], 'w2', 'fallbackMessageCenters.Host' );
    is( $aMCs->[0]->[1], '10062', 'fallbackMessageCenters.Port' );

    my $aMonitors = $Config->requestRouterMonitors;
    $aMonitors = [sort { $a->[0] cmp $b->[0] or $a->[1] <=> $b->[1] } @$aMonitors];
    my $Expected = [['w1', 10041],['w1', 10061],['w2', 10061]];
    is_deeply( $aMonitors, $Expected, 'requestRouterMonitors' );

    my $Servers = $Config->asPoolDBCacheServers;
    is( $Servers->[0]->{Host}, 'w1', 'asPoolDBCacheServer.Host1' );
    is( $Servers->[0]->{Port}, 17277, 'asPoolDBCacheServer.Port1' );
    is( $Servers->[1]->{Host}, 'w2', 'asPoolDBCacheServer.Host2' );
    is( $Servers->[1]->{Port}, 17277, 'asPoolDBCacheServer.Port2' );

    $self->testImportExport();
}

__PACKAGE__->runScript;
